IMPLEMENTATION MODULE MathLib;

(* These functions implement the algorithms found in Cheney & Kincaid,
   pp 71-73 and p 96.
*)

CONST   ln2 = 0.69314718;
        pi = 3.1415926535;
        rt2 = 0.70710678;

PROCEDURE  Exp( x : REAL) : REAL;

VAR     s, w, u, z, u2, u4 : REAL;
        m, k : INTEGER;

BEGIN   z := x / ln2;
        IF x < 0.0 THEN m := Trunc(z - 0.5) ELSE m := Trunc(z + 0.5) END;
        w := z - Float(m);
        u := w*ln2;
        u2 := u*u;
        u4 := u2*u2;
        z := (2520.0 + 28.0*u2)/(15120.0 + 420.0*u2 + u4);
        s := 2.0 + u2*z;
        w := (s+u)/(s-u);
        z := 1.0;
        FOR k := 1 TO m DO z := z * 2.0 END;
        FOR k := m TO -1 DO z := z / 2.0 END;
        RETURN z*w;
END Exp;

PROCEDURE  Sin( x : REAL) : REAL;

VAR     a, z, w, u, u2, u4, u6 : REAL;
        m : INTEGER;

BEGIN   z := x/pi;
        IF x < 0.0 THEN m := Trunc(z - 0.5) ELSE m := Trunc(z + 0.5) END;
        w := z - Float(m);
        x := w * pi;
        a := ABS(x);
        IF a < 1.0E-8 THEN RETURN x END;
        IF a > pi/6.0 THEN u := x/3.0 ELSE u := x END;
        u2 := u*u; u4 := u2*u2; u6 := u2*u4;
        w := u*(1.0 - (325523.0/2283996.0)*u2 + (34911.0/7613320.0)*u4 
                + (479249.0/11511339840.0)*u6)/
               (1.0 + (18381.0/761332.0)*u2 + (1261.0/4567992.0)*u4 
                + (2623.0/1644477120.0)*u6);
        IF u <> x THEN w := (3.0 - 4.0 * w*w)*w END;
        IF ODD(m) THEN w := -1.0*w END;
        RETURN w;
END Sin;

PROCEDURE  Cos( x : REAL) : REAL;

BEGIN   RETURN Sin(x + pi/2.0);
END Cos;
        
PROCEDURE  Log( x : REAL) : REAL;

VAR     k : INTEGER;
        u, u2, u4, u6, w : REAL;

BEGIN   x := ABS(x);
        IF x = 0.0 THEN RETURN 0.0 END;  (* ? *)
        IF x = 1.0 THEN RETURN 0.0 END;
        k := 0;
        WHILE x >= 1.0 DO x := x/2.0; INC(k) END;
        WHILE x <  0.5 DO x := x*2.0; DEC(k) END;
        u := (x-rt2)/(x+rt2);
        u2 := u*u; u4 := u2*u2; u6 := u4*u2;
        w := u*(20790.0 - 21545.27*u2 + 4223.9187*u4)/
               (10395.0 - 14237.635*u2 + 4778.8377*u4 - 230.41913*u6);
        RETURN (Float(k) - 0.5)*ln2 + w;
END Log;

PROCEDURE  Sqrt( x : REAL) : REAL;

VAR     t, x0, x1, x2 : REAL;

BEGIN   x := ABS(x);
        IF x = 0.0 THEN RETURN 0.0 END;
        t := 1.0;
        WHILE x >= 1.0 DO x := x / 4.0; t := t * 2.0 END;
        WHILE x < 0.25 DO x := x * 4.0; t := t / 2.0 END;
        x0 := 1.27235367 + 0.242693281*x - 1.02966039/(1.0+x);
        x1 := (x0 + x/x0)/2.0;
        x2 := (x1 + x/x1)/2.0;
        RETURN t*x2;
END Sqrt;

PROCEDURE  Arctan( x : REAL) : REAL;

CONST   r1 = 0.414213562;
        c1 = 0.1963495409;
        t1 = 0.1989123674;
        c2 = 0.5890486226;
        t2 = 0.6681786380;

VAR     n : BOOLEAN;
        u, u2, u4, u6, a, b, c, d, y : REAL;
        k : INTEGER;

BEGIN   IF ABS(x) <= 1.7E-9 THEN RETURN x END;
        IF x < 0.0 THEN x := ABS(x); n := TRUE ELSE n := FALSE END;
        IF x <= 2.0E-2 THEN
          a := x;
          u := 0.0;
          FOR k := 1 TO 4 DO
            u := u + a/Float(2*k-1);
            a := -1.0*a*x*x;
            END;
          IF n THEN u := -1.0*u END;
          RETURN u;
          END;
        IF x <= 1.0 
          THEN y := x; a := 0.0; b := 1.0;
          ELSE y := 1.0/x; a := pi/2.0; b := -1.0;
          END;
        IF y <= r1 
          THEN c := c1; d := t1;
          ELSE c := c2; d := t2;
          END;
        u := (y-d)/(1.0+d*y);
        u2 := u*u; u4 := u2*u2; u6 := u2*u4;
        y := u*((135135.0 + 171962.46*u2 + 52490.4832*u4 + 2218.1*u6)/
               (135135.0 + 217007.46*u2 + 97799.3033*u4 + 10721.3745*u6));
        u := a+b*(c+y);
        IF n THEN u := -1.0*u END;
        RETURN u;
END Arctan;

PROCEDURE  Entier( x : REAL) : INTEGER;

VAR     t : INTEGER;

BEGIN   t := Trunc(x);
        IF Float(t) > x THEN RETURN t-1 ELSE RETURN t END;
END Entier;

PROCEDURE  Float( x : INTEGER) : REAL;

BEGIN   IF x >= 0 
          THEN RETURN FLOAT(CARDINAL(x))
          ELSE RETURN -1.0*FLOAT(CARDINAL(-1*x));
          END;
END Float;

PROCEDURE  Trunc( x : REAL) : INTEGER;

BEGIN   IF x >= 0.0 
          THEN RETURN INTEGER(TRUNC(x))
          ELSE RETURN -1*INTEGER(TRUNC(-1.0*x))
          END;
END Trunc;

PROCEDURE  Round( x : REAL) : INTEGER;

BEGIN   RETURN Entier(x+0.5);
END Round;

END MathLib.
