IMPLEMENTATION MODULE L2Cartridge;  (* QIC-02 Driver for cartridge tape *)

(* This is a synchronous, half-duplex driver for the Archive tape drive.
The driver accepts requests to read or write large blocks of data - which
must be in multiples of 512 bytes.  The tapes are structured with just one
file.  Flush() causes end of file to be written and the tape to be rewound. *)

(* The control port QicWinCB2 on the Stride is a unified port which also
contains bits for the Winchester and Parity control.  LUMOS-2 does not
use those bits at present.  Both L2Cartridge and L2Winchester initialise
the port to zero (so that neither depends on the other's presence). This
port is also write only (!) and cannot be read/rewrite programmed in both
drivers. *)

(* As some operations on the tape can be very long a process may die
while in this driver.  It is not meaningful for more than one process to
access this drive at a time.  By storing the current process id, it is possible
to reject multiple users and to detect aborts. *)


FROM L2Errors IMPORT Status;                                      
FROM L2Kernel IMPORT Semaphore, SemV, SemP, SemInit, Guard, UnGuard,
                     CurrentProcess, Process;
FROM L2Exceptions IMPORT Establish;

CONST   nExcept = 0;     (* Bits on QicIport *)
        nReady  = 1;
        qErase  = 022H;  (* QIC commands    *)
        qRetens = 024H;
        qWrite  = 040H;
        qRead   = 080H;
        qStatus = 0C0H;
        nOnline = 2;     (* WinCB2 bits     *)
        nRqst   = 3;
        nReset  = 4;
        nExcInh = 5;
        nRdyInh = 6;
        sFMark  = 0;     (* Status bits     *)
        sWrProt = 4;
        sCNI    = 6;

TYPE    States = (Cold, Offline, Reading, Writing);

VAR     QicIport  [0FFC788H] : BITSET;
        QicCport  [0FFC782H] : CARDINAL;
        QicSport  [0FFC786H] : CARDINAL;
        QicInDat  [0FFC785H] : CHAR;
        QicOutDat [0FFC781H] : CHAR;
        QicIntClr [0FFC78AH] : CARDINAL;
        QicWinCB2 [0FFC580H] : BITSET;

        Control : BITSET;  (* image of QicWinCB2 *)
        State : States;
        Summary : BITSET;  (* of drive state *)
        ExceptSeen : BOOLEAN;
        TapeGuard, Sync : Semaphore;
        LastProcess : Process;

PROCEDURE  ReadBlock( VAR b : ARRAY OF CHAR) : Status;

(* User's buffer will be rounded down to a 512 byte boundary *)

VAR     i, count, k : CARDINAL;

BEGIN   Guard(TapeGuard);
        IF (LastProcess <> 0) AND (CurrentProcess <> LastProcess) THEN
          UnGuard(TapeGuard);
          RETURN DeviceBusy;
          END;
        LastProcess := CurrentProcess;
        count := (HIGH(b)+1) DIV 512;
        k := 0;
        IF State = Writing THEN
          SetControl({nOnline});
          Wait;
          IF ExceptSeen 
            THEN State := Cold 
            ELSE State := Offline 
            END;
          END;
        IF State = Cold THEN
          Initialise;
          ReadStatus;
          IF NOT (nExcept IN QicIport) THEN 
            UnGuard(TapeGuard);
            RETURN DevError;
            END;
          Retension;
          IF ExceptSeen THEN 
            UnGuard(TapeGuard);
            RETURN DevError;
            END;
          State := Offline;
          END;
        IF State = Offline THEN
          ClrControl({nOnline});
          QicCport := Not(qRead);
          ClrControl({nRqst});
          Wait;
          IF ExceptSeen THEN 
            SetControl({nOnline});
            ReadStatus;
            State := Cold;
            UnGuard(TapeGuard);
            RETURN DevError;
            END;
          SetControl({nRqst});
          State := Reading;
          END;
        IF State = Reading THEN
          WHILE count > 0 DO
            IF NOT (nExcept IN QicIport) THEN
              ReadStatus;
              IF sFMark IN Summary THEN 
                SetControl({nOnline});
                Wait; (* for the rewind *)
                ReadStatus;
                IF ExceptSeen THEN 
                  State := Cold;
                  UnGuard(TapeGuard);
                  RETURN DevError;
                  END;
                State := Offline;
                UnGuard(TapeGuard);
                RETURN EndFile;
                END;
              SetControl({nOnline});
              ReadStatus;
              State := Cold;
              UnGuard(TapeGuard);
              RETURN DevError;
              END;
            Wait;
            FOR i := 0 TO 511 DO b[i+k] := QicInDat END;
            k := k + 512;
            count := count - 1;
            END;
          END;
        UnGuard(TapeGuard);
        RETURN Ok;
END ReadBlock;

PROCEDURE  WriteBlock( VAR b : ARRAY OF CHAR) : Status;

(* User's buffer will be rounded down to a 512 byte boundary *)

VAR     i, count, k : CARDINAL;

BEGIN   Guard(TapeGuard);
        IF (LastProcess <> 0) AND (LastProcess <> CurrentProcess) THEN
          UnGuard(TapeGuard);
          RETURN DeviceBusy;
          END;
        LastProcess := CurrentProcess;
        count := (HIGH(b)+1) DIV 512;
        k := 0;
        IF State = Reading THEN
          SetControl({nOnline});
          Wait;
          IF ExceptSeen 
            THEN State := Cold 
            ELSE State := Offline 
            END;
          END;
        IF State = Cold THEN
          Initialise;
          ReadStatus;
          IF NOT (nExcept IN QicIport) THEN 
            UnGuard(TapeGuard);
            RETURN DevError;
            END;
          Retension;
          IF ExceptSeen THEN
            UnGuard(TapeGuard);
            RETURN DevError;
            END;
          State := Offline;
          END;
        IF State = Offline THEN     
          ReadStatus;
          IF sWrProt IN Summary THEN 
            UnGuard(TapeGuard);
            RETURN Protect;
            END;
          ClrControl({nOnline});
          QicCport := Not(qWrite);
          ClrControl({nRqst});
          Wait;
          IF ExceptSeen THEN 
            SetControl({nOnline});
            ReadStatus;
            State := Cold;
            UnGuard(TapeGuard);
            RETURN DevError;
            END;
          SetControl({nRqst});
          State := Writing;
          END;
        IF State = Writing THEN
          WHILE count > 0 DO
            IF NOT (nExcept IN QicIport) THEN
              SetControl({nOnline});
              ReadStatus;
              State := Cold;
              UnGuard(TapeGuard);
              RETURN DevError 
              END;
            Wait;
            FOR i := 0 TO 511 DO QicOutDat := b[i+k] END;
            k := k + 512;
            count := count - 1;
            END;
          END;
        UnGuard(TapeGuard);
        RETURN Ok;
END WriteBlock;                       

PROCEDURE  Flush() : Status;

BEGIN   Guard(TapeGuard);
        IF (CurrentProcess <> LastProcess) AND (LastProcess <> 0) THEN
          ClrControl(BITSET(0FFH));  (* Abort *)
          State := Cold;
          LastProcess := 0;
          UnGuard(TapeGuard);
          RETURN Ok;
          END;
        IF State <= Offline THEN 
          LastProcess := 0;
          UnGuard(TapeGuard);
          RETURN Ok;
          END;
        IF NOT (nExcept IN QicIport) THEN ReadStatus END; (* Clear Exception *)
        SetControl({nOnline});  (* Set offline *)
        Wait;
        ReadStatus;
        IF ExceptSeen THEN 
          State := Cold;
          UnGuard(TapeGuard);
          RETURN DevError;
          END;   
        State := Offline; 
        LastProcess := 0;
        UnGuard(TapeGuard);
        RETURN Ok;
END Flush;

PROCEDURE  Format() : Status;

BEGIN   Guard(TapeGuard);
        IF (State > Offline) OR (LastProcess <> 0) THEN   (*redundant tests?*)
          UnGuard(TapeGuard);
          RETURN DeviceBusy;
          END;
        LastProcess := CurrentProcess;
        State := Cold;
        Initialise;
        ReadStatus;
        IF sCNI IN Summary THEN 
          UnGuard(TapeGuard); 
          RETURN DevError;
          END;
        IF sWrProt IN Summary THEN 
          UnGuard(TapeGuard); 
          RETURN Protect;
          END;
        QicCport := Not(qErase);
        ClrControl({nRqst});
        Wait;
        SetControl({nRqst});
        Wait;
        IF ExceptSeen THEN 
          UnGuard(TapeGuard); 
          RETURN DevError;
          END;
        State := Offline; 
        UnGuard(TapeGuard);
        RETURN Ok;
END Format;

PROCEDURE  Not( n : CARDINAL ) : CARDINAL;

BEGIN   (*$T-*) RETURN CARDINAL(-1-INTEGER(n)); (*$T=*)
END Not;

PROCEDURE  SetControl( b : BITSET);

BEGIN   Control := Control + b;
        QicWinCB2 := Control;
END SetControl;

PROCEDURE  ClrControl( b : BITSET);

BEGIN   Control := Control - b;
        QicWinCB2 := Control;
END ClrControl;        

PROCEDURE  Initialise;

VAR     i : CARDINAL;

BEGIN   ClrControl(BITSET(0FFH));
        SetControl({nRqst});
        FOR i := 1 TO 30 DO END;
        SetControl({nReset});
END Initialise;

PROCEDURE  Retension;

BEGIN   QicCport := Not(qRetens);
        ClrControl({nRqst});
        Wait;
        SetControl({nRqst});
        Wait;
END Retension;

PROCEDURE  Wait;  (* For interrupt from controller *)

BEGIN   LOOP SetControl({nExcInh,nRdyInh}); (* Enable interrupts *)
             SemP(Sync);
             IF NOT ((nReady IN QicIport) AND (nExcept IN QicIport))
               THEN EXIT END;
             END;
END Wait;

PROCEDURE  ReadStatus;

VAR     i, j  : CARDINAL;
        sb : ARRAY [0..5] OF BITSET;

BEGIN   QicCport := Not(qStatus);
        QicIntClr := 0;
        ClrControl({nRqst});
        Wait;
        SetControl({nRqst});
        FOR i := 0 TO 5 DO
          Wait;
          sb[i] := BITSET(Not(QicSport) MOD 256);
          ClrControl({nRqst});
          FOR j := 1 TO 30 DO END;
          SetControl({nRqst});
          END;
        Wait;
        Summary := sb[0];
END ReadStatus;

MODULE Ints[4];  (* Catch the QIC-02 interrupts and convert to V(Sync) *)

IMPORT  Establish, Sync, SemV, QicIntClr, ClrControl, nExcInh,
        nRdyInh, ExceptSeen;

PROCEDURE QicInt( e : CARDINAL);

BEGIN   QicIntClr := 0;   
        ExceptSeen := e = 1;
        ClrControl({nExcInh,nRdyInh});
        SemV(Sync);
END QicInt;

BEGIN   Establish(61H,QicInt,1);
        Establish(60H,QicInt,0);
END Ints;

BEGIN   ClrControl(BITSET(0FFH));
        SemInit(Sync,0);
        SemInit(TapeGuard,1);
        State := Cold;
        LastProcess := 0;
END L2Cartridge.
