IMPLEMENTATION MODULE L2Clock[4];

FROM L2Exceptions IMPORT IntControl, Establish, Handler;
FROM L2Storage IMPORT ALLOCATE, DEALLOCATE;
FROM L2Errors IMPORT Status;
FROM SYSTEM IMPORT ADDRESS;

TYPE    Event = POINTER TO Ticks;  (* Kludge - compiler won't take ADDRESS *)
        Qptr = POINTER TO Qelt;
        Qelt = RECORD A : Handler;
                      P : CARDINAL;
                      Delta, AlarmTime : Ticks;
                      Next, Prev : Qptr;
                      Index : Event;
               END;

VAR     Now : Time;
        OldCount : Ticks;
        Head, Tail : Qptr;
        EventIndex : ADDRESS;

        PITMode    [0FFC246H] : CARDINAL;
        PITCount0  [0FFC240H] : CARDINAL;

        MonthLengths : ARRAY [0..11] OF CARDINAL;

PROCEDURE  GetTime( VAR t : Time );

BEGIN   Update;
        t := Now;
END GetTime;

PROCEDURE  SetTime( t : Time );

BEGIN   Now := t;
        OldCount := TickCount;
END SetTime;

PROCEDURE  RequestAlarm( a : Handler; param : CARDINAL; delta : Ticks;
                         resched : BOOLEAN; VAR e : Event ) : Status;

(* Arranges for a(param) to be called after delta ticks.  If resched is TRUE
   the alarm will be reset for another delta ticks.  e is a handle in case
   an event must be cancelled. *)

VAR     q : Qptr;

BEGIN   NEW(q);
        e := Event(EventIndex);
        INC(EventIndex);
        IF q = NIL THEN RETURN NoHeap END;
        WITH q^ DO
          Index := e;
          A := a;
          P := param;
          IF resched THEN Delta := delta ELSE Delta := 0 END;
          AlarmTime := TickCount + delta;
          END;
        Insert(q);
        RETURN Ok;
END RequestAlarm;

PROCEDURE  Insert( q : Qptr);

VAR     p : Qptr;

BEGIN   p := Head;
        WHILE (p <> NIL) AND (p^.AlarmTime < q^.AlarmTime) DO
          p := p^.Next END;
        IF p = NIL 
          THEN q^.Prev := Tail;
               q^.Next := NIL;
               IF Tail = NIL 
                 THEN Head := q
                 ELSE Tail^.Next := q
                 END;
               Tail := q;
          ELSE q^.Next := p;
               q^.Prev := p^.Prev;
               IF p^.Prev = NIL
                 THEN Head := q
                 ELSE p^.Prev^.Next := q
                 END;
               p^.Prev := q;
          END;
END Insert;

PROCEDURE  Remove( q : Qptr);

BEGIN   IF q^.Prev = NIL
          THEN Head := q^.Next
          ELSE q^.Prev^.Next := q^.Next
          END;
        IF q^.Next = NIL
          THEN Tail := q^.Prev
          ELSE q^.Next^.Prev := q^.Prev
          END
END Remove;

PROCEDURE  CancelAlarm( VAR e : Event) : Status;

VAR     q : Qptr;

BEGIN   q := Head;
        WHILE (q <> NIL) AND (q^.Index <> e) DO q := q^.Next END;
        IF q = NIL THEN RETURN NoSuch END;
        Remove(q);
        DISPOSE(q);
        RETURN Ok;
END CancelAlarm;

PROCEDURE  Ring;

VAR     p : Qptr; 

BEGIN   p := Head; 
        WHILE (Head <> NIL) AND (Head^.AlarmTime <= TickCount) DO
          p := Head;
          Remove(p);
          WITH p^ DO
            IF Delta <> 0 THEN
              AlarmTime := TickCount + Delta;
              Insert(p);
              END;
            A(P);  (*  Sound the alarm *) 
            IF Delta = 0 THEN DISPOSE(p) END;
            END;
          END;
END Ring;

PROCEDURE  ClockInt;            (* The PIT interrupt handler *)

BEGIN   EXCL(IntControl,4); (* Turn off interrupt latch *)
        INCL(IntControl,4); (* Re-enable interrupt *)
        INC(TickCount);
        IF Head <> NIL THEN Ring END;
END ClockInt;

PROCEDURE  Update;  (*  Corrects Now *)

VAR     tc, ot : Ticks;
        d, m, y, ml : CARDINAL;

BEGIN   tc := TickCount;
        ot := OldCount;
        OldCount := tc;
        tc := tc - ot;
        WITH Now DO
          y := day DIV 512;
          day := day MOD 512;
          m := day DIV 32;
          d := day MOD 32;
          INC(tc,csec);
          csec := CARDINAL(tc MOD 6000);  (*  Csecs per min *)
          tc := (tc DIV 6000) + Ticks(min);
          min := CARDINAL(tc MOD 1440);  (* Mins per day *)
          INC(d,CARDINAL(tc DIV 1440));
          LOOP ml := MonthLengths[m];
               IF (y MOD 4 = 0) AND (m = 2) THEN ml := 29 END;
               IF d > ml 
                 THEN INC(m);
                      DEC(d,ml);
                 ELSE EXIT END;
               y := y + ( m DIV 12 );
               m := m MOD 12;
               END;
          day := ((y * 16) + m) * 32 + d;
          END;
END Update;

BEGIN   MonthLengths[0] := 31;
        MonthLengths[1] := 28;
        MonthLengths[2] := 31;
        MonthLengths[3] := 30;
        MonthLengths[4] := 31;
        MonthLengths[5] := 30;
        MonthLengths[6] := 31;
        MonthLengths[7] := 31;
        MonthLengths[8] := 30;
        MonthLengths[9] := 31;
        MonthLengths[10] := 30;
        MonthLengths[11] := 31;

        INCL(IntControl,4);  (* Enable PIT0 *) 
        EXCL(IntControl,7);  (* Disable PIT2 *)
        Establish(46H,Handler(ClockInt),0);
        PITMode := 034H;
        PITCount0 := 640 MOD 256;
        PITCount0 := 640 DIV 256;
        WITH Now DO
          csec := 0;
          min := 0;
          day := 1;  (* 1-Jan-1980 00:00:00 *)
          END;
        TickCount := 0;
        OldCount := 0;
        Head := NIL;
        Tail := NIL;
        EventIndex := 1;
END L2Clock.
