IMPLEMENTATION MODULE L2Files;           (* Implements the LUMOS ODS *)

FROM L2Exceptions IMPORT Assert;
FROM L2Errors IMPORT Status;
FROM L2Strings IMPORT Substring, Compare, Copy, Find, Remove, Length;
FROM L2Clock IMPORT Time, GetTime;
FROM L2Kernel IMPORT Process, CurrentProcess, Resume, ProcMax,
                     Semaphore, SemV, SemP, SemInit;
FROM L2Storage IMPORT ALLOCATE, DEALLOCATE;
FROM L2Logicals IMPORT Translate;
FROM SYSTEM IMPORT ADR;
IMPORT L2Device;
FROM L2Cache IMPORT ReadCache, PurgeBlock, PurgeVolume;

CONST   MaxFiles = 100;
        MaxVols = 6;  (* simultaneously mounted *)
        EOF = 65535;
        FATStart = 3;  (* PBN of block 0 of the FAT *)

TYPE    NameString = ARRAY [0..19] OF CHAR;
        PathNameString = ARRAY [0..79] OF CHAR;
        DeviceName = ARRAY [0..9] OF CHAR;
        FATBlock = ARRAY [0..255] OF CARDINAL;
        DirectoryRecord = RECORD Name : NameString;
                                 First : CARDINAL;
                                 Date : Time;
                                 Owner : CARDINAL;
                                 Mode : FileMode;     (* 32 Bytes total *)
                          END;
        HomeBlock = RECORD Key : ARRAY [0..3] OF CHAR;
                           Label : NameString;
                           Date : Time;
                           Size : CARDINAL;
                           Owner : CARDINAL;
                           Mode : FileMode;
                    END;
        DIRBlock = ARRAY [0..15] OF DirectoryRecord;  (* 512 bytes *)
        VCBPointer = POINTER TO VCB;
        FATPointer = POINTER TO FATBuffer;
        FATBuffer = RECORD Dirty : BOOLEAN;
                           Volume, RefCount, Sector : CARDINAL;
                           Next, Prev : FATPointer;
                           CASE BOOLEAN OF
                             TRUE  : Data : FATBlock |
                             FALSE : DataA : ARRAY [0..511] OF CHAR;
                             END;
                    END;
        VCB = RECORD Device : DeviceName;
                     FATSectors, Rover, FileCount, User, FATCount : CARDINAL;
                     Mounted, Private, ReadOnly : BOOLEAN;
                     Head, Tail : FATPointer;
                     CASE BOOLEAN OF
                       TRUE  : Home : HomeBlock |
                       FALSE : HomeA : ARRAY [0..35] OF CHAR;
                       END;
              END;
        FCB = RECORD Volume : CARDINAL;
                     ProcNum : Process;
                     FullName : PathNameString;
                     DirEntry : DirectoryRecord;
                     DirSector, DirRecNum : CARDINAL;
                     CurrentPBN, PrevPBN, CurrentLBN, NextLBN : CARDINAL;
                     NFS, DirDirty, Write : BOOLEAN;
                     CurrentFAT : FATPointer;
                     DevAttribs : L2Device.DevAttrSet;
              END;
        FCBPointer = POINTER TO FCB;
 
VAR     OpenFiles : ARRAY [0..MaxFiles-1] OF FCBPointer;
        FileCount : CARDINAL;
        Volumes : ARRAY [1..MaxVols] OF VCB;
        Guard : Semaphore;
        Hyphen : ARRAY [0..1] OF CHAR;
        NewName : NameString;  (* Communication between Access and Enter *)
        Attachments : ARRAY Process OF PathNameString;

PROCEDURE  Open( VAR fn : ARRAY OF CHAR; write : BOOLEAN; 
                 VAR h : Handle ) : Status;

VAR     fcb : FCBPointer;
        i : CARDINAL;
        rs, ig : Status;
        devonly : BOOLEAN;
        acm : FileMode;

BEGIN   SemP(Guard);
        IF FileCount > MaxFiles THEN SemV(Guard); RETURN NoSlot END;
        h := 0;
        WHILE OpenFiles[h] <> NIL DO INC(h) END;
        NEW(OpenFiles[h]);
        fcb := OpenFiles[h];
        IF fcb = NIL THEN SemV(Guard); RETURN NoHeap END;
        WITH fcb^ DO
          Write := write;
          CurrentFAT := NIL;
          rs := Expand(fn,FullName);
          IF rs <> Ok THEN
            DISPOSE(OpenFiles[h]);
            SemV(Guard); 
            RETURN rs;
            END;
          i := 0;
          WHILE (i < MaxFiles) AND 
                ((OpenFiles[i] = NIL) OR (i = h) OR
                (Compare(OpenFiles[i]^.FullName,FullName) <> 0)) 
            DO INC(i) END;
          IF i < MaxFiles
            THEN IF (OpenFiles[i]^.Write OR write) AND NOT OpenFiles[i]^.NFS
                   THEN DISPOSE(OpenFiles[h]); 
                        SemV(Guard); 
                        RETURN Locked; 
                        END;
                 IF NOT OpenFiles[i]^.NFS THEN  (* Can we touch it ? *)
                   IF write THEN acm := FileMode{ORd,WRd,OWr,WWr}
                            ELSE acm := FileMode{ORd,WRd} END;
                   WITH OpenFiles[i]^.DirEntry DO
                     rs := CheckMode(Owner,Users[CurrentProcess],Mode,acm);
                     IF rs <> Ok THEN
                       DISPOSE(OpenFiles[h]);
                       SemV(Guard);
                       RETURN rs;
                       END;
                     END;
                   END;
                 fcb^ := OpenFiles[i]^;  (* Copy FCB *)
                 CurrentFAT := NIL; (* Can't copy FAT pointers *)
                 IF NOT fcb^.NFS THEN 
                   INC(Volumes[Volume].FileCount);  (* Easy to forget *)
                   END;
            ELSE devonly := Find(Hyphen,FullName) = Length(FullName);
                 rs := FindVCB(FullName,Volume,TRUE);
                 IF (rs = BadDevice) OR (rs = NotMounted)
                   THEN IF NOT devonly THEN
                          DISPOSE(OpenFiles[h]);
                          SemV(Guard); 
                          RETURN rs;
                          END;
                        NFS := TRUE;
                        DirEntry.Owner := 0;
                        DirEntry.Mode := FileMode(0FFH);
                   ELSE IF devonly THEN
                          DISPOSE(OpenFiles[h]);
                          SemV(Guard); 
                          RETURN BadName;
                          END;
                        IF Volumes[Volume].ReadOnly AND write THEN
                          DISPOSE(OpenFiles[h]);
                          SemV(Guard); 
                          RETURN NoWrite;
                          END;
                        NFS := FALSE;
                        rs := Access(fcb);
                        IF rs <> Ok THEN 
                          IF CurrentFAT <> NIL 
                            THEN ig := Release(CurrentFAT) END;
                          DISPOSE(OpenFiles[h]);
                          SemV(Guard); 
                          RETURN rs;
                          END;
                        INC(Volumes[Volume].FileCount);
                   END;
            END;
          INC(FileCount);
          CurrentLBN := 0;
          NextLBN := 0;
          CurrentPBN := DirEntry.First;
          PrevPBN := EOF;
          ProcNum := CurrentProcess;
          rs := L2Device.GetDevAttr(FullName,i,DevAttribs);
          DirDirty := FALSE;
          END;
        SemV(Guard); 
        RETURN Ok;
END Open;

PROCEDURE  Create( VAR fn : ARRAY OF CHAR; mode : FileMode; 
                   VAR h : Handle ) : Status;

CONST   sysmodes = FileMode{SYS,DIR};
        nullmode = FileMode{};

VAR     fcb : FCBPointer;
        i : CARDINAL;
        rs, ig : Status;

BEGIN   IF SYS IN mode THEN RETURN NoPriv END;
        SemP(Guard);
        IF FileCount > MaxFiles THEN SemV(Guard); RETURN NoSlot END;
        h := 0;
        WHILE OpenFiles[h] <> NIL DO INC(h) END;
        NEW(OpenFiles[h]);
        fcb := OpenFiles[h];
        IF fcb = NIL THEN SemV(Guard); RETURN NoHeap END;
        WITH fcb^ DO
          rs := Expand(fn,FullName);
          IF rs <> Ok THEN 
            DISPOSE(OpenFiles[h]);
            SemV(Guard); 
            RETURN rs;
            END;
          CurrentFAT := NIL;
          Write := TRUE;
          i := 0;
          WHILE (i < MaxFiles) AND 
                ((OpenFiles[i] = NIL) OR (i = h) OR
                (Compare(OpenFiles[i]^.FullName,FullName) <> 0)) 
            DO INC(i) END;
          IF i < MaxFiles
            THEN DISPOSE(OpenFiles[h]); 
                 SemV(Guard); 
                 RETURN Locked; 
                 END;
          IF Find(Hyphen,FullName) = Length(FullName) 
            THEN DISPOSE(OpenFiles[h]);
                 SemV(Guard); 
                 RETURN BadName;
                 END;
          rs := FindVCB(FullName,Volume,TRUE);
          IF rs <> Ok THEN
            DISPOSE(OpenFiles[h]);
            SemV(Guard); 
            RETURN rs;
            END;
          IF Volumes[Volume].ReadOnly THEN
            DISPOSE(OpenFiles[h]);
            SemV(Guard); 
            RETURN NoWrite;
            END;
          NFS := FALSE;
          rs := Access(fcb);
          IF rs <> Ok 
            THEN IF rs = NoFile THEN rs := Enter(fcb,mode) END;
                 IF rs <> Ok THEN 
                   IF CurrentFAT <> NIL 
                     THEN ig := Release(CurrentFAT) END;
                   DISPOSE(OpenFiles[h]);
                   SemV(Guard); 
                   RETURN rs;
                   END;
            ELSE IF (DirEntry.Mode * sysmodes) <> nullmode THEN
                   IF CurrentFAT <> NIL 
                     THEN ig := Release(CurrentFAT) END;
                   DISPOSE(OpenFiles[h]);
                   SemV(Guard); 
                   RETURN NoPriv;
                   END;
                 IF DirEntry.First <> EOF THEN
                   rs := Discard(Volume,DirEntry.First);
                   DirEntry.First := EOF;
                   GetTime(DirEntry.Date);
                   DirDirty := TRUE;
                   END;
            END;
          INC(Volumes[Volume].FileCount);
          INC(FileCount);
          CurrentLBN := 0;
          NextLBN := 0;
          CurrentPBN := EOF;
          PrevPBN := EOF;
          ProcNum := CurrentProcess;
          ig := L2Device.GetDevAttr(FullName,i,DevAttribs);
          END;
        SemV(Guard); 
        RETURN rs;
END Create;

PROCEDURE  Close( h : Handle; save : BOOLEAN) : Status;

CONST   delmode = FileMode{ODe,WDe};
        sysmodes = FileMode{SYS,DIR};
        nullmode = FileMode{};

VAR     rs, rs1, rs2, rs3 : Status;
        fcb : FCBPointer;
        u : CARDINAL;

BEGIN   SemP(Guard);
        rs := ValidateHandle(h,fcb);
        u := Users[CurrentProcess];
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        rs := Ok; rs1 := Ok; rs2 := Ok; rs3 := Ok;
        WITH fcb^ DO
          IF NOT NFS 
            THEN IF Write AND NOT save THEN
                   WITH DirEntry DO 
                     rs := CheckMode(Owner,u,Mode,delmode);
                     IF rs <> Ok THEN SemV(Guard); RETURN rs END;
                     IF Mode * sysmodes <> nullmode THEN 
                       SemV(Guard); 
                       RETURN NoPriv;
                       END;
                     END;
                   IF DirEntry.First <> EOF THEN
                     rs := Discard(Volume,DirEntry.First) END;
                   DirEntry.Name := "";
                   DirDirty := TRUE;
                   END;
                 IF Write AND DirDirty THEN rs1 := WriteBackDir(fcb) END;
                 DEC(Volumes[Volume].FileCount);
                 IF CurrentFAT <> NIL THEN rs2 := Release(CurrentFAT) END;
                 IF Write AND (L2Device.FLSH IN DevAttribs) THEN 
                   rs3 := L2Device.Flush(FullName) END;
            ELSE IF L2Device.FLSH IN DevAttribs THEN
                   IF NOT (L2Device.BLK IN DevAttribs)
                     THEN SemV(Guard) END;
                   rs := L2Device.Flush(FullName);
                   IF NOT (L2Device.BLK IN DevAttribs)
                     THEN SemP(Guard) END;
                   END;
            END;
          IF Write THEN Signal(h) END;
          END;
        DISPOSE(OpenFiles[h]);
        DEC(FileCount);
        SemV(Guard); 
        IF rs1 <> Ok THEN RETURN rs1 END;
        IF rs2 <> Ok THEN RETURN rs2 END;
        IF rs3 <> Ok THEN RETURN rs3 END;
        RETURN rs;
END Close;

PROCEDURE  Delete( VAR fn : ARRAY OF CHAR) : Status;

VAR     h : Handle;
        rs, ig : Status;

BEGIN   rs := Open(fn,TRUE,h);
        IF rs = Ok THEN 
          rs := Close(h,FALSE); 
          IF rs <> Ok THEN ig := Close(h,TRUE) END;
          END;
        RETURN rs;
END Delete;

PROCEDURE  Rename( VAR old, new : ARRAY OF CHAR) : Status;

VAR     ofcb, nfcb : FCBPointer;
        oFCB, nFCB : FCB;
        i : CARDINAL;
        rs, ig : Status;

BEGIN   SemP(Guard);
        ofcb := ADR(oFCB);
        nfcb := ADR(nFCB);
        WITH ofcb^ DO
          rs := Expand(old,FullName);
          IF rs <> Ok THEN SemV(Guard); RETURN rs END;
          CurrentFAT := NIL;
          Write := TRUE;
          i := 0;
          WHILE (i < MaxFiles) AND 
                ((OpenFiles[i] = NIL) OR 
                 (Compare(OpenFiles[i]^.FullName,FullName) <> 0))
            DO INC(i) END;
          IF i < MaxFiles THEN SemV(Guard); RETURN Locked END;
          IF Find(Hyphen,FullName) = Length(FullName) THEN 
            SemV(Guard); 
            RETURN BadName;
            END;
          rs := FindVCB(FullName,Volume,TRUE);
          IF rs <> Ok THEN SemV(Guard); RETURN rs END;
          rs := Access(ofcb);
          IF CurrentFAT <> NIL THEN ig := Release(CurrentFAT) END;
          IF rs <> Ok THEN SemV(Guard); RETURN rs END;
          IF SYS IN DirEntry.Mode THEN SemV(Guard); RETURN NoPriv END;
          END;  (* Old file now set up *)
        WITH nfcb^ DO
          rs := Expand(new,FullName);
          IF rs <> Ok THEN SemV(Guard); RETURN rs END;
          CurrentFAT := NIL;
          Write := FALSE;
          i := 0;
          WHILE (i < MaxFiles) AND 
                ((OpenFiles[i] = NIL) OR 
                 (Compare(OpenFiles[i]^.FullName,FullName) <> 0))
            DO INC(i) END;
          IF i < MaxFiles THEN SemV(Guard); RETURN Locked END;
          IF Find(Hyphen,FullName) = Length(FullName) THEN 
            SemV(Guard); 
            RETURN BadName;
            END;
          rs := FindVCB(FullName,Volume,TRUE);
          IF rs <> Ok THEN SemV(Guard); RETURN rs END;
          i := Length(ofcb^.FullName);
          IF (Find(ofcb^.FullName,FullName) = 0) AND
             (Length(FullName) > i) AND
             (FullName[i] = '-') THEN SemV(Guard); RETURN BadName END;
          IF ofcb^.Volume <> Volume THEN SemV(Guard); RETURN OffVolume END;
          IF Volumes[Volume].ReadOnly THEN SemV(Guard); RETURN NoWrite END;
          rs := Access(nfcb);
          IF rs = Ok THEN 
            IF CurrentFAT <> NIL THEN 
              rs := Release(CurrentFAT) END;
            SemV(Guard); 
            RETURN Exists;
            END;
          rs := Enter(nfcb,ofcb^.DirEntry.Mode);
          IF rs <> Ok THEN 
            IF CurrentFAT <> NIL THEN 
              ig := Release(CurrentFAT) END;
            SemV(Guard); 
            RETURN rs;
            END;
          IF CurrentFAT <> NIL THEN
            rs := Release(CurrentFAT) END;
          WITH DirEntry DO
            Mode := ofcb^.DirEntry.Mode;
            Owner := ofcb^.DirEntry.Owner;
            Date := ofcb^.DirEntry.Date;
            First := ofcb^.DirEntry.First;
            END;
          END;  (* New file now set up *)
        ofcb^.DirEntry.Name[0] := 0C;
        rs := WriteBackDir(nfcb);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        rs := WriteBackDir(ofcb);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;                 
        IF L2Device.FLSH IN ofcb^.DevAttribs THEN
          rs := L2Device.Flush(ofcb^.FullName) END;
        SemV(Guard); 
        RETURN rs;
END Rename;

PROCEDURE  GetMode( h : Handle; VAR date : Time; 
                    VAR mode : FileMode; VAR owner : CARDINAL ) : Status;

VAR     rs : Status;
        fcb : FCBPointer;

BEGIN   SemP(Guard);
        rs := ValidateHandle(h,fcb);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        WITH fcb^ DO
          IF NFS THEN SemV(Guard); RETURN BadFunction END;
          WITH DirEntry DO
            date := Date;
            mode := Mode;
            owner := Owner;
            END;
          END;
        SemV(Guard); 
        RETURN Ok;
END GetMode;
          
PROCEDURE  ViewHandle( h : Handle; VAR p : Process; 
                       VAR fn : ARRAY OF CHAR ) : Status;

BEGIN   IF h >= MaxFiles THEN RETURN NoSuch END;
        SemP(Guard);
        IF OpenFiles[h] = NIL THEN
          SemV(Guard); RETURN BadHandle END;
        WITH OpenFiles[h]^ DO
          IF (Users[CurrentProcess] <> Users[ProcNum]) AND
             (Users[CurrentProcess] <> 0) THEN
             SemV(Guard); RETURN NoPriv END;
          p := ProcNum;
          Copy(FullName,fn);
          END;
        SemV(Guard);
        RETURN Ok;
END ViewHandle;

PROCEDURE  SetMode( h : Handle; mode : FileMode; owner : CARDINAL) : Status;

VAR     rs : Status;
        fcb : FCBPointer;

BEGIN   IF SYS IN mode THEN RETURN NoPriv END;
        SemP(Guard);
        rs := ValidateHandle(h,fcb);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        WITH fcb^ DO
          IF NFS THEN SemV(Guard); RETURN BadFunction END;
          IF NOT Write THEN SemV(Guard); RETURN NoPriv END;
          IF SYS IN DirEntry.Mode THEN SemV(Guard); RETURN NoPriv END;
          IF (DIR IN DirEntry.Mode) AND NOT (DIR IN mode) AND NOT DirEmpty(fcb)
            THEN SemV(Guard); RETURN NoPriv END;
          IF NOT (DIR IN DirEntry.Mode) AND (DIR IN mode)
            THEN SemV(Guard); RETURN NoPriv END;
          IF Users[CurrentProcess] = 0 THEN 
            DirEntry.Owner := owner;
            DirDirty := TRUE;
            END;
          IF (DirEntry.Owner = Users[CurrentProcess]) 
            OR (Users[CurrentProcess] = 0)
            THEN DirEntry.Mode := mode;
                 DirDirty := TRUE;
            ELSE SemV(Guard); RETURN NoPriv
            END;
          END;
        SemV(Guard); 
        RETURN Ok;
END SetMode;

PROCEDURE SetFileDate( h : Handle; VAR t : Time) : Status;

VAR     rs : Status;
        fcb : FCBPointer;

BEGIN   SemP(Guard);
        rs := ValidateHandle(h,fcb);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        WITH fcb^ DO
          IF NFS THEN SemV(Guard); RETURN BadFunction END;
          IF NOT Write THEN SemV(Guard); RETURN NoPriv END;
          IF SYS IN DirEntry.Mode THEN SemV(Guard); RETURN NoPriv END;
          IF (DirEntry.Owner = Users[CurrentProcess]) 
            OR (Users[CurrentProcess] = 0)
            THEN DirEntry.Date := t;
                 DirDirty := TRUE;
            ELSE SemV(Guard); RETURN NoPriv
            END;
          END;
        SemV(Guard); 
        RETURN Ok;
END SetFileDate;

PROCEDURE  Size( h : Handle; VAR size : CARDINAL) : Status;

VAR     rs : Status;
        fcb : FCBPointer;
        e : BOOLEAN;

BEGIN   SemP(Guard);
        rs := ValidateHandle(h,fcb);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        WITH fcb^ DO
          IF NFS THEN SemV(Guard); RETURN BadFunction END;
          size := 0;
          LOOP rs := SetPosition(fcb,size,FALSE,e);
               IF (rs <> Ok) OR (CurrentPBN = EOF) THEN EXIT END;
               INC(size);
               END;
          END;
        IF rs = EndFile THEN rs := Ok END;
        SemV(Guard); 
        RETURN rs;
END Size;

PROCEDURE  FastSize( VAR dn : ARRAY OF CHAR; first : CARDINAL;
                     VAR size : CARDINAL) : Status;

(* An alternate way of getting the size of a file if you can read the
   directory and obtain the PBN of LBN 0 *)

VAR     rs : Status;
        vi, i, b : CARDINAL;
        fat : FATPointer;
        dvn : PathNameString;

BEGIN   SemP(Guard);
        rs := Expand(dn,dvn);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        rs := FindVCB(dvn,vi,TRUE);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        size := 0;
        i := first;
        fat := NIL;
        LOOP IF i = EOF THEN EXIT END;
             IF size = 65535 THEN EXIT END;
             IF (i = 0) AND (first <> 0) THEN EXIT END;
             b := i DIV 256;
             IF (fat = NIL) OR (b <> fat^.Sector) THEN
               IF fat <> NIL THEN
                 rs := Release(fat);
                 IF rs <> Ok THEN EXIT END;
                 END;
               IF b >= Volumes[vi].FATSectors THEN EXIT END;
               rs := Fetch(vi,b,fat);
               IF rs <> Ok THEN EXIT END;
               END;
             INC(size);
             i := fat^.Data[i MOD 256];
             END;
        IF fat <> NIL THEN rs := Release(fat) END;
        IF i = 0 THEN rs := BadVolume END;  (* sizing unallocated block *)
        IF size = 65535 THEN rs := BadVolume END; (* "infinite" loop *)
        SemV(Guard);
        RETURN rs;
END FastSize;

PROCEDURE  ReadBlock( h : Handle; VAR b : ARRAY OF CHAR) : Status;

VAR     rs : Status;
        fcb : FCBPointer;
        e : BOOLEAN;

BEGIN   SemP(Guard);
        rs := ValidateHandle(h,fcb);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        WITH fcb^ DO
          IF NOT (L2Device.RD IN DevAttribs)
            THEN SemV(Guard); RETURN NoFunction END;
          IF NFS 
            THEN CurrentPBN := NextLBN
            ELSE rs := SetPosition(fcb,NextLBN,FALSE,e);
                 IF rs <> Ok THEN SemV(Guard); RETURN rs END;
            END;
          IF NOT (L2Device.BLK IN DevAttribs) THEN SemV(Guard) END; 
          rs := L2Device.ReadBlock(FullName,CurrentPBN,b);
          INC(NextLBN);
          IF (L2Device.BLK IN DevAttribs) THEN SemV(Guard) END; 
          END;            (* The BLK attribute is overloaded ! *)
        RETURN rs;
END ReadBlock;

PROCEDURE  WriteBlock( h : Handle; VAR b : ARRAY OF CHAR; 
                       trunc : BOOLEAN) : Status;

VAR     rs : Status;
        fcb : FCBPointer;
        e : BOOLEAN;
        tb : CARDINAL;

BEGIN   SemP(Guard);
        rs := ValidateHandle(h,fcb);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        WITH fcb^ DO
          IF NOT (L2Device.WR IN DevAttribs) 
            THEN SemV(Guard); RETURN NoFunction END;
          IF NOT Write THEN SemV(Guard); RETURN NoPriv END;
          IF NFS 
            THEN CurrentPBN := NextLBN
            ELSE IF DIR IN DirEntry.Mode
                   THEN SemV(Guard); RETURN NoPriv END;
                 rs := SetPosition(fcb,NextLBN,TRUE,e);
                 IF rs <> Ok THEN SemV(Guard); RETURN rs END;
            END;
          IF NOT (L2Device.BLK IN DevAttribs) THEN SemV(Guard) END; 
          rs := L2Device.WriteBlock(FullName,CurrentPBN,b);
          IF NOT (L2Device.BLK IN DevAttribs) THEN SemP(Guard) END; 
          INC(NextLBN);
          IF rs <> Ok THEN SemV(Guard); RETURN rs END;
          IF trunc AND NOT NFS THEN 
            rs := SetPosition(fcb,NextLBN,FALSE,e);
            IF rs = EndFile THEN SemV(Guard); RETURN Ok END;
            IF rs <> Ok THEN SemV(Guard); RETURN rs END;
            WITH CurrentFAT^ DO
              tb := Data[PrevPBN MOD 256];
              Data[PrevPBN MOD 256] := EOF;
              Dirty := TRUE;
              END;
            IF tb <> EOF THEN rs := Discard(Volume,tb); END;
            CurrentPBN := EOF;
            IF rs <> Ok THEN SemV(Guard); RETURN rs END;
            END;
          END;
        SemV(Guard); 
        RETURN Ok;
END WriteBlock;

PROCEDURE  Seek( h : Handle; blocknum : CARDINAL) : Status;

VAR     rs : Status;
        fcb : FCBPointer;
        e : BOOLEAN;

BEGIN   SemP(Guard); 
        rs := ValidateHandle(h,fcb);
        IF rs = Ok THEN fcb^.NextLBN := blocknum END;
        SemV(Guard); 
        RETURN rs;
END Seek;

PROCEDURE  GetDevAttr( VAR dn : ARRAY OF CHAR; VAR devsize : CARDINAL;
                       VAR attr : L2Device.DevAttrSet ) : Status;

VAR     dvn : PathNameString;
        rs : Status;

BEGIN   rs := Expand(dn,dvn);
        IF rs <> Ok THEN RETURN rs END;
        RETURN L2Device.GetDevAttr(dvn,devsize,attr);
END GetDevAttr;

PROCEDURE  FormatTrack( h : Handle; track : CARDINAL) : Status;

VAR     rs : Status;
        fcb : FCBPointer;

BEGIN   SemP(Guard); 
        rs := ValidateHandle(h,fcb);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        WITH fcb^ DO
          IF NOT NFS OR NOT Write THEN rs := NoPriv END;
          IF NOT (L2Device.BLK IN DevAttribs) THEN SemV(Guard) END;
          IF rs = Ok THEN rs := L2Device.FormatTrack(FullName,track); END;
          IF (L2Device.BLK IN DevAttribs) THEN SemV(Guard) END;
          END;
        RETURN rs;
END FormatTrack;

PROCEDURE  Mount( VAR dn, label : ARRAY OF CHAR; private : BOOLEAN) : Status;

CONST   Mro = FileMode { ORd, WRd };
        Mrw = FileMode { ORd, WRd, OWr, WWr };

VAR     vi, fr, i : CARDINAL;
        rs : Status;
        at : L2Device.DevAttrSet;
        dvn : PathNameString;

BEGIN   SemP(Guard); 
        rs := Expand(dn,dvn);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        IF Length(dvn) <> Find(Hyphen,dvn) 
          THEN SemV(Guard); RETURN BadName END;
        i := 0;
        WHILE (i < MaxFiles) AND 
              ((OpenFiles[i] = NIL) OR 
              (Compare(OpenFiles[i]^.FullName,dvn) <> 0)) DO INC(i) END;
        IF i < MaxFiles THEN SemV(Guard); RETURN DeviceBusy END;
        rs := L2Device.GetDevAttr(dvn,i,at);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        IF NOT (L2Device.BLK IN at) THEN SemV(Guard); RETURN BadDevice END;
        fr := 0;
        vi := 1;
        LOOP IF NOT Volumes[vi].Mounted THEN fr := vi END;
             IF Compare(dvn,Volumes[vi].Device) = 0 THEN EXIT END;
             INC(vi);
             IF vi > MaxVols THEN
               IF fr = 0 THEN SemV(Guard); RETURN NoSlot END;
               vi := fr;
               Copy(dvn,Volumes[vi].Device);
               EXIT;
               END;
             END;           (* Now we have a VCB for this device *)
        WITH Volumes[vi] DO
          IF Mounted THEN SemV(Guard); RETURN DeviceBusy END;
          Mounted := TRUE;
          ReadOnly := FALSE;
          User := Users[CurrentProcess];
          Private := private;
          rs := L2Device.ReadBlock(dvn,128,HomeA);  (* trick to clear cache *)
          IF rs = Ok THEN
            rs := L2Device.ReadBlock(dvn,2,HomeA); END; (* Get HomeBlock *)
          IF rs = Ok THEN
            IF Compare(Home.Key,"HOME") <> 0
              THEN rs := BadVolume;
            ELSIF (Compare(Home.Label,label) <> 0) AND (User <> 0)
              THEN rs := NoPriv
            ELSE rs := CheckMode(Home.Owner,User,Home.Mode,Mrw);
                 IF rs <> Ok THEN
                   ReadOnly := TRUE;
                   rs := CheckMode(Home.Owner,User,Home.Mode,Mro);
                   END;
              END;
            END;
          IF rs <> Ok THEN Mounted := FALSE END;
          Rover := 0;
          FileCount := 0;
          FATSectors := (Home.Size+255) DIV 256;
          Head := NIL;
          Tail := NIL;
          FATCount := 0;
          END;
        SemV(Guard); 
        RETURN rs;
END Mount;

PROCEDURE  Dismount( VAR dn : ARRAY OF CHAR) : Status;

VAR     rs : Status;
        vi : CARDINAL;
        dvn : PathNameString;

BEGIN   SemP(Guard); 
        rs := Expand(dn,dvn);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        rs := FindVCB(dvn,vi,FALSE);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        WITH Volumes[vi] DO
          IF FileCount <> 0 THEN SemV(Guard); RETURN VolBusy END;
          Mounted := FALSE;
          Assert(FATCount=0,201);
          PurgeVolume(vi);
          END;
        SemV(Guard); 
        RETURN Ok;
END Dismount;

PROCEDURE  FreeSpace( VAR dn : ARRAY OF CHAR; VAR space : CARDINAL) : Status;

VAR     rs : Status;
        vi, i : CARDINAL;
        fat : FATPointer;

BEGIN   SemP(Guard); 
        rs := FindVCB(dn,vi,TRUE);
        IF rs <> Ok THEN SemV(Guard); RETURN rs END;
        WITH Volumes[vi] DO
          space := 0;
          i := 0;
          LOOP IF (i MOD 256) = 0 THEN
                 rs := Fetch(vi,i DIV 256,fat);
                 IF rs <> Ok THEN EXIT END;
                 END;
               IF fat^.Data[i MOD 256] = 0 THEN INC(space) END;
               INC(i);
               IF (i = Home.Size) OR ((i MOD 256) = 0) THEN
                 rs := Release(fat);
                 IF rs <> Ok THEN EXIT END;
                 IF i = Home.Size THEN EXIT END;
                 END;
               END;
          END;
        SemV(Guard); 
        RETURN rs;
END FreeSpace;

PROCEDURE  PurgeFiles( proc : Process); 

VAR     h : Handle;

BEGIN   IF (proc-1 < ProcMax) THEN Attachments[proc][0] := 0C END;
        FOR h := 0 TO MaxFiles-1 DO 
          IF (OpenFiles[h] <> NIL) AND (OpenFiles[h]^.ProcNum = proc) THEN
            OpenFiles[h]^.ProcNum := CurrentProcess;
            SpecialClose(h);
            END;
          END;
END PurgeFiles;

PROCEDURE  Stall;  (* Allows a process to synchronise with the file system *)

BEGIN   SemP(Guard);
END Stall;

PROCEDURE  UnStall;

BEGIN   SemV(Guard);
END UnStall;

PROCEDURE  Expand( VAR sn, tn : ARRAY OF CHAR) : Status;

VAR     i : CARDINAL;
        ch : CHAR;
        rs : Status;

BEGIN   FOR i := 0 TO Length(sn)-1 DO
          ch := CAP(sn[i]);
          IF NOT ((ch = '#') OR (ch = '$') OR (ch = '-') 
                             OR (ch = '.') OR (ch = '^') OR
             ((ch >= '0') AND (ch <= '9')) OR ((ch >= 'A') AND (ch <= 'Z')))
            THEN RETURN BadName END;
          IF ((ch = '#') OR (ch = '$')) AND (i <> 0) THEN RETURN BadName END;
          END;
        rs := Translate(sn,tn);
        IF rs <> Ok THEN RETURN rs END;
        IF tn[0] <> '#' THEN RETURN BadName END;
        RETURN Ok;
END Expand;

PROCEDURE  ValidateHandle( h : Handle; VAR fcb : FCBPointer) : Status;

BEGIN   IF (h >= MaxFiles) OR (OpenFiles[h] = NIL) OR
           (OpenFiles[h]^.ProcNum <> CurrentProcess) 
          THEN RETURN BadHandle END;
        fcb := OpenFiles[h];
        RETURN Ok;
END ValidateHandle;

PROCEDURE  Access( fcb : FCBPointer) : Status;

(* Assumes that the FullName has already been expanded and checked for access
   conflicts. Parses the file name and traverses the directory tree. *)

CONST   search = FileMode{OEx,WEx};
        accread = FileMode{ORd,WRd};
        accwrite = FileMode{OWr,WWr};

VAR     efn, comp : PathNameString;
        rs : Status;
        dn : BOOLEAN;
        u : CARDINAL;
        accmode : FileMode;

BEGIN   u := Users[CurrentProcess];
        accmode := accread;
        WITH fcb^ DO
          Copy(FullName,efn);
          IF NOT Component(efn,comp) THEN RETURN BadName END;
          rs := PointMFD(fcb);
          IF rs <> Ok THEN RETURN BadVolume END;
          LOOP WITH DirEntry DO rs := CheckMode(Owner,u,Mode,search) END;
               IF rs <> Ok THEN RETURN rs END;
               IF NOT Component(efn,comp) THEN RETURN BadName END;
               Copy(comp,NewName);  (* saved for Enter *)
               rs := LookUp(fcb,comp);
               dn := Compare(efn,"") = 0; 
               IF (rs = Ok) AND dn 
                 THEN IF Write THEN accmode := accmode + accwrite END;
                      WITH DirEntry DO 
                        rs := CheckMode(Owner,u,Mode,accmode) END;
                      RETURN rs
               ELSIF (rs = NoFile) AND NOT dn 
                 THEN RETURN NoDirect
               ELSIF rs <> Ok 
                 THEN RETURN rs
                 END;
               END;
          END;
END Access;

PROCEDURE  Component( VAR fn, comp : ARRAY OF CHAR) : BOOLEAN;

VAR     cp : CARDINAL;

BEGIN   cp := Find(Hyphen,fn);
        IF cp+1 = Length(fn) THEN RETURN FALSE END;  (* ends in hyphen *)
        Substring(fn,comp,0,cp);
        Remove(fn,0,cp+1);
        cp := Length(comp);
        RETURN (cp > 0) AND (cp <= 20);
END Component;

PROCEDURE  PointMFD( fcb : FCBPointer) : Status;

(* Sets up an FCB to point to the Master File Directory on a volume *)

VAR     mfd : CARDINAL;

BEGIN   WITH fcb^ DO
          mfd := Volumes[Volume].FATSectors + FATStart;
          DirEntry.Mode := FileMode{OEx,WEx,DIR};  (* temporary *)
          DirEntry.First := mfd;
          DirDirty := FALSE;  (* MFD is never empty ! *)
          CurrentPBN := mfd;
          PrevPBN := EOF;
          CurrentLBN := 0;
          END;
        RETURN LookUp(fcb,"DIRECTORY.SYS");  (* to get mode *)
END PointMFD;

PROCEDURE  LookUp( fcb : FCBPointer; VAR name : ARRAY OF CHAR) : Status;

(* Searches the directory on fcb for name, and resets the fcb to that
   entry *)

VAR     d : RECORD CASE BOOLEAN OF
              TRUE  : a : ARRAY [0..511] OF CHAR |
              FALSE : d : DIRBlock;
              END;
            END;
        rs : Status;
        bl, rc : CARDINAL;
        e : BOOLEAN;
        dvn : DeviceName;

BEGIN   WITH fcb^ DO
          IF NOT (DIR IN DirEntry.Mode) THEN RETURN BadDirect END;
          bl := 0;
          dvn := Volumes[Volume].Device;
          LOOP rs := SetPosition(fcb,bl,FALSE,e);
               IF (rs <> Ok) AND (rs <> EndFile) THEN RETURN rs END;
               IF CurrentPBN = EOF THEN RETURN NoFile END;
               rs := ReadCache(dvn,Volume,CurrentPBN,d.a);
               IF rs <> Ok THEN RETURN rs END;
               FOR rc := 0 TO 15 DO
                 IF Compare(d.d[rc].Name,name) = 0 THEN
                   DirEntry := d.d[rc];
                   DirSector := CurrentPBN;
                   DirRecNum := rc;
                   CurrentPBN := DirEntry.First;
                   PrevPBN := EOF;
                   CurrentLBN := 0;
                   RETURN Ok;
                   END;
                 END;
               INC(bl);
               END;
          END;
END LookUp;

PROCEDURE  Enter( fcb : FCBPointer; mode : FileMode) : Status;

(* should be called only if LookUp fails during a create *)

CONST   accwrite = FileMode{OWr,WWr};

VAR     d : RECORD CASE BOOLEAN OF
              TRUE  : a : ARRAY [0..511] OF CHAR |
              FALSE : d : DIRBlock;
              END;
            END;
        rs : Status;
        bl, rc, u : CARDINAL;
        extend, savedwrite : BOOLEAN;
        dvn : DeviceName;

BEGIN   WITH fcb^ DO
          IF NOT (DIR IN DirEntry.Mode) THEN RETURN BadDirect END;
          u := Users[CurrentProcess];
          WITH DirEntry DO rs := CheckMode(Owner,u,Mode,accwrite) END;
          IF rs <> Ok THEN RETURN rs END;
          bl := 0;
          dvn := Volumes[Volume].Device;
          LOOP rs := SetPosition(fcb,bl,TRUE,extend);
               IF rs <> Ok THEN RETURN rs END;
               IF extend 
                 THEN FOR rc := 0 TO 511 DO d.a[rc] := 0C END;
                      IF DirDirty THEN  (* the directory was empty *)
                        rs := WriteBackDir(fcb);
                        IF rs <> Ok THEN RETURN rs END;
                        END;
                 ELSE rs := L2Device.ReadBlock(dvn,CurrentPBN,d.a);
                      IF rs <> Ok THEN RETURN rs END;
                 END;
               FOR rc := 0 TO 15 DO
                 IF d.d[rc].Name[0] = 0C THEN
                   WITH DirEntry DO
                     Name := NewName;
                     Mode := mode;
                     First := EOF;
                     GetTime(Date);
                     Owner := Users[CurrentProcess];
                     END;
                   d.d[rc] := DirEntry;
                   rs := L2Device.WriteBlock(dvn,CurrentPBN,d.a);
                   PurgeBlock(Volume,CurrentPBN);
                   DirSector := CurrentPBN;
                   DirRecNum := rc;
                   RETURN rs;
                   END;
                 END;
               INC(bl);
               END;
          END;
END Enter;

PROCEDURE  DirEmpty( fcb : FCBPointer) : BOOLEAN;

VAR     d : RECORD CASE BOOLEAN OF
              TRUE  : a : ARRAY [0..511] OF CHAR |
              FALSE : d : DIRBlock;
              END;
            END;
        rs : Status;
        bl, rc : CARDINAL;
        e, empty : BOOLEAN;
        dvn : DeviceName;

BEGIN   WITH fcb^ DO
          bl := 0;
          empty := TRUE;
          dvn := Volumes[Volume].Device;
          LOOP rs := SetPosition(fcb,bl,FALSE,e);
               IF (rs <> Ok) AND (rs <> EndFile)
                 THEN empty := FALSE; EXIT END;
               IF CurrentPBN = EOF THEN EXIT END;
               rs := L2Device.ReadBlock(dvn,CurrentPBN,d.a);
               PurgeBlock(Volume,CurrentPBN);
               IF rs <> Ok THEN empty := FALSE; EXIT END;
               FOR rc := 0 TO 15 DO
                 IF d.d[rc].Name[0] <> 0C THEN empty := FALSE; EXIT END;
                 END;
               INC(bl);
               END;
          END;
        RETURN empty;
END DirEmpty;

PROCEDURE  WriteBackDir( fcb : FCBPointer) : Status;

(* Used by close *)

VAR     d : RECORD CASE BOOLEAN OF
              TRUE  : a : ARRAY [0..511] OF CHAR |
              FALSE : d : DIRBlock;
              END;
            END;
        rs : Status;
        dvn : DeviceName;

BEGIN   WITH fcb^ DO
          dvn := Volumes[Volume].Device;
          rs := L2Device.ReadBlock(dvn,DirSector,d.a);
          IF rs <> Ok THEN RETURN rs END;
          d.d[DirRecNum] := DirEntry;
          rs := L2Device.WriteBlock(dvn,DirSector,d.a);
          PurgeBlock(Volume,DirSector);
          END;
        RETURN rs;
END WriteBackDir;

PROCEDURE  SetPosition( VAR fcb : FCBPointer; lbn : CARDINAL; extend : BOOLEAN;
                        VAR extended : BOOLEAN) : Status;

(*  Sets the fcb to the indicated block number, extending the file if
    necessary and requested *)

VAR     fbn, fsn : CARDINAL;
        tmpfat : FATPointer;
        rs : Status;

BEGIN   extended := FALSE;
        WITH fcb^ DO
          IF lbn < CurrentLBN THEN (* reset the file *)
            CurrentPBN := DirEntry.First;
            CurrentLBN := 0;
            PrevPBN := EOF;
            END;
          LOOP
            IF CurrentPBN = EOF THEN
              IF NOT extend THEN RETURN EndFile END;
              rs := GetFree(Volume,fbn,tmpfat);
              IF rs <> Ok THEN RETURN rs END;
              IF PrevPBN = EOF
                THEN DirEntry.First := fbn;
                     DirDirty := TRUE;
                ELSE Assert(CurrentFAT <> NIL,203);
                     CurrentFAT^.Data[PrevPBN MOD 256] := fbn;
                END;
              IF CurrentFAT <> NIL THEN
                CurrentFAT^.Dirty := CurrentFAT <> tmpfat;
                rs := Release(CurrentFAT);
                IF rs <> Ok THEN RETURN rs END;
                END;
              CurrentFAT := tmpfat;
              CurrentPBN := fbn;
              CurrentFAT^.Data[fbn MOD 256] := EOF;
              CurrentFAT^.Dirty := TRUE;
              extended := TRUE;
              END;
            IF CurrentLBN = lbn THEN EXIT END;
            fsn := CurrentPBN DIV 256;
            IF (CurrentFAT = NIL) OR (fsn <> CurrentFAT^.Sector) THEN
              IF CurrentFAT <> NIL THEN
                rs := Release(CurrentFAT);
                IF rs <> Ok THEN RETURN rs END;
                END;
              IF fsn >= Volumes[Volume].FATSectors
                THEN RETURN BadVolume END;
              rs := Fetch(Volume,fsn,CurrentFAT);
              IF rs <> Ok THEN RETURN rs END;
              END;
            PrevPBN := CurrentPBN;
            CurrentPBN := CurrentFAT^.Data[CurrentPBN MOD 256];
            INC(CurrentLBN);
            END;
          RETURN Ok;
          END;
END SetPosition; 

PROCEDURE  GetFree( vi : CARDINAL; VAR fbn : CARDINAL; 
                    VAR tmpfat : FATPointer) : Status;

VAR     rs : Status;

BEGIN   WITH Volumes[vi] DO
          fbn := Rover;
          rs := Fetch(vi,fbn DIV 256,tmpfat);
          IF rs <> Ok THEN RETURN rs END;
          WHILE tmpfat^.Data[fbn MOD 256] <> 0 DO
            fbn := (fbn + 1) MOD Home.Size;
            IF fbn = Rover THEN 
              rs := Release(tmpfat);
              IF rs <> Ok THEN RETURN rs END;
              RETURN NoBlocks;
              END;
            IF (fbn MOD 256) = 0 THEN
              rs := Release(tmpfat);
              IF rs <> Ok THEN RETURN rs END;
              rs := Fetch(vi,fbn DIV 256,tmpfat);
              IF rs <> Ok THEN RETURN rs END;
              END;
            END;   (* If we pass here, fbn is a free block *)
          Rover := fbn;
          END;
        RETURN Ok;
END GetFree;

PROCEDURE  Discard( vol, dbn : CARDINAL) : Status;

VAR     tmpfat : FATPointer;
        next : CARDINAL;
        rs : Status;

BEGIN   rs := Fetch(vol,dbn DIV 256,tmpfat);
        IF rs <> Ok THEN RETURN rs END;
        LOOP WITH tmpfat^ DO
               next := Data[dbn MOD 256];
               Data[dbn MOD 256] := 0;
               Dirty := TRUE;
               END;
             IF next = EOF THEN EXIT END;
             IF (next DIV 256) <> (dbn DIV 256) THEN 
               rs := Release(tmpfat);
               IF rs <> Ok THEN RETURN rs END;
               rs := Fetch(vol,next DIV 256,tmpfat);
               IF rs <> Ok THEN RETURN rs END;
               END;
             dbn := next;
             END;
        rs := Release(tmpfat);
        RETURN rs;
END Discard;

PROCEDURE  Fetch( vi, sn : CARDINAL; VAR fat : FATPointer) : Status;

VAR     rs : Status;

BEGIN   WITH Volumes[vi] DO
          fat := Head;
          WHILE (fat <> NIL) AND (fat^.Sector <> sn) DO fat := fat^.Next END;
          IF fat = NIL THEN
            NEW(fat);
            IF fat = NIL THEN RETURN NoHeap END;
            rs := L2Device.ReadBlock(Device,sn+FATStart,fat^.DataA);
            IF rs <> Ok THEN DISPOSE(fat); RETURN rs END;
            WITH fat^ DO
              Next := NIL;
              Prev := Tail;
              IF Head = NIL THEN Head := fat ELSE Tail^.Next := fat END;
              Tail := fat;
              Sector := sn;
              Dirty := FALSE;
              RefCount := 0;
              Volume := vi;
              END;
            END;
          INC(fat^.RefCount);
          INC(FATCount);
          END;
        RETURN Ok;
END Fetch;

PROCEDURE  Release( VAR fat : FATPointer) : Status;

VAR     rs : Status;

BEGIN   IF fat = NIL THEN RETURN NoSuch END;
        rs := Ok;
        WITH fat^ DO
          WITH Volumes[Volume] DO
            IF Dirty THEN
              rs := L2Device.WriteBlock(Device,Sector+FATStart,DataA);
              IF rs = Ok THEN Dirty := FALSE END;
              END;
            DEC(RefCount);
            DEC(FATCount);
            IF RefCount = 0 THEN
              IF Next = NIL THEN Tail := Prev ELSE Next^.Prev := Prev END;
              IF Prev = NIL THEN Head := Next ELSE Prev^.Next := Next END;
              DISPOSE(fat);
              END;
            END;
          END;
        fat := NIL;
        RETURN rs;
END Release;

PROCEDURE  FindVCB( VAR dn : ARRAY OF CHAR; VAR vi : CARDINAL;
                    public : BOOLEAN ) : Status;

VAR     rs : Status;
        u : CARDINAL;
        at : L2Device.DevAttrSet;
        dvn : DeviceName;

BEGIN   vi := 1;
        u := Find(Hyphen,dn);
        Substring(dn,dvn,0,u);
        rs := L2Device.GetDevAttr(dvn,u,at);
        IF rs <> Ok THEN RETURN rs END;
        IF NOT (L2Device.BLK IN at) THEN RETURN BadDevice END;
        WHILE (vi <= MaxVols) AND ( NOT Volumes[vi].Mounted OR 
               ( Compare(Volumes[vi].Device,dvn) <> 0 )) DO INC(vi) END;
        IF vi > MaxVols THEN RETURN NotMounted END;
        u := Users[CurrentProcess];
        WITH Volumes[vi] DO
          IF (u <> 0) AND (u <> User) AND (Private OR NOT public)
            THEN RETURN NoPriv END;
          END;
        RETURN Ok;
END FindVCB;

PROCEDURE  CheckMode( owner, user : CARDINAL;  mode, need : FileMode) : Status;

BEGIN   IF user = 0 THEN RETURN Ok END;  (* SuperUser *)
        IF owner = user 
          THEN need := need - FileMode{WRd,WWr,WEx,WDe}
          ELSE need := need - FileMode{ORd,OWr,OEx,ODe} 
          END;
        need := need - mode;
        IF need = FileMode{} 
          THEN RETURN Ok
          ELSE RETURN NoAccess
          END;
END CheckMode;

PROCEDURE  SpecialClose( h : Handle);  (* for Purging *)

VAR     rs : Status;
        fcb : FCBPointer;

BEGIN   rs := ValidateHandle(h,fcb);
        IF rs <> Ok THEN RETURN END;
        WITH fcb^ DO
          IF NOT NFS 
            THEN IF Write AND DirDirty THEN rs := WriteBackDir(fcb) END;
                 DEC(Volumes[Volume].FileCount);
                 IF CurrentFAT <> NIL THEN rs := Release(CurrentFAT) END;
                 IF Write AND (L2Device.FLSH IN DevAttribs) THEN 
                   rs := L2Device.Flush(FullName) END;
            ELSE IF L2Device.FLSH IN DevAttribs
                   THEN rs := L2Device.Flush(FullName) END;
            END;
          IF Write THEN Signal(h) END;
          END;
        DISPOSE(OpenFiles[h]);
        DEC(FileCount);
END SpecialClose;

PROCEDURE  AttachTree( VAR root : ARRAY OF CHAR) : Status;

BEGIN   RETURN Translate(root,Attachments[CurrentProcess]);
END AttachTree;

PROCEDURE  DetachTree;

BEGIN   Attachments[CurrentProcess][0] := 0C;
END DetachTree;

PROCEDURE  Signal( h : Handle ); (* Resumes any eligible process *)

VAR     p : Process;
        rs : Status;

BEGIN   WITH OpenFiles[h]^ DO
          FOR p := 1 TO ProcMax DO
            IF Attachments[p][0] <> 0C THEN
              IF Find(Attachments[p],FullName) = 0 THEN
                WITH DirEntry DO
                  IF ((Users[p] = Owner) AND (ORd IN Mode)) OR (WRd IN Mode)
                    THEN rs := Resume(p) 
                    END;
                  END;
                END;
              END;
            END;
          END;
END Signal;

VAR     i : CARDINAL;

BEGIN   FOR i := 1 TO MaxVols DO 
          WITH Volumes[i] DO
            Mounted := FALSE;
            Device := "";
            END;
          END;
        FOR i := 1 TO ProcMax DO Attachments[i][0] := 0C END;
        FOR i := 0 TO MaxFiles-1 DO OpenFiles[i] := NIL END;
        FileCount := 0;
        Hyphen[0] := "-";  Hyphen[1] := 0C;
        SemInit(Guard,1);
END L2Files.
