IMPLEMENTATION MODULE L2Floppy;         (* Driver for the 765 controller *)

(* The controller uses interrupts but not DMA.  Transfers to and from the
   disk medium are in units of 8 block tracks.  A track buffer is maintained
   in the driver for each unit. *)

FROM SYSTEM IMPORT WORD, ADDRESS, ADR;
FROM L2Exceptions IMPORT Establish, Handler, Vectors, IntControl;
FROM L2Clock IMPORT RequestAlarm, CancelAlarm, Event, Ticks;
FROM L2Kernel IMPORT Semaphore, SemV, SemP, SemQ, SemInit;
FROM L2Errors IMPORT Status;
FROM L2Machine IMPORT TestBit, BeqS, MoveBWL, MoveBLW, Add1, RTE, JMP;

CONST   MaxUnit = 1;

TYPE    Units = [0..MaxUnit];
        Block = ARRAY [0..511] OF CHAR;
        Header = ARRAY [0..3] OF CHAR;  (* C, H, S, N *)

VAR     FloppyAux     [0FFC1C6H] : BITSET;  (* read only  *)
        MasterStatus  [0FFC2C0H] : BITSET;  (* of 765     *)
        FloppyData    [0FFC2C2H] : WORD;    (* of 765     *)
        TransferInt   [400H]  : ARRAY [0..19] OF CARDINAL;

        HData : ARRAY [1..8] OF Header;  (* for formatting *)
        Escape, Initialised, MotorOn : BOOLEAN;
        Drive : ARRAY Units OF 
                  RECORD Track : ARRAY [0..7] OF Block;
                         TrackNum : CARDINAL;
                         Dirty, Recalibrated, WriteProtect : BOOLEAN;
                  END;
        SavedProc : PROC;
        MotorOff : Event;

PROCEDURE  TimeOutAlarm;

BEGIN   Escape := TRUE;
END TimeOutAlarm;

PROCEDURE  SetTimeOut( t : Ticks; VAR e : Event) : Status;

BEGIN   Escape := FALSE;
        RETURN RequestAlarm(Handler(TimeOutAlarm),0,t,FALSE,e);
END SetTimeOut;

PROCEDURE  ApplyCommand( VAR com : ARRAY OF WORD) : Status;

VAR     r : Status;
        i, t : CARDINAL;
        e : Event;

BEGIN   r := SetTimeOut(100,e);
        IF r <> Ok THEN RETURN r END;
        FOR i := 0 TO HIGH(com) DO
          WHILE NOT (7 IN MasterStatus) AND NOT Escape DO END;
          IF NOT (7 IN MasterStatus) THEN RETURN TimedOut END;
          IF 6 IN MasterStatus THEN
            r := CancelAlarm(e);
            RETURN DevError 
            END;
          FloppyData := com[i];
          FOR t := 1 TO 4 DO END;  (* 24 micro-secs *)
          END;
        r := CancelAlarm(e);
        RETURN Ok;
END ApplyCommand;

PROCEDURE  ReadResult( VAR res : ARRAY OF WORD) : Status;

VAR     r : Status;
        i, t : CARDINAL;
        e : Event;

BEGIN   r := SetTimeOut(100,e);
        IF r <> Ok THEN RETURN r END;
        FOR i := 0 TO HIGH(res) DO
          WHILE NOT (7 IN MasterStatus) AND NOT Escape DO END;
          IF NOT (7 IN MasterStatus) THEN RETURN TimedOut END;
          IF NOT (6 IN MasterStatus) THEN 
            r := CancelAlarm(e);
            RETURN DevError 
            END;
          res[i] := FloppyData;
          FOR t := 1 TO 4 DO END;  (* 24 micro-secs *)
          END;
        r := CancelAlarm(e);
        RETURN Ok;
END ReadResult;

PROCEDURE  SeekStatus() : Status;

VAR     r : Status;
        SIScom : CARDINAL;
        status : RECORD ST0 : BITSET;
                        PCN : CARDINAL;
                 END;

BEGIN   SIScom := 8; (* Sense interrupt status *)
        r := ApplyCommand(SIScom);
        IF r <> Ok THEN RETURN r END;
        r := ReadResult(status);
        IF r <> Ok THEN RETURN r END;
        WITH status DO
          IF ST0 * {6,7,4} = {} THEN RETURN Ok
            ELSIF 7 IN ST0 THEN RETURN IllegalOp 
            ELSIF 6 IN ST0 THEN RETURN SeekFail
            ELSE RETURN SeekFail
            END;
          END;
END SeekStatus;

PROCEDURE  CompletionStatus() : Status;

VAR     r : Status;
        status : RECORD ST0, ST1, ST2 : BITSET;
                        C, H, R, N : CARDINAL;
                 END;

BEGIN   r := ReadResult(status);
        IF r <> Ok THEN RETURN r END;
        WITH status DO
          IF ST0 * {6,7} = {} THEN RETURN Ok
            ELSIF 7 IN ST0 THEN RETURN IllegalOp
            ELSIF 5 IN ST1 THEN RETURN CRCError
            ELSIF 4 IN ST1 THEN RETURN OverRun
            ELSIF 7 IN ST1 THEN RETURN EndTrack 
            ELSE RETURN BadFormat
            END;
          END;
END CompletionStatus;

PROCEDURE  Recalibrate( unit : CARDINAL) : Status;

VAR     recal : ARRAY [0..1] OF CARDINAL;
        r : Status;
        retries : CARDINAL;
 
BEGIN   recal[0] := 7; recal[1] := unit;  (* Recalibrate *)
        retries := 0;
        LOOP SetControl({2});
             r := ApplyCommand(recal);
             IF r <> Ok THEN EXIT END;
             r := Wait(100);
             IF r <> Ok THEN EXIT END;
             r := SeekStatus();
             IF (r = Ok) OR (retries = 3) THEN EXIT END;
             INC(retries);
             END;
        RETURN r;
END Recalibrate;

PROCEDURE  Seek( unit, head, cyl : CARDINAL ) : Status;

VAR     seek : ARRAY [0..2] OF CARDINAL;
        r : Status;
        retries : CARDINAL;
 
BEGIN   seek[0] := 15; 
        seek[1] := 4*head + unit;
        seek[2] := cyl;
        retries := 0;
        LOOP SetControl({2});
             r := ApplyCommand(seek);
             IF r <> Ok THEN EXIT END;
             r := Wait(100);
             IF r <> Ok THEN EXIT END;
             r := SeekStatus();
             IF r = Ok THEN EXIT END;
             IF retries = 0 THEN 
               r := Recalibrate(unit);
               IF r <> Ok THEN EXIT END;
               END;
             IF retries = 3 THEN EXIT END;
             INC(retries);
             END;
        RETURN r;
END Seek;

PROCEDURE  Transfer( read : BOOLEAN; unit, track : CARDINAL) : Status; 

VAR     tcom : RECORD op, hd, C, H, R, N,
                      EOT, GPL, DTL : CARDINAL;
               END;
        r : Status;
        h, c, retries, mode : CARDINAL;

BEGIN   r := Select(unit);
        IF r <> Ok THEN RETURN r END;
        IF NOT read AND Drive[unit].WriteProtect
          THEN Deselect(unit,Protect); RETURN Protect END;
        c := track DIV 2;
        h := track MOD 2;
        IF c >= 40 THEN SetControl({6}) ELSE ClearControl({6}) END;
        r := Seek(unit,h,c);
        IF r <> Ok THEN Deselect(unit,r); RETURN r END;
        WITH tcom DO
          IF read THEN op := 46H ELSE op := 45H END;
          hd := 4*h + unit;
          C := c;
          H := h;
          R := 1;
          N := 2;
          EOT := 8;
          GPL := 42;
          DTL := 255;
          END;
        retries := 0;
        LOOP IF read THEN mode := 0 ELSE mode := 1 END;
             SetHandler(mode,ADR(Drive[unit].Track));
             SetControl({2});
             r := ApplyCommand(tcom);
             IF r <> Ok THEN EXIT END;
             r := Wait(100);
             IF r <> Ok THEN EXIT END;
             RemoveHandler;
             r := CompletionStatus();
             IF r = EndTrack THEN r := Ok END;
             IF (r <> Ok) AND (retries = 2) THEN EXIT END;
             IF (r = Ok) AND NOT read THEN (* check the write *)
               tcom.op := 46H;  (* Read *)
               SetHandler(2,0); (* Readcheck *)
               SetControl({2});
               r := ApplyCommand(tcom);
               IF r <> Ok THEN EXIT END;
               r := Wait(100);
               IF r <> Ok THEN EXIT END;
               RemoveHandler;
               r := CompletionStatus();
               IF r = EndTrack THEN r := Ok END;
               IF (r <> Ok) AND (retries = 2) THEN EXIT END;
               END;
             IF r = Ok THEN EXIT END;
             INC(retries);
             END;
        Deselect(unit,r); 
        RETURN r;
END Transfer;

PROCEDURE  FormatTrack( unit, track : CARDINAL ) : Status;

VAR     fcom : RECORD op, hd, N, SC, GPL, D : CARDINAL;
               END;
        r : Status;
        i, c, h, retries : CARDINAL;

BEGIN   IF unit > MaxUnit THEN RETURN NoSuch END;
        IF track >= 160 THEN RETURN NoTrack END;
        Drive[unit].TrackNum := 65535;
        Drive[unit].Dirty := FALSE;
        r := Select(unit);
        IF r <> Ok THEN RETURN r END;
        IF Drive[unit].WriteProtect 
          THEN Deselect(unit,Protect); RETURN Protect END;
        c := track DIV 2;
        h := track MOD 2;
        FOR i := 1 TO 8 DO
          HData[i][0] := CHR(c);
          HData[i][1] := CHR(h);
          HData[i][2] := CHR(i);
          HData[i][3] := CHR(2);  (* 512 byte sectors *)
          END;
        IF c >= 40 THEN SetControl({6}) ELSE ClearControl({6}) END;
        r := Seek(unit,h,c);
        IF r <> Ok THEN Deselect(unit,r); RETURN r END;
        WITH fcom DO
          op := 4DH;
          hd := h*4 + unit;
          N := 2;
          SC := 8;
          GPL := 80;
          D := 0C5H;  (* 10100101 *)
          END;
        retries := 0;
        LOOP SetHandler(1,ADR(HData));
             SetControl({2});
             r := ApplyCommand(fcom);
             IF r <> Ok THEN EXIT END;
             r := Wait(100);
             IF r <> Ok THEN EXIT END;
             RemoveHandler;
             r := CompletionStatus();
             IF (r = Ok) OR (retries = 1) THEN EXIT END;
             INC(retries);
             END;
        Deselect(unit,r); 
        RETURN r;
END FormatTrack;

PROCEDURE  Initialise() : Status; 

VAR     e : Event;
        r : Status;
        i : CARDINAL;
        scom : ARRAY [0..2] OF CARDINAL;

BEGIN   RemoveHandler;
        SetControl({1}); (* Turn RDY on *)
        FOR i := 1 TO 10 DO END;
        SetControl({7});  (* FRES *)
        FOR i := 1 TO 10 DO END;
        ClearControl({7});  (* Turn off reset *)
        SetControl({2});
        r := Wait(100);
        IF r <> Ok THEN RETURN DevError END;
        r := SeekStatus();  (* ignore it, this is just to clear the int *)
        scom[0] := 3;
        scom[1] := 0E1H;  (* SRT = 4 ms, HUT = 32 ms, HLT = 12 ms *)
        scom[2] := 7;
        r := ApplyCommand(scom);
        IF r <> Ok THEN RETURN r END;
        Initialised := TRUE;
        RETURN Ok;
END Initialise;

PROCEDURE  ShutOffMotor;

BEGIN   MotorOn := FALSE;
        SetControl({5});
END ShutOffMotor;

PROCEDURE  Select( unit : CARDINAL ) : Status;

VAR     e : Event;
        r : Status;

BEGIN   IF NOT Initialised THEN 
          r := Initialise();
          IF r <> Ok THEN RETURN r END;
          END;
        IF unit = 0 THEN ClearControl({3}) ELSE ClearControl({4}) END;
        IF MotorOn 
          THEN r := CancelAlarm(MotorOff)
          ELSE ClearControl({5});
               MotorOn := TRUE;
               r := Wait(25);
          END;
        IF NOT Drive[unit].Recalibrated THEN
          r := Recalibrate(unit);
          IF r <> Ok THEN Deselect(unit,r); RETURN r END;
          Drive[unit].Recalibrated := TRUE;
          END;
        Drive[unit].WriteProtect := NOT (1 IN FloppyAux);
        RETURN Ok;
END Select;

PROCEDURE  Deselect( unit : CARDINAL; errorcode : Status);

VAR     r : Status;

BEGIN   IF (errorcode = TimedOut) OR (errorcode = BadFormat) OR
           (errorcode = DevError) OR (errorcode = SeekFail) OR
           (errorcode = IllegalOp) THEN
          WITH Drive[unit] DO
            Recalibrated := FALSE;
            Dirty := FALSE;
            TrackNum := 65535;
            END;
          Initialised := FALSE;
          ClearControl({1,2});
          END;
        SetControl({3,4});
        r := RequestAlarm(Handler(ShutOffMotor),0,1500,FALSE,MotorOff);
        IF r <> Ok THEN ShutOffMotor END;
END Deselect;

(* To handle the IPL 6 interrupts from the floppy fast enough, we borrow
   an idea from the Stride BIOS -- a special self-modifying handler is
   located in the empty part of page 0.  This handler fields execution 
   phase interrupts but passes control to the normal routine if we are in 
   the result phase.  The following routine sets up this special handler *)

PROCEDURE  SetHandler( mode : CARDINAL; bufaddr : ADDRESS);

(*  Modes:  0 = Read,  1 = Write,  2 = ReadCheck (do not alter buffer) *)

BEGIN   TransferInt[0] := TestBit;
        TransferInt[1] := 5;
        TransferInt[2] := 0C2C1H;  (* MSR, Will sign extend *)
        TransferInt[3] := BeqS + 14;
        CASE mode OF
          0 :  TransferInt[4] := MoveBWL;
               TransferInt[5] := 0C2C3H;   (* FDR *)
               TransferInt[6] := CARDINAL(bufaddr DIV 10000H);
               TransferInt[7] := CARDINAL(bufaddr MOD 10000H);
               TransferInt[8] := Add1;
               TransferInt[9] := 400H + 12;                      |
          1 :  TransferInt[4] := MoveBLW;
               TransferInt[5] := CARDINAL(bufaddr DIV 10000H);
               TransferInt[6] := CARDINAL(bufaddr MOD 10000H);
               TransferInt[7] := 0C2C3H;
               TransferInt[8] := Add1;
               TransferInt[9] := 400H + 10;                      |
          2 :  TransferInt[4] := MoveBWL;
               TransferInt[5] := 0C2C3H;
               TransferInt[6] := 0;
               TransferInt[7] := 400H + 28; (* Out of harm's way *)
               TransferInt[8] := RTE;
          END;
        TransferInt[10] := RTE;
        TransferInt[11] := JMP;
        TransferInt[12] := CARDINAL(ADDRESS(SavedProc) DIV 10000H);
        TransferInt[13] := CARDINAL(ADDRESS(SavedProc) MOD 10000H);
        Vectors[52H] := PROC(400H);
END SetHandler;

PROCEDURE  RemoveHandler;

BEGIN    Vectors[52H] := SavedProc;
END RemoveHandler;

(* The floppy interrupt handler needs to do a V operation, but it is 
   running at IPL 6 and if the kernel runs that high, then the transfer
   interrupts will get missed ! So, we use the 'software interrupt' 
   provided by the Stride.  On the IntControl port bits 1 and 2 high
   cause an interrupt on vector 53H at IPL 1 .  This can be cleared by 
   lowering either bit.  We leave 2 high, and set or clear 1.  *)

MODULE  Level6[6];

  IMPORT  IntControl;

  EXPORT FloppyInt, SetControl, ClearControl;
 
  VAR     FloppyControl [0FFC202H] : BITSET;  (* write only *)
          FControl : BITSET;

  PROCEDURE  SetControl( m : BITSET );

  BEGIN   FControl := FControl + m;
          FloppyControl := FControl;
  END SetControl;

  PROCEDURE  ClearControl( m : BITSET );

  BEGIN   FControl := FControl - m;
          FloppyControl := FControl;
  END ClearControl;

  PROCEDURE  FloppyInt;

  BEGIN   ClearControl({2});  (* Switch off Interrupt *)
          INCL(IntControl,1);
  END FloppyInt;

BEGIN   FControl := {3,4,5};  (* No drive selected, interrupts off *)
        FloppyControl := FControl;
END Level6;

MODULE  Level4[4];   (* A mini-kernel to supply the special timed wait *)

  IMPORT  Semaphore, SemV, SemP, SemQ, SemInit, CancelAlarm, RequestAlarm,
          Ticks, Status, Event, ClearControl, IntControl;

  EXPORT  Wait, Signal;

  VAR   Result : Status;
        Sync : Semaphore;
        e : Event;
        Pending : BOOLEAN;

  PROCEDURE  Wait( t : Ticks ) : Status;

  VAR   r : Status;

  BEGIN IF SemQ(Sync) <= 0 THEN
          r := RequestAlarm(Signal,CARDINAL(TimedOut),t,FALSE,e);
          IF r <> Ok 
            THEN ClearControl({2});
                 Result := r;
                 IF 1 IN IntControl THEN SemP(Sync) END
            ELSE Pending := TRUE;
                 SemP(Sync); 
            END;
          ELSE SemP(Sync) END;
        RETURN Result;
  END Wait;

  PROCEDURE  Signal( r : CARDINAL);

  VAR   x : Status;

  BEGIN ClearControl({2});
        IF r = CARDINAL(TimedOut) THEN Pending := FALSE END;
        IF Pending THEN 
          x := CancelAlarm(e);
          Pending := FALSE;
          END;
        IF 1 IN IntControl THEN RETURN END; (* The real one is on the way *)
        Result := Status(r);
        SemV(Sync);
  END Signal;

BEGIN   SemInit(Sync,0);
        Result := Ok;
        Pending := FALSE;
END Level4;

PROCEDURE  SoftInt;

BEGIN   EXCL(IntControl,1);
        Signal(CARDINAL(Ok));
END SoftInt;

PROCEDURE  SetTrack( unit, t : CARDINAL) : Status;

VAR     r : Status;

BEGIN   WITH Drive[unit] DO
          IF t <> TrackNum THEN
            r := Flush(unit);
            IF r <> Ok THEN RETURN r END;
            r := Transfer(TRUE,unit,t);
            IF r <> Ok THEN RETURN r END;
            TrackNum := t;
            END;
          END;
        RETURN Ok;
END SetTrack;

PROCEDURE  ReadBlock( unit, sector : CARDINAL; VAR b : ARRAY OF CHAR) : Status;

VAR     r : Status;
        t, s : CARDINAL;

BEGIN   IF unit > MaxUnit THEN RETURN NoSuch END;
        IF sector >= 1280 THEN RETURN NoSector END;
        s := sector MOD 8;
        t := sector DIV 8;
        WITH Drive[unit] DO
          r := SetTrack(unit,t);
          IF r <> Ok THEN RETURN r END;
          Copy(Track[s],b);
          END;
        RETURN Ok;
END ReadBlock;

PROCEDURE  WriteBlock(unit, sector: CARDINAL; VAR b : ARRAY OF CHAR) : Status;

VAR     r : Status;
        t, s : CARDINAL;

BEGIN   IF unit > MaxUnit THEN RETURN NoSuch END;
        IF sector >= 1280 THEN RETURN NoSector END;
        s := sector MOD 8;
        t := sector DIV 8;
        WITH Drive[unit] DO
          r := SetTrack(unit,t);
          IF r <> Ok THEN RETURN r END;
          IF WriteProtect THEN RETURN Protect END;
          Copy(b,Track[s]);
          Dirty := TRUE;
          END;
        RETURN Ok;
END WriteBlock;

PROCEDURE  Copy( VAR s, t : ARRAY OF CHAR);

VAR     i, k : CARDINAL;

BEGIN   k := HIGH(s);
        IF k > HIGH(t) THEN k := HIGH(t) END;
        FOR i := 0 TO k DO t[i] := s[i] END;
        FOR i := k+1 TO HIGH(t) DO t[i] := 0C END;
END Copy;

PROCEDURE  Flush( unit : CARDINAL ) : Status;  (* Write buffered track *)

VAR     r : Status;

BEGIN   IF unit > MaxUnit THEN RETURN NoSuch END;
        WITH Drive[unit] DO
          IF Dirty THEN 
            r := Transfer(FALSE,unit,TrackNum);
            IF r <> Ok THEN RETURN r END;
            Dirty := FALSE;
            TrackNum := 65535;
            END;
          END;
        RETURN Ok;
END Flush;

VAR     i : Units;

BEGIN   Establish(52H,Handler(FloppyInt),0);
        Establish(53H,Handler(SoftInt),0);
        INCL(IntControl,2); (* Enable the IPL 1 interrupt *)
        SavedProc := Vectors[52H];
        Initialised := FALSE;
        MotorOn := FALSE;
        FOR i := 0 TO MaxUnit DO
          WITH Drive[i] DO
            Dirty := FALSE;
            TrackNum := 65535;
            Recalibrated := FALSE;
            WriteProtect := FALSE;
            END;
          END;
END L2Floppy.
