IMPLEMENTATION MODULE L2Memory[4];

FROM SYSTEM IMPORT ADDRESS, REGISTER, SETREG;
FROM L2Storage IMPORT ALLOCATE, DEALLOCATE;
FROM L2Errors IMPORT Status;
FROM L2Exceptions IMPORT Assert;

CONST   MaxPage = 1024;  (* bit map size - 4 megs - increase if needed *)
        HeapPages = 28 + 64;  (* heap, RamDisk *)

TYPE    PageMap = RECORD CodeSize, StackSize : CARDINAL;
                         Frames : ARRAY [1..UserFrames] OF CARDINAL;
                  END;
        AddressSpace = POINTER TO PageMap;

VAR     PageTable [0FF4000H] : ARRAY [0..1] OF ARRAY [0..1023] OF CARDINAL;
        MMUControl [0FFCA00H] : BITSET;
        MemTop [0FF8300H] : ADDRESS;   (* left by ROM *)
 
        BitMap : ARRAY [0..(MaxPage DIV 16)] OF BITSET;
          (* Page n is used iff (n MOD 16) IN BitMap[n DIV 16] *)
        FreeCount  : CARDINAL;  (* Number of free pages *)
        CurrentMap : AddressSpace;

PROCEDURE  NewSpace( VAR a : AddressSpace; code, stack : CARDINAL) : Status;

VAR     i : CARDINAL;
        rs : Status;

BEGIN   IF code + stack > FreeCount THEN RETURN NoMemory END;
        IF code + stack > UserFrames THEN RETURN NoSpace END;
        NEW(a);
        IF a = NIL THEN RETURN NoHeap END;
        WITH a^ DO
          FOR i := 1 TO UserFrames DO a^.Frames[i] := 1000H END;  (* Invalid *)
          CodeSize := 0;
          StackSize := 0;
          END;
        rs := ExpandCode(a,code);
        rs := ExpandStack(a,stack);
        RETURN Ok;
END NewSpace;

PROCEDURE  ExpandCode( a : AddressSpace; code : CARDINAL) : Status;

VAR     p : CARDINAL;

BEGIN   IF code > FreeCount THEN RETURN NoMemory END;
        IF a = NIL THEN RETURN NoSuch END;
        WITH a^ DO 
          IF CodeSize + StackSize + code > UserFrames 
            THEN RETURN NoSpace END;
          p := 0;
          WHILE code > 0 DO
            WHILE (p MOD 16) IN BitMap[p DIV 16] DO INC(p) END;
            INCL(BitMap[p DIV 16],p MOD 16); DEC(FreeCount);
            Frames[CodeSize + 1] := p;
            IF a = CurrentMap THEN 
              PageTable[0][CodeSize + 1] := p;
              PageTable[1][CodeSize + 1] := p;
              END;
            INC(CodeSize);
            DEC(code);
          END;
        END;
        RETURN Ok;
END ExpandCode;

PROCEDURE  ExpandStack( a : AddressSpace; stack : CARDINAL) : Status;

VAR     p : CARDINAL;

BEGIN   IF stack > FreeCount THEN RETURN NoMemory END;
        IF a = NIL THEN RETURN NoSuch END;
        WITH a^ DO 
          IF CodeSize + StackSize + stack > UserFrames 
            THEN RETURN NoSpace END;
          p := 0;
          WHILE stack > 0 DO
            WHILE (p MOD 16) IN BitMap[p DIV 16] DO INC(p) END;
            INCL(BitMap[p DIV 16],p MOD 16); DEC(FreeCount);
            Frames[UserFrames - StackSize] := p;
            IF a = CurrentMap THEN 
              PageTable[0][UserFrames - StackSize] := p;
              PageTable[1][UserFrames - StackSize] := p;
              END;
            INC(StackSize);
            DEC(stack);
          END;
        END;
        RETURN Ok;
END ExpandStack;

PROCEDURE  MapSpace( a : AddressSpace) : Status;

VAR     i : CARDINAL;

BEGIN   IF a = NIL THEN RETURN NoSuch END;
        IF a <> CurrentMap THEN
          FOR i := 1 TO UserFrames DO
            PageTable[0][i] := a^.Frames[i];
            PageTable[1][i] := a^.Frames[i];
            END;
          CurrentMap := a;
          END;
        RETURN Ok;
END MapSpace;
 
PROCEDURE  DisposeSpace( VAR a : AddressSpace) : Status;

VAR     i, p : CARDINAL;

BEGIN   IF a = NIL THEN RETURN NoSuch END;
        FOR i := 1 TO UserFrames DO 
          p := a^.Frames[i];
          IF p <> 1000H THEN 
            EXCL(BitMap[p DIV 16],p MOD 16);
            INC(FreeCount);
            END;
          END;
        DISPOSE(a);
        RETURN Ok;
END DisposeSpace;

(* This module should be initialised very early.  It assumes that
   the linker bound the image to pages beginning at UserFrames + 1
   and that the boot loader left the first free address in A5. The
   stack is moved, and the system heap is initialised. *)

VAR     ff : ADDRESS;
        lp, i : CARDINAL;

(*$S-*)  (* We play with the stack here *)
(*$P-*)  (* Avoid the LINK instruction  *)

BEGIN   ff := REGISTER(13);  (* A5 *)
        lp := CARDINAL((ff + 4*ADDRESS(StackMax)) DIV 1000H); (* Last page *)
        ff := ADDRESS(lp + 1) * ADDRESS(1000H) - ADDRESS(2);  (* SP *)
        SETREG(15,ff);  (* Move the SuperStack *)
        FOR i := 1 TO HeapPages DO
          INC(lp);
          ff := ADDRESS(lp)*ADDRESS(1000H);
          DEALLOCATE(ff,1000H);
          END;
        FOR i := 0 TO 1023 DO
          PageTable[0][i] := 1000H;
          PageTable[1][i] := 1000H;  (* Invalid *)
          END;
        FOR i := 0 TO MaxPage DIV 16 DO BitMap[i] := { } END;
        FOR i := UserFrames+1 TO lp DO
          PageTable[1][i] := i;  (* 1-1 Mapping in Supervisor *)
          INCL(BitMap[i DIV 16],i MOD 16);
          END;
        PageTable[1][0] := 0;  (* Map the Vectors in SuperMode *)
        INCL(BitMap[0],0); (* Mark vector page used *)
        Assert(CARDINAL(MemTop DIV 1000H) <= MaxPage,10);
        FreeCount := CARDINAL(MemTop DIV 1000H) - 1 - ( lp - UserFrames);
        CurrentMap := NIL;
        MMUControl := {   };
        MMUControl := { 7 };
        MMUControl := { 2, 5 };  (* Initiate Mapping *)
END L2Memory.
