IMPLEMENTATION MODULE L2Winchester;  

(* The Stride's winchester controller is home-brew.  The following code
   represents knowledge gleaned from Stride's BIOS source and the controller
   schematics. We assume that the boot ROM has correctly recalibrated the 
   drive and that a valid format exists.  Note that the control port 
   WinCB2 is a write-only port shared with the QIC-02 controller and with
   the parity controller.  At present only L2Cartridge makes real use of
   this port. *)    

(* The driver loads the number of heads, the list of badblocks and the
   partition information from track0 of the disk.  The partitions are
   offset by one - #W0 refers to partition number 1, #W1 to number 2, etc.
   LUMOS-2 does not provide access to partition number 0 - the whole disk.
   The badblock list is presumed to be in order. The blockmove operations
   are used to speed data transfers.
*)

FROM L2Errors IMPORT Status;
FROM L2Kernel IMPORT Semaphore, SemV, SemP, SemInit;
FROM L2Exceptions IMPORT Establish, AssertOk;
FROM L2Clock IMPORT Time, GetTime;
FROM L2BlockMove IMPORT FastIn, FastOut;
FROM SYSTEM IMPORT ADR;

CONST   SeekIn         = 0;  (* Bits on WinControl *)
        InhibitEndInt  = 1;
        InhibitSeekInt = 2;
        Verify         = 3;
        Open           = 4;
        NotStep        = 5;
        WriteDisk      = 6;
        CPU            = 7;
        NotSeekDone    = 0;  (* Bits on WinStatus  *)
        NotTrack0      = 1;
        NotWriteFault  = 2;
        NotReady       = 3;
        CRCErr         = 4;
        AEND           = 5;
        NotIndex       = 6;
        Drive0         = {0,1,2,3,5,6,7};    (* on WinSelect *)
        NoSelect       = {0,1,2,3,4,5,6,7};
        MaxParts       = 14;                                 
        MaxJunk        = 5;
        MaxBad         = 218;  (* Track0 must total 512 bytes *)

VAR     WinBufCnt  [0FFC400H] : CARDINAL;
        WinTrCnt   [0FFC440H] : CARDINAL;
        WinData    [0FFC481H] : CHAR;
        WinStatus  [0FFC4C0H] : BITSET;
        WinSelect  [0FFC500H] : BITSET;
        WinControl [0FFC540H] : BITSET;
        WinCB2     [0FFC580H] : CARDINAL;   (* Unified port - unused *)

        Track0 : RECORD Heads : CARDINAL;
                        junk : ARRAY [0..MaxJunk] OF CARDINAL;
                        Part : ARRAY [0..MaxParts] OF
                                 RECORD start, len : CARDINAL; END;
                        Bad  : ARRAY [0..MaxBad] OF CARDINAL;
                 END;  (* 512 bytes *)

        Control : BITSET;  (* Cache of WinControl *)
        Sync : Semaphore;
        CurrentPTN, CurrentHead : CARDINAL;
        CRCBad, Dirty, Initialised : BOOLEAN;

PROCEDURE  ReadBlock( unit, sector: CARDINAL; 
                          VAR b : ARRAY OF CHAR) : Status;

VAR     ptn, sn, i : CARDINAL;
        rs : Status;
 
BEGIN   IF NOT Initialised THEN InitWinch END;
        rs := Map(unit,sector,ptn,sn);
        IF rs <> Ok THEN RETURN rs END;
        rs := SetTrack(ptn);
        IF rs <> Ok THEN RETURN rs; END;
        WinBufCnt := 512*sn + 018H;   
        IF HIGH(b) = 511 
          THEN FastIn(ADR(WinData),ADR(b))
          ELSE i := 0;
               WHILE (i < 512) AND (i <= HIGH(b)) DO 
               b[i] := WinData;
               INC(i);
               END;
          END;
        RETURN Ok;
END ReadBlock;

PROCEDURE  WriteBlock( unit, sector: CARDINAL; 
                          VAR b : ARRAY OF CHAR) : Status;

VAR     ptn, sn, i : CARDINAL;
        rs : Status;
 
BEGIN   IF NOT Initialised THEN InitWinch END;
        rs := Map(unit,sector,ptn,sn);
        IF rs <> Ok THEN RETURN rs END;
        rs := SetTrack(ptn);
        IF rs <> Ok THEN RETURN rs; END;
        WinBufCnt := 512*sn + 018H;
        IF HIGH(b) = 511
          THEN FastOut(ADR(b),ADR(WinData));
          ELSE i := 0;
               WHILE (i < 512) AND (i <= HIGH(b)) DO 
               WinData := b[i];
               INC(i);
               END;
          END;
        Dirty := TRUE;
        RETURN Ok;
END WriteBlock;                

PROCEDURE  SectorCount( unit : CARDINAL; VAR size : CARDINAL) : Status;

BEGIN   IF NOT Initialised THEN InitWinch END;
        IF unit > MaxParts THEN RETURN NoUnit END;
        size := Track0.Part[unit].len * 16;  (* 16 sectors per track *)
        RETURN Ok;
END SectorCount;

PROCEDURE  Flush( unit : CARDINAL) : Status;  (* Write buffered track *)

(* unit is ignored at present *)

VAR     retries, optn : CARDINAL;
        rs : Status;

BEGIN   IF NOT Initialised THEN InitWinch END;
        retries := 0;
        rs := Ok;  
        IF Dirty THEN
          WinSelect := Drive0 - BITSET(CurrentHead);
          LOOP optn := CurrentPTN;
               StoreTrack;
               LoadTrack(TRUE);
               IF NOT CRCBad THEN
                 rs := Ok;
                 IF optn <> CurrentPTN THEN rs := DevError END;
                 EXIT;
                 END;
               IF retries = 2 THEN EXIT END;
               INC(retries);
               END;
          Dirty := FALSE;
          END;
          WinSelect := NoSelect;
        RETURN rs;
END Flush;

PROCEDURE  InitWinch;

(* Loads the tables from the disk and modifies them for use by Map, etc. *)

(* The partition table on disk contains start,end pairs whereas we need
   start,len pairs and in terms of mapped tracks.  The computations here
   map out the bad tracks from the partitions. *)

VAR     p, n, b, l, s : CARDINAL;
        rs : Status;

BEGIN   Track0.Heads := 1;  (* dummy value to load 0 *)
        AssertOk(SetTrack(0),999);
        WinBufCnt := 018H;  (* sector 0 *)
        FastIn(ADR(WinData),ADR(Track0));
        WITH Track0 DO
          n := 0;
          WHILE Bad[n] <> 0 DO INC(n) END;  
          Bad[n] := 65535;
          FOR p := 0 TO MaxParts DO
            WITH Part[p] DO
              IF len > 0 THEN
                s := start; l := len - start + 1;
                FOR b := 0 TO n DO
                  IF Bad[b] < start THEN DEC(s)
                  ELSIF Bad[b] < len THEN DEC(l) END;
                  END;
                start := s; len := l;
                END;
              END;
            END;
          FOR b := 1 TO n-1 DO DEC(Bad[b],b) END;
          END;
        Initialised := TRUE;
END InitWinch;

PROCEDURE  Map( unit, sector : CARDINAL; VAR ptn, sn : CARDINAL) : Status;

VAR     k : CARDINAL;

BEGIN   IF unit > MaxParts THEN RETURN NoUnit END;
        WITH Track0 DO
          IF sector >= Part[unit].len * 16 THEN RETURN NoSector END;
          sn := sector MOD 16;
          ptn := (sector DIV 16) + Part[unit].start;
          k := 0;
          WHILE ptn >= Bad[k] DO INC(k) END;
          INC(ptn,k);
          END;
        RETURN Ok;
END Map;
        
PROCEDURE  LoadTrack( vfy : BOOLEAN);

(* Loads the track at the current cylinder and head into
   the controller's buffer.  The track number is obtained from the 
   track header and saved in CurrentPTN.  CRCBad is updated. *)
                                                             
(* vfy instructs the controller to do a fast compare and set the error codes *)

BEGIN   WinBufCnt := -4;
        WinTrCnt := -018H;
        IF NOT vfy THEN EXCL(Control,Verify) END;
        Control := Control - {InhibitEndInt,WriteDisk,CPU} 
                           + {Open,NotStep};
        WinControl := Control;
        SemP(Sync);   (* on return CPU control is enabled etc. *)
        CRCBad := CRCErr IN WinStatus;
        WinBufCnt := 3; (* Location of PTN in header *)
        CurrentPTN := 256*ORD(WinData) + ORD(WinData);  (* reads twice *)
END LoadTrack;

PROCEDURE  StoreTrack;

(* Writes the current buffered track back to the disk.  Writing the time into 
   the header ensures that the data is different. *)

VAR     i : CARDINAL;
        ts : RECORD CASE BOOLEAN OF
                 TRUE  : t : Time |
                 FALSE : s : ARRAY [0..5] OF CHAR;
                 END;
               END;

BEGIN   GetTime(ts.t);
        WinBufCnt := -64;
        FOR i := 1 TO 60 DO WinData := 0C END;
        WinData := 100C;  (* Sync bit *)
        FOR i := 1 TO 6 DO WinData := 0C END;
        WinBufCnt := 6;
        FOR i := 0 TO 5 DO WinData := ts.s[i] END;
        WinBufCnt := 2018H; (* Trailer *)
        FOR i := 1 TO 4 DO WinData := 0C END;
        WinBufCnt := -64;
        WinTrCnt := -1AH;
        Control := Control - {Verify,InhibitEndInt,CPU}
                           + {Open,NotStep,WriteDisk};
        WinControl := Control;
        SemP(Sync);
END StoreTrack;

PROCEDURE  SetTrack( ptn : CARDINAL) : Status;

VAR     cyl, cc, hd : CARDINAL;
        seekretries, loadretries : CARDINAL;
        rs : Status;

BEGIN   IF ptn = CurrentPTN THEN RETURN Ok END;
        IF Dirty THEN 
          rs := Flush(0);  (* dummy parameter is ignored *)
          IF rs <> Ok THEN RETURN rs END;
          END;
        IF CurrentPTN = 65535 THEN
          WinSelect := Drive0;
          LoadTrack(FALSE); 
          END;    
        WITH Track0 DO
          cyl := ptn DIV Heads;
          hd := ptn MOD Heads;
          END;
        CurrentHead := hd;  (* save for future flush *)
        seekretries := 0;
        LOOP WinSelect := Drive0 - BITSET(hd);
             cc := CurrentPTN DIV Track0.Heads;
             Step(cyl-cc);
             loadretries := 0;
             LOOP LoadTrack(FALSE);
                  IF NOT CRCBad THEN EXIT END;
                  INC(loadretries);
                  IF loadretries = 4 THEN RETURN CRCError END;
                  END;
             WinSelect := NoSelect;
             IF CurrentPTN = ptn THEN EXIT END;
             INC(seekretries);
             IF seekretries = 4 THEN RETURN SeekFail END;
             END;
        RETURN Ok;
END SetTrack;

PROCEDURE  Step( n : INTEGER);  (* Move the head by n steps *)

BEGIN   IF n = 0 THEN RETURN END;
        IF n > 0 
          THEN n := -n;
               EXCL(Control,SeekIn);
          ELSE INCL(Control,SeekIn);
          END;
        WinBufCnt := 0;
        WinTrCnt := n;
        EXCL(Control,NotStep);
        WinControl := Control;
        WHILE NOT (AEND IN WinStatus) DO END;
        EXCL(Control,InhibitSeekInt);
        WinControl := Control;
        SemP(Sync);
END Step;

MODULE Level4[4];

  IMPORT  SemV, Control, WinControl, Establish, Open, CPU, NotStep, 
          WinTrCnt, WinBufCnt, Sync, InhibitSeekInt, InhibitEndInt,
          Verify;

  PROCEDURE  WinchInt( inh : CARDINAL);

  BEGIN EXCL(Control,Open);
        INCL(Control,inh);
        Control := Control + {CPU,NotStep,Verify};
        WinControl := Control;
        WinBufCnt := 0;
        WinTrCnt := 0;
        SemV(Sync);
  END WinchInt;

BEGIN   Establish(62H,WinchInt,InhibitSeekInt);
        Establish(63H,WinchInt,InhibitEndInt);
END Level4;

BEGIN   WinCB2 := 0CH;  (* disables PreComp - our only interest *)
        WinSelect := NoSelect;
        Control := {InhibitSeekInt,InhibitEndInt,Verify,NotStep};
        SemInit(Sync,0);
        Dirty := FALSE;
        CurrentPTN := 65535;
        Initialised := FALSE;
        CurrentHead := 0;
END L2Winchester.
