IMPLEMENTATION MODULE N2ICMP;

FROM L2Clock      IMPORT Ticks, TickCount;
FROM L2Exceptions IMPORT Assert;         
FROM L2Kernel     IMPORT CurrentProcess, Semaphore, SemP, SemV, SemInit, 
                         Process, SetNoSwitch;
FROM N2Base       IMPORT IP1, IPMinLen, IPhdrlen, IPtlen, IPsaddr, IPdaddr,
                         IPproto, HostIP, IPAddr, NetHost2, HostNet2, 
                         HostNet4, ChkSum;
FROM N2Errors     IMPORT Status;
FROM N2IP         IMPORT IPRegister, IPSend;
FROM N2Packets    IMPORT PacketLen, PacketPtr, Packet, PacketQ, NewPacket, 
                         PutPacket, GetPacket, DisposePacket, InitPacketQ, 
                         DeleteQ;

CONST                              

 (* ICMP field offsets for datagrams generated at this host *)

  ICMP1       = IP1 + IPMinLen;
  ICMPtype    = ICMP1;
  ICMPcode    = ICMP1 + 1;
  ICMPchksum  = ICMP1 + 2;
  ICMPident   = ICMP1 + 4;
  ICMPseq     = ICMP1 + 6;
  Data1       = ICMP1 + 8;

 (*  ICMP  Types  *)

  EchoReply   = 0C;
  DestUnreach = 3C;
  Quench      = 4C;
  Redirect    = 5C;
  EchoMess    = 10C;
  TimeExceed  = 13C;
  ParamProb   = 14C;
  TStampMess  = 16C;
  InfoReq     = 20C;
                  
  ICMPProtocol = 1C;
  MaxTmo       = 700;  (* 7 seconds *)
  MaxData      = PacketLen - Data1;

VAR                                                       
  icmp1,                          (* offset of first octet of ICMP hdr *)
  icmpLen : CARDINAL;             (* length of ICMP hdr & data *)
  stats   : RECORD
              echo,
              other,
              bad   : CARDINAL;
            END;
  pingCB :  RECORD
              guard : Semaphore;
              seq   : CARDINAL;
              time  : Ticks;
              pktQ  : PacketQ;
              pkt   : PacketPtr;
            END;

PROCEDURE LockICMP;
  BEGIN
    SemP( pingCB.guard );
  END LockICMP;

PROCEDURE UnLockICMP;
  BEGIN
    SemV( pingCB.guard );
  END UnLockICMP;

PROCEDURE Ping( forHost : IPAddr; tmo : Ticks; size : CARDINAL;
                                              VAR rtt : Ticks ) : Status;
  VAR                             
    res          : Status;
    i, chk, rseq : CARDINAL;
  BEGIN
    WITH pingCB DO
      SemP( guard );
      NewPacket( pkt );
      DeleteQ( pktQ );
      InitPacketQ( pktQ, 1 );

     (* ...set IP fields -> protocol, source address, foreign address *)

      pkt^.data[IPproto] := ICMPProtocol;
      pkt^.data[IPsaddr] := HostIP[0];
      pkt^.data[IPsaddr + 1] := HostIP[1];
      pkt^.data[IPsaddr + 2] := HostIP[2];
      pkt^.data[IPsaddr + 3] := HostIP[3];
      HostNet4( forHost, IPdaddr, pkt^.data );

     (* ...add data if required & calculate packet length *)

      IF size > MaxData THEN size := MaxData; END;
      pkt^.plen := Data1 + size;
      FOR i := Data1 TO pkt^.plen+1 DO pkt^.data[i] := CHR(i-Data1); END;

     (* ...fill in ICMP fields -> type, code, identifier, sequence, chksum *)

      pkt^.data[ICMPtype] := EchoMess;
      pkt^.data[ICMPcode] := 0C;
      HostNet2( CARDINAL(CurrentProcess), ICMPident, pkt^.data );
      seq := CARDINAL( TickCount MOD 10000H );
      HostNet2( seq, ICMPseq, pkt^.data );
      pkt^.data[ICMPchksum] := 0C;
      pkt^.data[ICMPchksum + 1] := 0C;
      ChkSum( chk, ICMP1, size + 8, pkt^.data, FALSE );
      HostNet2( chk, ICMPchksum, pkt^.data );

     (* ...send it, record send time *)

      res := IPSend( pkt );
      time := TickCount;
      IF res <> Ok THEN
        SemV( guard );
        RETURN res;
      END; (* IF *)

     (* ...wait for reply *)

      IF (tmo = 0) OR (tmo > MaxTmo) THEN tmo := MaxTmo; END;
      IF GetPacket( pkt, pingCB.pktQ, tmo ) <> Ok THEN 
        SemV( guard ); 
        RETURN TimedOut; 
      END; (* IF *)
      rtt := TickCount - time;

     (* ...verify sequence number *)

      NetHost2( rseq, ICMPseq, pkt^.data );
      DisposePacket( pkt );
      IF rseq <> seq THEN SemV( guard ); RETURN BadSeq; END;
      SemV( guard );
    END; (* WITH *)
    RETURN Ok;
  END Ping;

PROCEDURE RcvEchoReply( p : PacketPtr );
  VAR
    ident : Process;
  BEGIN
    NetHost2( ident, ICMP1 + 4, p^.data );
    SetNoSwitch( TRUE );
    IF pingCB.pktQ.pid = ident THEN
      PutPacket( p, pingCB.pktQ );
    ELSE
      DisposePacket( p );
    END; (* IF *)
    SetNoSwitch( FALSE );
  END RcvEchoReply;

PROCEDURE SendEchoReply( p : PacketPtr );
 (* ICMP Type 0 .. Echo Reply *)
  VAR
    i, chk : CARDINAL;
    res    : Status;
  BEGIN
    p^.data[ICMP1] := EchoReply;                     (* set ICMP type *)
    FOR i := 0 TO 3 DO
      p^.data[IPdaddr + i] := p^.data[IPsaddr + i];  (* swap source and *)
      p^.data[IPsaddr + i] := HostIP[i];             (* dest address    *)
    END; (* FOR *)
    p^.data[ICMP1 + 2] := 0C;                        (* compute chksum *)
    p^.data[ICMP1 + 3] := 0C;
    ChkSum( chk, icmp1, icmpLen, p^.data, FALSE );
    HostNet2( chk, icmp1 + 2, p^.data );
    p^.plen := icmp1 + icmpLen;
    res := IPSend( p );                              (* return to sender *)
  END SendEchoReply;

PROCEDURE ICMPIn( p : PacketPtr );
 (* Incoming packets from IP *)
  VAR
    chk1, chk2 : CARDINAL;
  BEGIN
        
   (* calculate ICMP start offset & ICMP total length *)

    icmp1 := IP1 + 4 * ( CARDINAL( p^.data[IPhdrlen] ) MOD 10H );
    NetHost2( icmpLen, IPtlen, p^.data );
    icmpLen := IP1 + icmpLen - icmp1;

   (* recalculate checksum *)

    NetHost2( chk1, icmp1 + 2, p^.data );    (* save transmitted checksum *)
    p^.data[icmp1 + 2] := 0C;
    p^.data[icmp1 + 3] := 0C;
    ChkSum( chk2, icmp1, icmpLen, p^.data, FALSE );

   (* process ICMP message if checksum is ok *)

    IF chk1 = chk2 THEN
      CASE p^.data[icmp1] OF
        EchoMess     : INC( stats.echo ); SendEchoReply( p );                |
        EchoReply    : INC( stats.echo ); RcvEchoReply( p );                 |
        DestUnreach  : IF ORD( p^.data[icmp1 + 1] ) > 3 THEN
                        INC( stats.other );
                       ELSE
                        INC( stats.other );
                       END; (* IF *)
                       DisposePacket( p );                                    |
        Quench       : INC( stats.other ); DisposePacket( p );              |
        Redirect     : IF ORD( p^.data[icmp1 + 1] ) > 5 THEN
                        INC( stats.other );
                       ELSE
                        INC( stats.other);
                       END; (* IF *)
                       DisposePacket( p );                                    |
        TimeExceed   : INC( stats.other ); DisposePacket( p );        |
        ParamProb    : INC( stats.other ); DisposePacket( p );        |
        TStampMess   : INC( stats.other ); DisposePacket( p );           |
        InfoReq      : INC( stats.other ); DisposePacket( p );
      ELSE
        INC( stats.bad ); DisposePacket( p );
      END; (* CASE *)
    ELSE
      INC( stats.bad ); DisposePacket( p );
    END; (* IF *)
  END ICMPIn;

PROCEDURE ICMPStats( VAR echo, other, bad : CARDINAL );
  BEGIN
    echo := stats.echo;                               
    other := stats.other;
    bad := stats.bad;
  END ICMPStats;

VAR
  i : CARDINAL;
BEGIN
  Assert( IPRegister( ICMPIn, ICMPProtocol ) = Ok, 501 );
  stats.echo := 0;
  stats.other := 0;
  stats. bad := 0;
  InitPacketQ( pingCB.pktQ, 1 );
  SemInit( pingCB.guard, 1 );
END N2ICMP.
