IMPLEMENTATION MODULE N2Packets [4];
FROM L2Kernel IMPORT Process,Suspend,Resume,CurrentProcess;
FROM L2Clock IMPORT Ticks,Event,RequestAlarm,CancelAlarm;
FROM L2Storage IMPORT ALLOCATE,DEALLOCATE;
FROM L2Errors IMPORT Status;    
FROM L2Exceptions IMPORT Handler;
IMPORT N2Errors;

TYPE
    Stack = RECORD
              top:PacketPtr;
            END;
VAR
    i : CARDINAL;
    s : Stack;
    r : PacketPtr;
    L2rs : Status;
    proc : Process;                       
    StackPackets,TotalPackets : CARDINAL;

(*This procedure will initialize a new packet queue. Qsize will be the 
  maximum number of packets that can be on the queue.*)

PROCEDURE InitPacketQ(VAR q:PacketQ;Qsize:CARDINAL);
VAR
   i:CARDINAL;
   p:PacketPtr;
BEGIN
   q.head:=NIL;
   q.tail:=NIL;
   q.maxsize:=Qsize;
   q.cnt:=0;
   q.pid:=CurrentProcess;   
   q.blocked:=FALSE;
END InitPacketQ;
             
(*This procedure will place a packet on the specified packet queue if
  the queue has not reached its maximum limit*)

PROCEDURE PutPacket(VAR p:PacketPtr;VAR q:PacketQ);
BEGIN
   IF q.maxsize=q.cnt THEN
      Push(p,s);                    (*Queue has maximum limit, dispose of*)
   ELSE                             (*packet*)
      IF q.head=NIL THEN
        q.head:=p;                  (*Else add the packet onto the queue*)
        p^.prev:=NIL;
        q.cnt:=1;
      ELSE
        p^.prev:=q.tail;
        q.tail^.next:=p;
        INC(q.cnt);
      END(*IF*);
      p^.next:=NIL; 
      q.tail:=p;
   END(*IF*);
   IF q.blocked THEN                (*If a process was waiting for a packet*)
      L2rs:=Resume(q.pid);          (*then awaken the process*)
      q.blocked:=FALSE;
   END(*IF*);     
   p:=NIL;
END PutPacket;
                                                              

(*This procedure will retreive a packet from a packet queue, if a packet
  exists on the queue. Otherwise the process will wait for tmo, and then
  give up*)

PROCEDURE GetPacket(VAR p:PacketPtr;VAR q:PacketQ;tmo:Ticks):N2Errors.Status;
VAR
   proc:Process;
   cnt:CARDINAL;
BEGIN
   cnt:=0;
   LOOP
      cnt:=cnt+1;
      IF q.head=NIL THEN
        IF cnt=2 THEN               (*If cnt=2 then we have already timed out*)
           RETURN N2Errors.TimedOut;
           EXIT;
        END(*IF*);
        q.blocked:=TRUE;            (*Wait for at most tmo seconds*)
        L2rs:=Sleep(tmo,q);
        IF L2rs<>Ok THEN
           q.blocked:=FALSE;
        END(*IF*);
      ELSE                          
        p:=q.head;                  (*There is a packet on the queue, thus*)
        q.head:=q.head^.next;       (*retreive packet and exit*)
        IF q.head=NIL THEN
          q.tail:=NIL;
        ELSE
          q.head^.prev:=NIL;
        END(*IF*);
        q.cnt:=q.cnt-1;
        EXIT;
      END(*IF*);
   END(*LOOP*);           
   RETURN N2Errors.Ok;
END GetPacket;
                                                               
(*This procedure is used to store unwanted packets until they are needed again*)

PROCEDURE Push(VAR p:PacketPtr;VAR s:Stack);
BEGIN
   p^.next:=s.top;
   s.top:=p;
   StackPackets:=StackPackets+1;
   p:=NIL;
END Push;
              
(*This procedure will retreive a packet from the stack, if none exists then
  a new packet will be allocated.*)

PROCEDURE Pop(VAR p:PacketPtr;VAR s:Stack);
BEGIN        
   
   IF s.top=NIL THEN
      NEW(p);                            
      TotalPackets:=TotalPackets+1;
   ELSE
      p:=s.top;
      s.top:=s.top^.next;
      StackPackets:=StackPackets-1;
   END(*IF*);
END Pop;

(*This procedure calls procedure pop, to see if there are any available
  packets, if not then a new one is made*)

PROCEDURE NewPacket(VAR p:PacketPtr);
BEGIN
   Pop(p,s);   
   IF p<>NIL THEN               (*Check to make sure we never ran out of heap*)
     p^.plen:=0;
     p^.next:=NIL;
     p^.prev:=NIL;
   END(*IF*);
END NewPacket;

(*This procedure will place unwanted packets on the stack*)

PROCEDURE DisposePacket(VAR p:PacketPtr);                               
BEGIN                      
   p^.plen:=0;
   Push(p,s);       
   p:=NIL;
END DisposePacket;

(*This procedure will return statistics about the number of packets in the
  system and how many are on the stack*)

PROCEDURE Qquery(VAR TPackets,SPackets:CARDINAL);
BEGIN
   TPackets:=TotalPackets;
   SPackets:=StackPackets;
END Qquery;

(*This procedure will delete a queue and will place any packets on the
  queue in the stack*)

PROCEDURE DeleteQ(VAR q:PacketQ);
VAR
   p:PacketPtr;
BEGIN
   WHILE q.head<>NIL DO           (*While packets in the queue do*)
     p:=q.head;
     q.head:=q.head^.next;
     Push(p,s);                   (*Return packets to the stack for future use*)
   END(*WHILE*);
   L2rs:=CancelAlarm(q.alarm);      (*Cancel any alarms if there is a process*)
END DeleteQ;                      (*blocked*)
                                            
(*This procedure will allow a process to sleep for so much time. If this
  time expires or if a packet is put on the queue then the process is
  awakened*)
 
PROCEDURE Sleep(intervals:Ticks; q:PacketQ):Status;
BEGIN
   IF intervals = 0 THEN Suspend; RETURN Ok; END;
   L2rs:=RequestAlarm(Handler(Resume),CurrentProcess,intervals,
                                                     FALSE,q.alarm);
   IF L2rs<>Ok THEN
      RETURN L2rs;
   END(*IF*);
   Suspend;                          (*until some signal, probably timed out*)
   L2rs:=CancelAlarm(q.alarm);       (*Alarm is cancelled*)
   RETURN Ok;
END Sleep;

BEGIN                                                                      
   StackPackets:=0;
   s.top:=NIL;
   FOR i:=1 TO 5 DO
     NEW(r);
     Push(r,s);
   END(*FOR*);
   TotalPackets:=5;
END N2Packets.
