IMPLEMENTATION MODULE N2TCPLib;

FROM N2TCPTimer IMPORT ClearTimerEvents, DeleteTimerEvent, SetTimerEvent,
                       Event;
FROM N2TCPMain  IMPORT TCBPtr,BuffSz,TCPMINRXT,TCPTWOMSL,activeSocs, OutState,
                       TCBState, TCB, Sleeper, MaxSegSz, FragPtr, SocHandle,
                       socList;
FROM N2Errors   IMPORT Status;
FROM SYSTEM     IMPORT ADDRESS;
FROM L2Clock    IMPORT Ticks;               
FROM L2Kernel   IMPORT Resume,SemV, SemP, SemQ;
FROM L2Storage  IMPORT DEALLOCATE;
IMPORT L2Errors;

CONST
  HalfSeq = 80000000H;

(* Wake up processes sleeping. Sleeper determines which processes are
   woken *)

PROCEDURE WakeUp( tcb : TCBPtr; sleeper : Sleeper );
VAR
  rs:L2Errors.Status;
  BEGIN
    IF tcb^.pid = 0 THEN RETURN; END;
    CASE sleeper OF
      READER : IF tcb^.rdBlocked THEN
                SemV(tcb^.socPtr^.ioGuard);
                tcb^.rdBlocked:=FALSE;
               ELSIF tcb^.attached THEN
                rs:=Resume(tcb^.pid);
               END(*IF*)              |
      WRITER : IF tcb^.wrBlocked THEN
                SemV( tcb^.socPtr^.ioGuard);
                tcb^.wrBlocked := FALSE;
               END(*IF*)     |
      CLOSER,
      OPENER : IF tcb^.ocBlocked THEN
                SemV(tcb^.socPtr^.ioGuard);
                tcb^.ocBlocked:=FALSE;
               END(*IF*);    |
      ALL    : IF SemQ(tcb^.socPtr^.ioGuard)<0 THEN
                SemV(tcb^.socPtr^.ioGuard);
                tcb^.rdBlocked:=FALSE;
                tcb^.wrBlocked:=FALSE;
                tcb^.ocBlocked:=FALSE;
               ELSIF tcb^.attached THEN
                rs:=Resume(tcb^.pid);
               END(*IF*);
      END(*CASE*)
END WakeUp;

(* This procedure will abort an active TCP connection *)

PROCEDURE TCPAbort( VAR tcb:TCBPtr; error:Status );
BEGIN
  ClearTimerEvents(tcb);
  tcb^.flags.rcvDone:=TRUE;
  tcb^.flags.sendDone:=TRUE;
  tcb^.status:=error;
  WakeUp(tcb,ALL);
END TCPAbort;

(* This procedure will set a Delete timer event for a specific TCB *)

PROCEDURE TCPWait(tcb:TCBPtr);
BEGIN
  ClearTimerEvents(tcb);
  SetTimerEvent(tcb,DELETETCB,TCPTWOMSL);
END TCPWait;

(* This procedure calculates a window advertisement and implements 
   receiver side silly window avoidance. *)

PROCEDURE TCPRWindow(VAR tcb:TCBPtr):CARDINAL;
VAR
  windSz,value:CARDINAL;
BEGIN
  windSz:=BuffSz-tcb^.rcvBuffCnt;
  WITH tcb^ DO
    IF tcbState<ESTABLISHED THEN
      RETURN windSz;
    END(*IF*);

(*Apply silly window avoidance*)
(*wait for 1/4 receiver buffer and 1 max-sized seg. before opening a
  window. ie. make sure a non trivial space is available*)

    IF (windSz*4<BuffSz) OR (windSz<MaxSegSz) THEN
      windSz:=0;
    END;
    value:=CARDINAL(tcb^.winSeq-tcb^.rcvNext);
    IF value>windSz THEN
      windSz:=value;
    END;
    tcb^.winSeq:=tcb^.rcvNext+ADDRESS(windSz);
    RETURN windSz;
  END(*WITH*);
END TCPRWindow;
                                          
(* Calculate the round trip time for a packet *)

PROCEDURE TCPRtt(VAR tcb:TCBPtr);
VAR
  rtt : Ticks;                     (*Raw round trip*)
  delta : ADDRESS;                 (*Deviation from smoothed*)
BEGIN
  WITH tcb^ DO
   DeleteTimerEvent(tcb,RETRANSMITSEG,rtt);
   IF (rtt<>0) AND (tcb^.outState<>RETRANSMIT) THEN
     IF (smoothRndTripTime=0) THEN
        smoothRndTripTime:=ADDRESS(rtt)*8;
     END(*IF*);
     delta:=rtt-smoothRndTripTime DIV 8;
     smoothRndTripTime:=smoothRndTripTime+delta;
     IF (delta>HalfSeq) THEN
        delta:=0FFFFFFFFH - delta;
     END(*IF*);
     rndTripDevEst:=rndTripDevEst+Ticks(delta-(rndTripDevEst DIV 4));
     retransTmoVal:=(smoothRndTripTime DIV 4 + rndTripDevEst) DIV 2;
     IF retransTmoVal<TCPMINRXT THEN
       retransTmoVal:=TCPMINRXT;
     END(*IF*);
   END(*IF*);
   IF congWinSz<slowStartThresh THEN
     congWinSz:=congWinSz+MaxSegSz;
   ELSE
     congWinSz:=congWinSz+CARDINAL(MaxSegSz*MaxSegSz DIV ADDRESS(congWinSz));
   END(*IF*);
  END(*WITH*);
END TCPRtt;

(*This procedure will remove a TCB from the data structures*)

PROCEDURE TCBDeallocate( VAR tcb:TCBPtr );              
VAR              
  soc : SocHandle;
  f : FragPtr;
BEGIN    
  IF tcb=NIL THEN RETURN END;
  SemP(activeSocs.guard);
  ClearTimerEvents(tcb);
  IF tcb^.prev = NIL THEN
    activeSocs.head:=tcb^.next;
  ELSE
    tcb^.prev^.next := tcb^.next;
  END; (* IF *)
  IF tcb^.next <> NIL THEN
    tcb^.next^.prev := tcb^.prev;
  END; (* IF *)
  LOOP
    f := tcb^.fragHead;
    IF f = NIL THEN EXIT; END;
    tcb^.fragHead := tcb^.fragHead^.next;
    DISPOSE( f );
  END; (* LOOP *)
  IF tcb^.sendBuffer <> NIL THEN DISPOSE( tcb^.sendBuffer ); END;
  IF tcb^.rcvBuffer <> NIL THEN DISPOSE( tcb^.rcvBuffer ); END;
  soc := tcb^.soc;
  DISPOSE( tcb );
  socList[soc].tcb := NIL;
  SemV(activeSocs.guard);
END TCBDeallocate;

BEGIN
END N2TCPLib.
