IMPLEMENTATION MODULE N2TCPTimer;

FROM L2Clock      IMPORT Ticks, TickCount, RequestAlarm, CancelAlarm;
                  IMPORT L2Clock;
                  IMPORT L2Errors;
FROM L2Exceptions IMPORT Assert, AssertOk, Handler;
FROM L2Files      IMPORT Users;
FROM L2Kernel     IMPORT MaxPriority, Process, Semaphore, SemInit, SetNoSwitch,
                         SemP, SemV, SemQ, Suspend, Resume, CreateProcess,
                         CurrentProcess;
FROM L2Storage    IMPORT ALLOCATE, DEALLOCATE;
FROM N2TCPMain    IMPORT MaxSocs, SocPtr, TCBPtr, TCB;
FROM N2Errors     IMPORT Status;


  (*********************************

       Message Passing Routines
              to TCOut

   *********************************)


CONST
  MaxMessages = 4 * MaxSocs;

TYPE
  Message  = RECORD
               sc     : SocPtr;
               event  : Event;
             END;
  MessageQ = RECORD
               head,
               tail    : CARDINAL;
               cnt,
               guard   : Semaphore;
               message : ARRAY [0..MaxMessages-1] OF Message;
             END;

VAR
  messQ    : MessageQ;

PROCEDURE PutMessage( sc : SocPtr; event : Event );
  VAR
    l2res : L2Errors.Status;
  BEGIN
    WITH messQ DO
      SemP( guard );
      IF SemQ(cnt) < MaxMessages THEN
        message[head].sc := sc;
        message[head].event := event;
        INC( head );
        IF head = MaxMessages THEN 
          head := 0; 
        END; (* IF *)
        SemV( cnt );
      END; (* IF *)
      SemV( guard );
    END; (* WITH *)
  END PutMessage;

PROCEDURE GetMessage( VAR sc : SocPtr; VAR event : Event );
  BEGIN
    WITH messQ DO
      SemP( cnt );
      SemP( guard );
      sc := message[tail].sc;
      event := message[tail].event;
      INC( tail );
      IF tail = MaxMessages THEN tail := 0; END;
      SemV( guard );
    END; (* WITH *)
  END GetMessage;                           

  (*********************************

          Timer Routines

   *********************************)

TYPE
  TPtr     = POINTER TO TRec;
  TRec     = RECORD
               dt,
               inTime : Ticks;
               event  : Event;
               sc     : SocPtr;
               next   : TPtr;
             END;
  TQueue   = RECORD
               head   : TPtr;
               guard  : Semaphore;
             END;

VAR
  timerProc     : Process;
  timerGrain,                          (* timer process sleep interval *)
  maxDeltaT,                           (* used to check if clock reset *)
  defaultDeltaT : Ticks;               (* used if clock was reset *)
  timeQ         : TQueue;

PROCEDURE TimerEventExists( tcb : TCBPtr; event : Event ) : BOOLEAN;
  VAR
    t : TPtr;
  BEGIN
    SemP( timeQ.guard );
    t := timeQ.head;
    WHILE (t <> NIL) AND ((t^.sc <> tcb^.socPtr) OR (t^.event <> event)) DO
      t := t^.next;
    END; (* WHILE *)
    SemV( timeQ.guard );
    RETURN t <> NIL;
  END TimerEventExists;

PROCEDURE SetTimerEvent( tcb : TCBPtr; event : Event; time : Ticks );
  VAR                                                           
    l2res : L2Errors.Status;
    t, prev, next : TPtr;             
    ticks : Ticks;
  BEGIN
    NEW( t );
    Assert( t <> NIL, 550 );
    t^.event := event;
    t^.sc := tcb^.socPtr;
    t^.dt := time;

   (* ...remove any duplicate events from delta list *)

    DeleteTimerEvent( tcb, event, ticks );

   (*  ...insert event into the delta list. *)

    SemP( timeQ.guard );
    next := timeQ.head;
    prev := NIL;
    WHILE (( next <> NIL ) AND ( next^.dt < t^.dt )) DO
      t^.dt := t^.dt - next^.dt;
      prev := next;
      next := next^.next;
    END; (* DO *)
    IF next <> NIL THEN next^.dt := next^.dt - t^.dt; END;
    t^.inTime := TickCount;                         
    t^.next := next;                               
    IF prev = NIL THEN
      IF timeQ.head = NIL THEN l2res := Resume( timerProc ); END;
      timeQ.head := t;
    ELSE
      prev^.next := t;
    END; (* IF *)
    SemV( timeQ.guard );
  END SetTimerEvent;

PROCEDURE ClearTimerEvents( tcb : TCBPtr );
  VAR
   prev, tmp, t : TPtr;
  BEGIN
    SemP( timeQ.guard );
    prev := NIL;
    t := timeQ.head;
    LOOP
      IF t = NIL THEN SemV( timeQ.guard ); RETURN; END;
      IF t^.sc = tcb^.socPtr THEN
        IF prev = NIL THEN
          timeQ.head := t^.next;
        ELSE
          prev^.next := t^.next;
        END; (* IF *)
        tmp := t;
        t := t^.next;
        IF t <> NIL THEN t^.dt := t^.dt + tmp^.dt; END;(*adjust trailing dt *)
        DISPOSE( tmp );
      ELSE
        prev := t;
        t := t^.next;
      END; (* IF *)
    END; (* LOOP *)
  END ClearTimerEvents;

PROCEDURE DeleteTimerEvent( tcb : TCBPtr; event : Event; VAR ticks : Ticks );
  VAR
   prev, t : TPtr;
  BEGIN
    ticks := 0;
    SemP( timeQ.guard );
    prev := NIL;
    t := timeQ.head;
    WHILE (t <> NIL) AND ((t^.sc <> tcb^.socPtr) OR (t^.event <> event)) DO
      prev := t;
      t := t^.next;
    END; (* WHILE *)
    IF t = NIL THEN SemV( timeQ.guard ); RETURN; END;  (* there was none! *)
    ticks := TickCount - t^.inTime;
    IF prev = NIL THEN
      timeQ.head := t^.next;
    ELSE
      prev^.next := t^.next;
    END; (* IF *)
   (* ...must adjust trailing dt if there is one! *)
    IF t^.next <> NIL THEN t^.next^.dt := t^.next^.dt + t^.dt; END;
    DISPOSE( t );
    SemV( timeQ.guard );
  END DeleteTimerEvent;

PROCEDURE SetTimerGrain( tGrainSz : Ticks ) : Status;
  BEGIN
    IF Users[CurrentProcess] <> 0 THEN RETURN PrivViol; END;
    SemP( timeQ.guard );
    timerGrain := tGrainSz;
    maxDeltaT := 100 * tGrainSz;
    defaultDeltaT := 10 * tGrainSz;
    SemV( timeQ.guard );                                   
    RETURN Ok;
  END SetTimerGrain;

PROCEDURE TCPTimer;
  VAR
    now,
    lastTime,
    dt        : Ticks;
    temp      : TPtr;  
    e         : L2Clock.Event;
    l2res     : L2Errors.Status;
  BEGIN
    lastTime := TickCount;
    LOOP      
      AssertOk( RequestAlarm( Handler( Resume ), timerProc, timerGrain,
                                                       FALSE, e ), 551 );
      Suspend;                 
      l2res := CancelAlarm( e );
      SemP( timeQ.guard );
      IF timeQ.head = NIL THEN 
        SemV( timeQ.guard ); 
        Suspend; 
        SemP( timeQ.guard );
        lastTime := TickCount;
      END; (* IF *)       
      SetNoSwitch( TRUE );
      
     (* calculate time elapsed since the last time we processed list *)

      now := TickCount;
      dt := now - lastTime;
      lastTime := now;
      IF dt > maxDeltaT THEN  (* just in case clock was modified *)
        dt := defaultDeltaT;
      END; (* IF *)

     (* signal TCPOut for all events which have become current *)
                               
      WHILE ( timeQ.head <> NIL ) AND ( timeQ.head^.dt <= dt ) DO
        dt := dt - timeQ.head^.dt;
        PutMessage( timeQ.head^.sc, timeQ.head^.event );
        temp := timeQ.head;
        timeQ.head := timeQ.head^.next;
        DISPOSE( temp );
      END; (* WHILE *)     
      IF timeQ.head <> NIL THEN
        timeQ.head^.dt := timeQ.head^.dt - dt;
      END; (* IF *)
      SemV( timeQ.guard );
    END; (* LOOP *)
  END TCPTimer;

VAR
  n2res : Status;
BEGIN
  SemInit( messQ.guard, 1 );
  SemInit( messQ.cnt, 0 );
  messQ.head := 0;
  messQ.tail := 0;
  timeQ.head := NIL;
  SemInit( timeQ.guard, 1 );
  timerGrain := 10;
  maxDeltaT := 1000;
  defaultDeltaT := 100;
  AssertOk( CreateProcess( timerProc, 0, 0, MaxPriority-1, TCPTimer ), 592 );
END N2TCPTimer.
