IMPLEMENTATION MODULE N2UDP;

               IMPORT N2Errors, L2Errors;
FROM N2Base    IMPORT IP1, IPMinLen, IPhdrlen, IPtlen, IPproto, IPsaddr,
                      IPdaddr, IPAddr, HostIP, NetHost2, HostNet2, HostNet4,
                      NetHost4, ChkSum;
FROM N2Packets IMPORT PacketLen, PacketQ, PacketPtr, Packet, InitPacketQ,
                      GetPacket, NewPacket, DisposePacket, PutPacket, DeleteQ;
FROM N2IP      IMPORT IPSend, IPRegister;
FROM L2Kernel  IMPORT CurrentProcess, Process, ProcMax, Semaphore, SemInit,
                      SemP, SemV, Guard, UnGuard, SetNoSwitch;
FROM L2Files   IMPORT Users;
FROM L2Clock   IMPORT Ticks, Time, GetTime; 
FROM L2Exceptions IMPORT Assert;
FROM L2Storage IMPORT ALLOCATE, DEALLOCATE;
FROM SYSTEM    IMPORT ADDRESS;

CONST
    UDPProtocol     = 21C;        (* UDP protocol number is 17 *)
    UDPHdrLen       = 8;
    UDP1            = IP1 + IPMinLen;    (* first octet UDP header *)
    Data1           = UDP1 + UDPHdrLen;  (* first octet UDP data   *)
    MaxDataLen      = PacketLen - Data1 + 1;
    MaxUDPQPackets  = 3;
    NoPrivFloorPort = 1000;       (* ports 1..999 are privileged *)

    (* ...following are UDP field pointers *)

    UDPsport        = UDP1;
    UDPforport      = UDP1 + 2;
    UDPlen          = UDP1 + 4;
    UDPchksum       = UDP1 + 6;

TYPE
    PCBPtr   =  POINTER TO PCB;
    PCB      =  RECORD                  (* Port Control Block *)
                   port  : Port;
                   pktQ  : PacketQ;
                   pkt   : PacketPtr;
                   tmos,                (*                   *)
                   rcvd,                (*   port            *)
                   sent  : ADDRESS;     (*     statistics    *)
                   start : Time;        (*                   *)
                   next  : PCBPtr;
                END;
    PCBRec   =  RECORD                  
                   guard : Semaphore;
                   pcb   : PCBPtr;
                END;
    PCBHead  =  RECORD                  
                 guard : Semaphore;
                 head  : PCBPtr;
                END;

VAR
   nextPortNum  : Port;       (* next unprivileged port *)
   portList     : ARRAY Process OF PCBRec;
   aPorts       : PCBHead;    (* active port list *)
   globalStats  : RECORD
                    maxPorts,
                    openPorts,
                    badPktsNoPort,
                    badPktsChksum : CARDINAL;
                    pktsQd,
                    pktsSent,
                    pktsRcvd      : ADDRESS;
                  END;

PROCEDURE SetUDPPort( port : Port ) : N2Errors.Status;
  VAR
    n2res : N2Errors.Status;
  BEGIN
    WITH portList[CurrentProcess] DO
      SemP( guard );

       (* ...port = 0 -> close caller's port. *)

      IF port = 0 THEN
        n2res := DeletePort( CurrentProcess );
        SemV( guard );
        RETURN n2res;
      END;

       (* ...privileged user -> singular attempt to secure port requested. *)

      IF Users[CurrentProcess] = 0 THEN
        n2res := AddPort( port );
        SemV( guard );
        RETURN n2res;
      END;

       (* ...unprivileged user -> allocate non reserved port. *)

      LOOP
        n2res := AddPort( nextPortNum );
        SetNoSwitch( TRUE );              (* make ops on nextPortNum atomic *)
        IF nextPortNum = 0FFFFH THEN
          nextPortNum := NoPrivFloorPort;
        ELSE
          INC(nextPortNum);
        END; (* IF *)
        SetNoSwitch( FALSE );
        IF n2res = N2Errors.Ok THEN EXIT; END;
      END; (* LOOP *)
      SemV( guard );
    END; (* WITH *)
    RETURN n2res;
  END SetUDPPort;

PROCEDURE Send( VAR data : ARRAY OF CHAR; len : CARDINAL;
                forPort : Port; forHost : IPAddr ) : N2Errors.Status;
  VAR
    i, j, chk  : CARDINAL;
    n2res      : N2Errors.Status;
  BEGIN
    WITH portList[CurrentProcess] DO

       (* ...check for errors -> port, length. *)

      IF pcb = NIL THEN
        RETURN N2Errors.NoPort;
      ELSIF len > (HIGH(data) + 1) THEN
        RETURN N2Errors.BoundsError;
      ELSIF len > MaxDataLen THEN
        RETURN N2Errors.BuffOverFlow;
      END; (* IF *)

     (* ...create packet to send. *)

      NewPacket( pcb^.pkt );
      IF pcb^.pkt = NIL THEN
        RETURN N2Errors.NoHeap;
      END;
      i := Data1;
      FOR j := 0 TO len - 1 DO    (* copy data to packet *)
        pcb^.pkt^.data[i] := data[j];
        INC( i );
      END;
      pcb^.pkt^.plen := i;

     (* ...set IP fields -> protocol, source address, foreign address *)

      pcb^.pkt^.data[IPproto] := UDPProtocol;
      pcb^.pkt^.data[IPsaddr] := HostIP[0];
      pcb^.pkt^.data[IPsaddr + 1] := HostIP[1];
      pcb^.pkt^.data[IPsaddr + 2] := HostIP[2];
      pcb^.pkt^.data[IPsaddr + 3] := HostIP[3];
      HostNet4( forHost, IPdaddr, pcb^.pkt^.data );

     (* ...fill in UDP header -> source port, foreign port, length, chksum *)

      HostNet2( pcb^.port, UDPsport, pcb^.pkt^.data );
      HostNet2( forPort, UDPforport, pcb^.pkt^.data );
      HostNet2( len + UDPHdrLen, UDPlen, pcb^.pkt^.data );
      pcb^.pkt^.data[UDPchksum] := 0C;      (* make sure chksum *)
      pcb^.pkt^.data[UDPchksum +1] := 0C;   (* fields are 0's   *)
      ChkSum( chk, UDP1, len + UDPHdrLen, pcb^.pkt^.data, TRUE );
      IF chk = 0 THEN chk := 0FFFFH; END;     (* ..we don't want +ve zero! *)
      HostNet2( chk, UDPchksum, pcb^.pkt^.data );

      (* ...send it! *)

      SemP( guard );
      n2res := IPSend( pcb^.pkt );
      pcb^.pkt := NIL;
      INC( globalStats.pktsSent );
      INC( pcb^.sent );
      SemV( guard );
      RETURN n2res;
    END; (* WITH *)
  END Send;

PROCEDURE Receive( VAR data : ARRAY OF CHAR; VAR len : CARDINAL; tmo : Ticks;
                VAR forPort : Port; VAR forHost : IPAddr ) : N2Errors.Status;
  VAR
    ipHLen, ipTLen, 
    udp1, pos, i    : CARDINAL;
    n2res           : N2Errors.Status;
  BEGIN
    WITH portList[CurrentProcess] DO

      (* ...error checking *)

      IF pcb = NIL THEN 
        RETURN N2Errors.NoPort;
      END;

      n2res := GetPacket( pcb^.pkt, pcb^.pktQ, tmo );
      IF n2res <> N2Errors.Ok THEN
        IF n2res = N2Errors.TimedOut THEN INC( pcb^.tmos ); END;
        RETURN n2res;
      END; (* IF *)     

     (* .. calculate offsets and data length. *)

      NetHost2( ipTLen, IPtlen, pcb^.pkt^.data );
      ipHLen := 4 * ( CARDINAL( pcb^.pkt^.data[IPhdrlen] ) MOD 10H );
      udp1 := IP1 + ipHLen;
      NetHost2( len, udp1 + 4, pcb^.pkt^.data );
      len := len - UDPHdrLen;
      NetHost2( forPort, udp1, pcb^.pkt^.data );
      NetHost4( forHost, IPsaddr, pcb^.pkt^.data ); 
      pos := udp1 + UDPHdrLen;       
                                           
     (* .. copy UDP data to callers buffer. *)

      IF len > HIGH( data ) + 1 THEN len := HIGH( data ) + 1; END;
      FOR i := 0 TO len - 1 DO
        data[i] := pcb^.pkt^.data[pos];
        INC( pos );
      END; (* FOR *)                         

      DisposePacket( pcb^.pkt );
      pcb^.pkt := NIL;
      INC( globalStats.pktsRcvd );
      INC( pcb^.rcvd );
      RETURN N2Errors.Ok;
    END; (* WITH *)
  END Receive;

PROCEDURE LockUDPPort( pid : Process );
  BEGIN
    SemP( portList[pid].guard );    (* .. locks pid out of N2UDP *)
  END LockUDPPort;

PROCEDURE ClearUDPPort( pid : Process );
  VAR
    n2res : N2Errors.Status;
  BEGIN
    n2res := DeletePort( pid );     (* .. remove pid's PCB if any *)
    SemV( portList[pid].guard );    (* .. NOT LockUDPPort( pid )  *)
  END ClearUDPPort;

PROCEDURE AddPort( port : Port ) : N2Errors.Status;
  VAR
    m, prev : PCBPtr;
    cnt     : CARDINAL;
  BEGIN
    SemP( aPorts.guard );

   (* .. check that 'port' is not already in use. *)

    m := aPorts.head;
    WHILE (m <> NIL) AND (m^.port <> port) DO m := m^.next; END;
    IF m <> NIL THEN
      SemV( aPorts.guard );
      RETURN N2Errors.PortInUse;
    END; (* IF *)

   (* .. allocate a PCB to this process. *)

    m := aPorts.head;
    prev := NIL;
    cnt := 0;
    WHILE (m <> NIL) AND (m^.pktQ.pid <> CurrentProcess) DO 
      prev := m;
      m := m^.next;
      INC( cnt );
    END; (* WHILE *)             
    IF m = NIL THEN                   (* process has no PCB so create one. *)
      NEW( m );
      IF m = NIL THEN
        SemV( aPorts.guard );
        RETURN N2Errors.NoHeap;
      END; (* IF *)
      m^.next := NIL;
      m^.pkt := NIL;
      portList[CurrentProcess].pcb := m;
      IF prev = NIL THEN
        aPorts.head := m;
      ELSE
        prev^.next := m;
      END; (* IF *)
      INC(cnt);
      globalStats.openPorts := cnt;
      IF globalStats.maxPorts < cnt THEN
        globalStats.maxPorts := cnt;
      END; (* IF *)
    ELSE      (* process has a PCB so just flush it's queue. *)
      DeleteQ( m^.pktQ );
    END; (* IF *)                  

   (* .. PCB initialization .. *)

    InitPacketQ( m^.pktQ, MaxUDPQPackets );
    m^.port := port;
    m^.tmos := 0;
    m^.sent := 0;
    m^.rcvd := 0;
    GetTime( m^.start ); 

    SemV( aPorts.guard );
    RETURN N2Errors.Ok;
  END AddPort;

PROCEDURE DeletePort( pid : Process ) : N2Errors.Status;
  VAR
    m, prev : PCBPtr;
  BEGIN
    SemP( aPorts.guard );
    m := aPorts.head;
    prev := NIL;
    WHILE (m <> NIL) AND (m^.pktQ.pid <> pid) DO (* .. search for pid's PCB *)
      prev := m;
      m := m^.next;
    END; (* WHILE *)
    IF m = NIL THEN                  (* pid has no PCB *)
      SemV( aPorts.guard );
      RETURN N2Errors.NoPort;
    ELSIF prev = NIL THEN            (* delete first PCB *)
      aPorts.head := m^.next;
    ELSE                             (* general case PCB deletion *)
      prev^.next := m^.next;
    END;
    DeleteQ( m^.pktQ );
    IF m^.pkt <> NIL THEN DisposePacket( m^.pkt ); END;
    DISPOSE( m );
    portList[pid].pcb := NIL;
    DEC( globalStats.openPorts );

    SemV( aPorts.guard );
    RETURN N2Errors.Ok;
  END DeletePort;

PROCEDURE UDPIn( p : PacketPtr );
  VAR
    m          : PCBPtr;
    port       : Port;
    j, chk1, chk2, ipLen, udp1, udpLen : CARDINAL;
  BEGIN
    SemP( aPorts.guard );    

   (* .. calculate offsets ( incoming datagrams may have ip options! ) 
       and extract destination port ( receiver port this host! )      *)

    NetHost2( ipLen, IPtlen, p^.data );
    udp1 := IP1 + 4 * ( CARDINAL( p^.data[IPhdrlen] ) MOD 10H );
    udpLen := IP1 + ipLen - udp1;
    NetHost2( port, udp1 + 2, p^.data ); 
                          
   (* .. search for receiving port. *)

    m := aPorts.head;
    WHILE (m <> NIL) AND (m^.port <> port) DO 
       m := m^.next; 
    END; 
    IF m = NIL THEN                                        (* Port not open *)
      DisposePacket( p );
      INC( globalStats.badPktsNoPort );
      SemV( aPorts.guard );
      RETURN;
    END; (* IF *)                                       

   (* checksum validation. *)
        
    NetHost2( chk1, udp1 + 6, p^.data );
    IF chk1 <> 0 THEN
      p^.data[udp1 + 6] := 0C;
      p^.data[udp1 + 7] := 0C;
      ChkSum( chk2, udp1, udpLen, p^.data, TRUE );
      IF NOT ( chk1 = chk2 ) OR (( chk1 = 0FFFFH ) AND ( chk2 = 0 )) THEN
        DisposePacket( p );                                  (* Bad Checksum *)
        INC( globalStats.badPktsChksum );
        SemV( aPorts.guard );
        RETURN;
      END; (* IF *)                 
    END; (* IF *)                                           

   (* queue packet on port *)
        
    PutPacket( p, m^.pktQ );
    INC( globalStats.pktsQd );

    SemV( aPorts.guard );
  END UDPIn;

PROCEDURE ProcessStats( pid : Process; VAR port : Port;
                         VAR tmos, sent, rcvd : ADDRESS;
                         VAR start : Time ) : N2Errors.Status;
  BEGIN
    WITH portList[pid] DO
      Guard( guard );
      IF pcb = NIL THEN
        UnGuard( guard );
        RETURN N2Errors.NoPort;
      END; (* IF *)
      tmos := pcb^.tmos;
      port := pcb^.port;
      sent := pcb^.sent;
      rcvd := pcb^.rcvd;
      start := pcb^.start;
      UnGuard( guard );
    END; (* WITH *)  
    RETURN N2Errors.Ok;
  END ProcessStats;

PROCEDURE GlobalStats( VAR maxPorts, openPorts,
                       badPktsNoPort, badPktsChksum : CARDINAL;
                       VAR pktsQd, pktsSent, pktsRcvd : ADDRESS );
  BEGIN
    maxPorts := globalStats.maxPorts;
    openPorts := globalStats.openPorts;
    badPktsNoPort := globalStats.badPktsNoPort;
    badPktsChksum := globalStats.badPktsChksum;
    pktsQd := globalStats.pktsQd;
    pktsSent := globalStats.pktsSent;
    pktsRcvd := globalStats.pktsRcvd;
  END GlobalStats;
                 
VAR
  i : CARDINAL;
BEGIN
  nextPortNum := NoPrivFloorPort;
  aPorts.head := NIL;
  SemInit( aPorts.guard, 1 );
  FOR i := 1 TO ProcMax DO
    portList[i].pcb := NIL;
    SemInit( portList[i].guard, 1 );
  END;
  WITH globalStats DO
    maxPorts := 0;
    openPorts := 0;
    badPktsNoPort := 0;
    badPktsChksum := 0;
    pktsQd := 0;
    pktsSent := 0;
    pktsRcvd := 0;
  END; (* WITH *)
  Assert( IPRegister( UDPIn, UDPProtocol ) = N2Errors.Ok, 517 );
END N2UDP.
