IMPLEMENTATION MODULE Compiler;  (* Compiler for Update *)

(* G.R. Guenther 1988 05 30  *)
(* Modified: GRG 1988 06 01  Added the string operations in Cassign *)

(* This module hides the syntax of the Update description files.
   It compiles the current input file onto a 'StatSeq' (as defined
   in Interpreter).  Any formal errors are detected and reported
   in this phase.

   The MODULE clauses are sorted into the required update order
   and compiled into guarded statements.  It is assumed that
   the interpreter will use the guards to determine if the
   particular step is required.

   The syntax of a description file is defined in the users' guide:

#START
     3. Tutorial Introduction to the Update Description Language

     A description file contains three kinds of clauses: MODULE,
     PROCEDURE and BEGIN clauses.  The clauses may be arranged in any
     order, as long as there is only one BEGIN clause and it is the
     last one in the description.

     A typical description file will contain several MODULE clauses.
     Each MODULE clause describes a particular manufacturing step in
     the rebuilding of the system.  The clause names the required
     components for this step, the products of the operations and the
     commands that must be executed to perform this step.

     The commands may be listed explicitly in the MODULE clause, or
     they may be encapsulated in a procedure that is named instead.
     For example:

          MODULE "OpCode.SYM" USES "OpCode.DEF"
            BY * M2C OpCode.DEF; END;

     describes a simple manufacturing step with an explicit command.
     For another example:

          MODULE "OpCode.OBJ" USES "OpCode.MOD", "OpCode.SYM"
            BY Build("OpCode.MOD") END;

     describes a more complicated manufacturing step in which one
     product requires two components for its construction.  The
     commands are referenced indirectly through the procedure named
     "Build". In yet another example:

          MODULE "OpCodes.DEF", "OpCodes.MOD" USES "OpCode.TAB"
            BY * MakeTable ; END;

     the MODULE clause describes a manufacturing step that produces
     two outputs at once from a single input.

     When Update interprets these clauses, it will select and order
     them automatically so that only necessary steps are performed and
     that required components are ready before they are used.

     The actual building steps, whether explicit or contained in a
     procedure, are expressed as a sequence of statements.  The
     command statement is the most important of these; it invokes the
     system's command interpreter to execute a program (such as a
     compiler or linker). Command statements begin with the '*' symbol
     and extend to the next ';'.  You will see other kinds of
     statements shortly.

     A collection of statements may be grouped together and named in a
     PROCEDURE clause, as in the example:

          PROCEDURE  BuildSystem();
            * Del System.LNK ;
            * M2C System ;
            END;

     To be really useful, procedures need parameters.  Update allows
     procedures to be qualified with any number of parameters.  These
     parameters are considered to be character string variables.  One
     way that parameters and variables can be used is to include their
     contents in command statements.  This is done by including the
     variable name, surrounded by "back ticks", in the command text.
     For example:

          PROCEDURE  Build(file);
            * M2C `file`;
            END;

     If this procedure is called by Build("Duke") then the command

          Modula Duke

     would be executed.

     Variables are not limited to procedure parameters, however.  They
     may be defined and given values when Update is invoked. For
     example:

          Update King map install="A:"

     starts Update with the variable 'map' set to "T" and the variable
     'install' set to "A:".

     Variables may also be used in two other kinds of statements,
     assignments and conditionals.  Assignments can be used to give a
     variable a value during the execution of the rebuilding process.
     (A variable name that has never been given a value in a parameter
     list, on the command line or in an assignment statement is
     assumed to be an empty string.) Here is a typical assignment
     statement:

          mapflag := "/NOM";

     To use conditional statements effectively, you should know that
     variables and literal strings may be combined in relational
     expressions using the operators: = (for equality), # (inequality)
     and <= (string inclusion).  Relational expressions can be
     combined using parentheses and the logical operators: AND, OR and
     NOT.

     A conditional statement looks like:

          IF map = "T" THEN * Link `file`;
                       ELSE * Link `file`/NOM; END;

     The ELSE clause is optional (as in Modula-2).  The previous
     example could also be written as:

          IF map # "T" THEN mapflag := "/NOM" END;
          * Link `file``mapflag` ;

     Many applications of Update will share common procedures. The
     INCLUDE directive tells Update to read from another file as if it
     were part of the description file.  INCLUDE directives are
     recognised as clauses; consequently, included files should
     contain complete clauses.  Sometimes, it is necessary to arrange
     for certain statements to execute before any of the automatically
     sequenced rebuilding operations.  This can be done by including
     these statements in a BEGIN clause.  There is a special
     statement, ABORT, that causes Update to stop (prematurely).
     Update description files may also include comments, in the usual
     Modula-2 syntax.

     A final example:

          (* A description file for a small program,
             with some complications *)
          MODULE "Earl.EXE" USES "Earl.OBJ", "InOut.OBJ"
             BY Build("Earl.OBJ") END;
          MODULE "Earl.OBJ" USES "Earl.MOD", "InOut.SYM"
             BY Build("Earl.MOD") END;
          PROCEDURE Build(file);
          IF ".OBJ" <= file THEN * M2L `file``mapflag` ;
            ELSE IF (".DEF" <= file) OR (".MOD" <= file)
                   THEN * M2C `file`;
                   ELSE * Echo Unexpected file type in: `file`;
                        ABORT
                   END
            END
          END;
          BEGIN IF map # "T" THEN mapflag="/NOM" END;
          END.

     4. Formal Syntax of the Update Description Language

     <description> ::= <cl_list> [ ';' <begin> ] '.'
     <cl_list>     ::= <clause> ( ';' <clause> )*
     <clause>      ::= <step> | <include> | <proc>
     <include>     ::= 'INCLUDE' <literal>
     <step>        ::= 'MODULE' <list> 'USES' <list>
                       'BY' <stat_seq> 'END'
     <list>        ::= <literal> ( ',' <literal> )*
     <begin>       ::= 'BEGIN' <stat_seq> 'END'
     <stat_seq>    ::= <statement> | <command> |
                       <command> <stat_seq> |
                       <statement> ';' <stat_seq>
     <statement>   ::= <cond> | <call> | 'ABORT' | <assign>
     <proc>        ::= 'PROCEDURE' <ident> '(' <ident_seq> ')' ';'
                       <stat_seq> 'END'
     <call>        ::= <ident> '(' <expr_seq> ')'
     <ident>       ::= <letter> ( <letter> | <digit> )*
     <ident_seq>   ::= <empty> | <ident> ( ',' <ident> )*
     <expr_seq>    ::= <empty> | <expr> ( ',' <expr> )*
     <command>     ::= '*' ( <com_char> | <subst> )* ';'
     <com_char>    ::= <printable> - ';' - '`'
     <subst>       ::= '`' <ident> '`'
     <expr>        ::= <ident> | <literal>
     <literal>     ::= '"' <string> '"'
     <string>      ::= <str_char>*
     <str_char>    ::= <printable> - '"'
     <cond>        ::= 'IF' <boolean> 'THEN' <stat_seq>
                           [ 'ELSE' <stat_seq> ] 'END'
     <assign>      ::= <ident> ':=' <expr> [ ('+' | '-') <expr> ]
     <boolean>     ::= <bool_term> ( 'OR' <bool_term> )*
     <bool_term>   ::= <bool_fact> ( 'AND' <bool_fact> )*
     <bool_fact>   ::= '(' <boolean> ')' | 'NOT' <bool_fact> | <rel>
     <rel>         ::= <expr> <relop> <expr>
     <relop>       ::= '=' | '#' | '<='
#STOP
*)                                          

FROM Interpreter IMPORT 
        StatKind, ExprKind, BoolKind, RelaKind, BTptr, BFptr, SopKind,
        BEptr, EXptr, REptr, ExprSeq, StrSeq, StatSeq, ProcList,
        Relation, BoolFact, BoolTerm, BoolExpr, Expr, ExprCell,
        ProcDecl, StrCell, StatCell, Procs;
FROM UpdLex IMPORT Lex, LexSym, LexText, StrPtr, StrType, Symbol;
FROM TextIO IMPORT Eof, ErrorDisplay, OpenInput, CloseInput;
FROM Errors IMPORT Error, Assert;
FROM Symbols IMPORT InitSymTab, SymTab, Enter, LookUp, Release, Null;
FROM TopoSort IMPORT Insert, GetNode;
FROM R2Storage IMPORT ALLOCATE, DEALLOCATE;

VAR     uniq : SymTab;   (* used to discover when there is more than
                            one way to build a module. *)
        begincode, beginlast : StatSeq;  (* First and last statements
                                            of the compiled BEGIN block. *)

PROCEDURE  Next;   (* Lex with auto-close (for INCLUDE) *)

BEGIN   Lex;
        IF Eof THEN 
          CloseInput;
          Lex;
          END;
END Next;

PROCEDURE  sin( VAR msg : ARRAY OF CHAR);

BEGIN   ErrorDisplay;
        Error("Syntax error",msg);
END sin;

PROCEDURE  Expect( sy : Symbol);

BEGIN   IF LexSym = sy THEN RETURN END;
        CASE sy OF
           semiSY    :  sin('";" expected');
        |  lparenSY  :  sin('"(" expected');
        |  rparenSY  :  sin('")" expected');
        |  bySY      :  sin('"BY" expected');
        |  endSY     :  sin('"END" expected');
        |  thenSY    :  sin('"THEN" expected');
        |  usesSY    :  sin('"USES" expected');
        |  identSY   :  sin('Identifier expected');
        |  stringSY  :  sin('String expected');
           ELSE Assert(FALSE,"Unexpected parameter to Expect");
           END;
END Expect;

(* The C... routines each compile (by recursive descent) one of
   the syntactic categories of the Update description language. *)

PROCEDURE  Cboolfact( VAR b : BFptr);

BEGIN   NEW(b);
        WITH b^ DO
          IF LexSym = lparenSY THEN
            kind := parenBF;
            Next; Cboolean(be);
            Expect(rparenSY);
            Next;
          ELSIF LexSym = notSY THEN
            kind := notBF;
            Next; Cboolfact(bf);
          ELSE
            kind := relopBF;
            NEW(re);
            WITH re^ DO
              Cexpr(e1);
              Next; 
              IF LexSym = eqSY THEN op := eqOP
              ELSIF LexSym = neqSY THEN op := neqOP
              ELSIF LexSym = subSY THEN op := subsOP
              ELSE sin('"=", "#" or "<=" expected');
                END;
              Next;
              Cexpr(e2);
              Next;
              END;
            END;
          END;
END Cboolfact;

PROCEDURE  Cboolterm( VAR b : BTptr);

VAR     p : BTptr;

BEGIN   NEW(b); p := b;
        LOOP WITH p^ DO
               Cboolfact(bf);
               nx := NIL;
               IF LexSym <> andSY THEN EXIT END;
               Next;
               NEW(nx);
               p := nx;
               END;
             END;
END Cboolterm;

PROCEDURE  Cboolean( VAR b : BEptr);

VAR     p : BEptr;

BEGIN   NEW(b); p := b;
        LOOP WITH p^ DO
               Cboolterm(bt);
               nx := NIL;
               IF LexSym <> orSY THEN EXIT END;
               Next;
               NEW(nx);
               p := nx;
               END;
             END;
END Cboolean;

PROCEDURE  Cexprseq( VAR s : ExprSeq);

VAR     p : ExprSeq;

BEGIN   s := NIL;
        LOOP Next;
             IF (LexSym <> identSY) AND (LexSym <> stringSY)
               THEN RETURN END;                            
             IF s = NIL
               THEN NEW(p); s := p;
               ELSE NEW(p^.nx); p := p^.nx;
               END;
             p^.nx := NIL;
             Cexpr(p^.ex);
             Next;
             IF LexSym <> commaSY THEN RETURN END;
             END;
END Cexprseq;

PROCEDURE  Cexpr( VAR x : EXptr);

BEGIN   NEW(x);
        WITH x^ DO
          IF LexSym = identSY THEN rk := idenXP
            ELSIF LexSym = stringSY THEN rk := litrXP
            ELSE sin("Literal or variable expected");
            END;
          pt := LexText;
          END;
END Cexpr;

PROCEDURE  Ccond( VAR t : StatSeq);

VAR     zz : StatSeq;

BEGIN   NEW(t);
        WITH t^ DO
          nx := NIL;
          sk := condST;
          Next;
          Cboolean(bx);
          Expect(thenSY);
          Cstatseq(tc,zz);
          IF LexSym = elseSY THEN Cstatseq(ec,zz) ELSE ec := NIL END;
          Expect(endSY);
          Next;
          END;
END Ccond;

PROCEDURE  Ccall( id : StrPtr; VAR t : StatSeq);

(* Ccall also enters the procedure name into the Procs list
   (with a NIL body).  At the end of the compile, the list
   is checked for incomplete entries. *)

VAR     q : ProcList;

BEGIN   NEW(t);
        WITH t^ DO
          nx := NIL;
          sk := callST;
          pn := id;
          Cexprseq(ps);
          Expect(rparenSY);
          Next;
          END;
        q := Procs;    (* now we have to enter this for the check *)
        WHILE (q <> NIL) AND (q^.nm <> id) DO q := q^.nx END;
        IF q = NIL THEN
          NEW(q);
          q^.nx := Procs;
          Procs := q;
          q^.nm := id;
          q^.ss := NIL;
          END;
END Ccall;

PROCEDURE  Cassign( id : StrPtr; VAR t : StatSeq);

(* id has already been parsed, as has the getsSY.  Build a
   StatSeq on t.  

   The language has been extended from the original.  The
   assignment statement now has the syntax:

        <assign> ::= <ident> := <expr> [ ('+'|'-') <expr>]
*)

BEGIN   NEW(t);
        WITH t^ DO
          nx := NIL;
          sk := assignST;
          so := noSO;
          vl := id;
          vx := NIL;
          Next;
          Cexpr(vr);
          Next; 
          IF (LexSym = plusSY) OR (LexSym = minusSY) THEN
            IF LexSym = plusSY THEN so := conSO ELSE so := remSO END;
            Next;
            Cexpr(vx);
            Next;
            END;
          END;
END Cassign;

PROCEDURE  Cabort( VAR t : StatSeq);

BEGIN   NEW(t);
        WITH t^ DO
          nx := NIL;
          sk := abortST;
          END;
        Next;
END Cabort;

PROCEDURE  Ccomm( VAR t : StatSeq);

BEGIN   NEW(t);
        WITH t^ DO
          nx := NIL;
          sk := commST;
          cs := LexText;
          END; 
        Next;
END Ccomm;

PROCEDURE  Cstatseq( VAR get, put : StatSeq);

(* queues statements onto get & put.  If the list is empty, a
   null statement is inserted.
*)

VAR     t : StatSeq;
        id : StrPtr;

BEGIN   NEW(get); put := get;
        WITH get^ DO
          nx := NIL;
          sk := nullST;
          END;
        LOOP Next; t := NIL;
             IF LexSym = ifSY THEN
               Ccond(t);
             ELSIF LexSym = abortSY THEN
               Cabort(t);
             ELSIF LexSym = comandSY THEN
               Ccomm(t);
             ELSIF LexSym = identSY THEN
               id := LexText;
               Next;
               IF LexSym = getsSY THEN
                 Cassign(id,t);
               ELSIF LexSym = lparenSY THEN
                 Ccall(id,t);
               ELSE sin('"(" or ":=" expected')
                 END;
             ELSIF LexSym = semiSY THEN ;  (* permit extras *)
             ELSE EXIT
             END;
             IF t <> NIL THEN
               put^.nx := t;
               put := t;
               END;
             IF LexSym <> semiSY THEN EXIT END;
             END;
        IF get^.nx <> NIL THEN  (* lose the null statement *)
          t := get;
          get := t^.nx;
          DISPOSE(t);
          END;
END Cstatseq;

PROCEDURE  Cstrseq( typ : Symbol; VAR s : StrSeq);

VAR     x : StrSeq;

BEGIN   x := NIL;
        s := NIL;
        LOOP  Next;  
              IF LexSym <> typ THEN RETURN END;
              IF s = NIL 
                THEN NEW(x); s := x;
                ELSE NEW(x^.nx); x := x^.nx;
                END;
              x^.nx := NIL;
              x^.st := LexText;
              Next;
              IF LexSym <> commaSY THEN RETURN END;
              END;
END Cstrseq;

PROCEDURE  Cmodule;

(* Modules disappear into toposort, only to reappear at the end.
   The attribute is a pointer to the guard.

   Although it would be possible to insert each possible
   (output,input) pair, the outputs are chained to each other
   and only one actually depends on the inputs.
*)

VAR     g, zz : StatSeq;
        i, o, po : StrSeq;

BEGIN   NEW(g);
        WITH g^ DO
          sk := guardST;
          Cstrseq(stringSY,mo);
          IF mo = NIL THEN sin("Output modules required") END;
          Expect(usesSY);
          Cstrseq(stringSY,mi);
          IF mi = NIL THEN sin("Input modules required") END;
          Expect(bySY);
          Cstatseq(ss,zz);
          Expect(endSY);
          (* now we have to put this stuff into toposort, and
             check for multiple rules on one output. *)
          o := mo; po := NIL;
          WHILE o <> NIL DO
            IF LookUp(uniq,o^.st) <> Null THEN
              Error("More than one way to rebuild",o^.st^);
              END;
            Enter(uniq,o^.st,o^.st);
            IF po <> NIL THEN Insert(po^.st,o^.st,NIL) END;
            po := o;
            o := o^.nx;
            END;
          i := mi;
          WHILE i <> NIL DO
            Insert(po^.st,i^.st,g);
            i := i^.nx;
            END;
          END;
END Cmodule;

PROCEDURE  Cproc;

VAR     q : ProcList;
        zz : StatSeq;

BEGIN   Next; Expect(identSY);  (* get name *)
        q := Procs;  (* Look to see if it is already there *)
        WHILE (q <> NIL) AND (q^.nm <> LexText) DO q := q^.nx END;
        IF q = NIL THEN
          NEW(q);
          q^.nx := Procs; 
          q^.nm := LexText;
          q^.ss := NIL;
          Procs := q;
          END;
        IF q^.ss <> NIL THEN
          Error("Procedure multiply defined",LexText^);
          END;
        Next; Expect(lparenSY);
        Cstrseq(identSY,q^.pl);
        Expect(rparenSY);
        Cstatseq(q^.ss,zz);
        Expect(endSY);
END Cproc;

PROCEDURE  Cdescr;

VAR     ip : StrPtr;

BEGIN   ip := NIL;
        LOOP  IF ip <> NIL THEN
                OpenInput(ip^,".UPD","Cannot open INCLUDE file",
                           (*force*)FALSE,(*fatal*)TRUE);
                ip := NIL;
                END;
              Next;
              IF LexSym = beginSY THEN
                 Cstatseq(begincode,beginlast);
                 Expect(endSY);
                 Next;
                 IF LexSym <> dotSY THEN
                   sin('BEGIN must be last, "." expected');
                   END;
                 EXIT;
              ELSIF LexSym = includSY THEN
                 Next; Expect(stringSY);
                 ip := LexText;
              ELSIF LexSym = procSY THEN
                 Cproc;
              ELSIF LexSym = moduleSY THEN
                 Cmodule;
              ELSIF LexSym = semiSY THEN ; (* allow extra ; *)
              ELSE 
                 sin("INCLUDE, MODULE, PROCEDURE or BEGIN expected");
                 END;
              IF LexSym <> semiSY THEN Next END;
              IF ip = NIL THEN
                IF LexSym = dotSY THEN EXIT END;
                Expect(semiSY);
                END;             (* otherwise, leave till we get back *)
              END;
        Next;
        IF NOT Eof THEN
          sin('"." not at end of file');
          END;
END Cdescr;

PROCEDURE  Parse() : StatSeq;

(* Parse returns a pointer to the internal representation of
   the current description file.
*)

VAR     g, p : StatSeq;
        s : StrPtr;
        pr : ProcList;

BEGIN   InitSymTab(uniq);
        Procs := NIL;
        begincode := NIL;
        Cdescr;                             (* parse *)
        p := NIL;
        LOOP GetNode(s,g);                  (* extract the guards *)
             IF s = NIL THEN EXIT END;      (* from toposort      *)
             IF g <> NIL THEN
               g^.nx := p;
               p := g;
               END;
             END;
        IF begincode = NIL
          THEN begincode := p       (* no BEGIN clause *)
          ELSE beginlast^.nx := p;  (* link guards to begin block *)
          END;
        Release(uniq);
        pr := Procs;
        WHILE pr <> NIL DO
          IF pr^.ss = NIL THEN 
            Error("Procedure not declared",pr^.nm^);
            END;
          pr := pr^.nx;
          END;
        RETURN begincode;           (* everything is ready now *)
END Parse;

END Compiler.
