IMPLEMENTATION MODULE DMode;  (* Dependence file generation mode of Uses *)

(* G.R. Guenther 1988 05 29  *)
(* Modified: GRG 1989 01 03  LUMOS-2 version *)

(* This module contains the procedures to generate a dependence
   file for a suite of modules.  The required steps are outlined
   in section 3 of the Users' Manual for Uses:

#START
     3. Principles of Operation - D mode

     When Uses examines a module it reads all the IMPORT and FROM
     clauses in the module and keeps a list of the external modules
     named there.  (Uses does not attempt to find local modules.)
     Having built such a list for the root module, the process repeats
     on both the definition and implementation parts of every module
     named.  If any new modules are encountered, they too will have
     such lists built.  If uses discovers an import from a module that
     it cannot locate in the current directory, it assumes that this
     is a library module and does not process it further.

     This process will eventually stop when all referenced modules
     have been located and examined.  The list of module names will
     then be sorted so that every name occurs before any module it
     imports.   This list is then written to the .DEP file in the
     format suggested by the following example:

          King:
            Queen, Duke, (Earl)
          Queen:
          Duke:
            Earl
          Earl:
            SysCalls

     Notice that Queen imports nothing and that SysCalls does not
     occur on the list (it could not be found). Names that are
     referenced in the implementation part of a module, but not in the
     definition part, are enclosed in parentheses (like Earl in the
     example above).
#STOP
*)       

FROM SYSTEM IMPORT ADDRESS;
FROM UsesLex IMPORT IdType, IdPtr, LexText;
FROM TopoSort IMPORT Insert, GetNode;
FROM IdLists IMPORT IdInit, IdList, IdGet, IdPut;
FROM TextIO IMPORT OpenInput, CloseInput, Done, CloseOutput,
                   CreateOutput, PutString, PutChar, PutLn;
FROM StrTable IMPORT Enter;
FROM UsesParse IMPORT Parse;
FROM Errors IMPORT Error;
FROM Extensions IMPORT MODext, DEFext;
IMPORT R2Strings;

VAR     ModList : IdList;  (* List of modules to be processed *)
        SYSTEMp : IdPtr;   (* SYSTEM is a phony module *)

PROCEDURE  ProcessModule( VAR mod, ext : ARRAY OF CHAR; DepList : IdList;
                          strong : BOOLEAN) : BOOLEAN;

(* Reads mod.ext to find all dependencies.  All modules seen are
   added to the ModList and to a dependence list for mod. Strong
   is used as the iat parameter to IdPut.  It is used to distinguish
   the H and V senses of Uses.  Each dependence is also stored
   in the TopoSort module (with the dependence list as the lattr of mod).  

   In order to be able to recover all the dependence lists from TopoSort,
   it is necessary for every module to depend on some other module.
   Conveniently, the special module SYSTEM doesn't actually exist.
   We make all modules depend on system, but ignore it when it is time
   to display the import lists.

   The return value indicates whether the file was found.
*)

VAR     this, dep : IdPtr;
        z : BOOLEAN;

BEGIN   OpenInput(mod,ext,(*no msg*)"",FALSE,FALSE);
        IF NOT Done THEN RETURN FALSE END;
        z := Parse();
        this := LexText;
        WHILE Parse() DO
          dep := LexText;
          IF dep <> SYSTEMp THEN
            IdPut(ModList,dep,FALSE);
            IdPut(DepList,dep,strong);
            END;
          Insert(this,dep,ADDRESS(DepList));  (* into toposort *)
          END;
        CloseInput;
        Insert(this,SYSTEMp,ADDRESS(DepList));
        RETURN TRUE;
END ProcessModule;

PROCEDURE  FormatDEP( root : IdType);
             
(* Formats the .DEP file by extracting the modules in order
   from TopoSort.  It uses the lattr to regain the dependence list
   and formats it into the output file.  The strong attribute
   determines whether to bracket a module name.

   All modules depend on SYSTEM - but we don't print this.
*)

CONST   margin = 72;
        indent = 5;

VAR     mod, dep : IdPtr;
        DepList : IdList;
        first, V : BOOLEAN;
        str : IdType;
        i : CARDINAL;
        a : ADDRESS;

BEGIN   CreateOutput(root,".DEP","Cannot create output file");
        LOOP GetNode(mod,a);
             IF mod = NIL THEN EXIT END;
             IF a <> NIL THEN
               DepList := IdList(a);
               PutString(mod^,0,0);
               PutChar(':'); PutLn;
               first := TRUE;
               FOR i := 1 TO indent DO PutChar(' ') END;
               WHILE IdGet(DepList,dep,V) DO
                 IF dep <> SYSTEMp THEN
                   IF first
                     THEN first := FALSE
                     ELSE PutString(", ",0,0)
                     END;
                   IF V 
                     THEN str := "";
                          R2Strings.Insert(dep^,str,0);
                     ELSE str := "()";
                          R2Strings.Insert(dep^,str,1);
                     END;
                   PutString(str,margin,indent);
                   END;
                 END;
               PutLn;
               END;
             END;
        CloseOutput;
END FormatDEP;

PROCEDURE  ModeD( root : IdType);
  
(* This procedure initialises the ModList, processes the root
   module, and then processes the .MOD and .DEF forms of each
   module on the ModList. It then calls FormatDEP to produce the
   report.

   The same IdList is used for both the H and V lists, but
   the attribute distinguishes the senses.
*)

CONST   fnf = "File not found";

VAR     mod : IdPtr;
        ok : BOOLEAN;
        DepList : IdList;  (* really nasty, this *)
            
BEGIN   IdInit(ModList);
        SYSTEMp := Enter("SYSTEM");
        IdInit(DepList);
        ok := ProcessModule(root,MODext,DepList,(*strong*)TRUE);
        IF NOT ok THEN Error(fnf,root) END;
        WHILE IdGet(ModList,mod,ok) DO
          IdInit(DepList);
          ok := ProcessModule(mod^,DEFext,DepList,(*strong*)TRUE);
          IF ok THEN
            ok := ProcessModule(mod^,MODext,DepList,(*weak*)FALSE);
            IF NOT ok THEN Error(fnf,mod^) END;
            END;
          END;
        FormatDEP(root);
END ModeD;

END DMode.
