IMPLEMENTATION MODULE FileTimes;  (* Supports file time stamp operations *)

(* G.R. Guenther  1988 05 26  *)
(* Modified: GRG  1988 05 31  Removed ConformTimes, replaced with 
                              MarkCurrent.
                  1988 06 06  Speed improvements (complete rewrite)
                  1989 01 03  LUMOS-2 version
                  1990 09 20  Added $USERLIB
*)

(* This module hides all information about how to obtain file creation
   time stamps and how to compare them.  It can also be told to 
   believe that a file is now current.

   This implementation maintains a list of files and their time
   stamps.  If reference is made to a file not on the list, it is
   looked up in its directory.  MarkCurrent adjusts the internal
   list only. 

   If a file is not present and it does not already contain a path
   it is searched for in $MODLIB.
*)

FROM R2Storage IMPORT ALLOCATE; 
FROM R2Strings IMPORT Concatenate;
FROM R2SysCalls IMPORT Time, Status, GetTime;
FROM R2Directories IMPORT SetWild, NextWild, NDate;
FROM UpdLex IMPORT StrPtr, StrType;


TYPE    TimePtr = POINTER TO TimeRec;
        TimeRec = RECORD Name : StrPtr;
                         Date : Time;
                         Next : TimePtr;
                  END;

VAR     TimeList : TimePtr;

PROCEDURE LoadTime( st : StrPtr);

VAR     t : TimePtr;
        n : StrType;
        rs : Status;

BEGIN   Concatenate("",st^,n);
        rs := SetWild(n);
        IF rs = Ok THEN rs := NextWild() END;
        IF rs <> Ok THEN
          Concatenate("$USERLIB-",st^,n);
          rs := SetWild(n);
          IF rs = Ok THEN rs := NextWild() END;
          END;
        IF rs <> Ok THEN
          Concatenate("$MODLIB-",st^,n);
          rs := SetWild(n);
          IF rs = Ok THEN rs := NextWild() END;
          END;
        NEW(t);
        WITH t^ DO
          IF rs <> Ok 
            THEN WITH Date DO
                   day := 0;
                   min := 0;
                   csec := 0;
                   END;
            ELSE Date := NDate;
            END;
          Next := TimeList;
          TimeList := t;
          Name := st;
          END;
END LoadTime;

PROCEDURE  FindEntry( fn : StrPtr) : TimePtr;

(* Search the list - lookup file if necessary *)

VAR     t : TimePtr;

BEGIN   t := TimeList;  
        WHILE t <> NIL DO
          IF fn = t^.Name THEN RETURN t END;
          t := t^.Next;
          END;
        LoadTime(fn);     (* look up this file's time *)
        RETURN TimeList;
END FindEntry;

PROCEDURE CompareTimes( f0, f1 : StrPtr) : BOOLEAN;

(* Returns TRUE if f0 is older than f1. *)
       
VAR     t0, t1 : TimePtr;
        d : INTEGER;

BEGIN   t0 := FindEntry(f0);
        t1 := FindEntry(f1);
        d :=  t0^.Date.day - t1^.Date.day;
        IF d <> 0 THEN RETURN d < 0 END;
        d :=  t0^.Date.min - t1^.Date.min;
        IF d <> 0 THEN RETURN d < 0 END;
        RETURN  t0^.Date.csec < t1^.Date.csec;
END CompareTimes;

PROCEDURE MarkCurrent( f0 : StrPtr);

(* Henceforth, FileTimes will believe that f0 has a time stamp
   of the time this call is executed. This will not alter any 
   operating system data about the file.
*)

VAR     t0 : TimePtr;
        rs : Status;

BEGIN   t0 := FindEntry(f0); 
        rs := GetTime(t0^.Date);
END MarkCurrent;

BEGIN   TimeList := NIL;
END FileTimes.
