IMPLEMENTATION MODULE Interpreter;  (* Execution module of Update *)

(* G.R. Guenther  1988 05 31  *)
(* Modified: GRG  1988 06 01  Added primitive string operations *)
(*           GRG  1988 06 06  Added product check               *)

(* This module interprets the internal form of an Update description.
   Its basic tasks are described in the users' manual (although the
   sort has already been done in Compiler, and the time and date
   code is in FileTimes):

#START
     5. Principles of Operation

     Update sorts the MODULE clauses into an order consistent with the
     dependencies between the manufacturing steps.  It then considers
     each clause.  If any of the product files have creation dates
     earlier than any of their required components then the rebuilding
     statements in that clause are executed.

     If Update is unable to locate a required file to obtain its date,
     it will try to find the file in the Modula-2 library directories.

     Update honours the value of a special variable, Test.  If Test is
     "T", then the selected rebuilding commands are not executed, they
     are only printed out.  Normally, each command is printed to the
     terminal as it is executed.  The special variable Quiet may be
     used to suppress this printing.

     Description files are often generated by the Uses program.  In
     its default mode of operation, Uses writes an INCLUDE statement
     for the file $Build.  This file normally contains a single
     procedure called Build, which is sufficient for ordinary compile
     and link operations.  You should examine $Build to be sure that
     you understand what it does.

     In Update, all variables are global.  Procedure parameters,
     however, are local in scope so procedures can call other
     procedures.
#STOP
*)

(* See definition module for types !  *)

(* This module consists mainly of iXXX routines that each
   interpret some part of the internal form.  There are three
   classes of symbols: procedures, variables and parameters.
   The same identifier can represent objects in more than one
   class.  Procedures are completely independent of the other
   two classes, the same identifier can be a procedure and a variable
   or parameter - the syntax distinguishes the uses.  When
   an identifier is to be evaluated in an expression, possible
   parameter names in the current procedure are checked first.  If
   this does not yield a value, then the global variables are checked.
   IF both searches fail, Null is returned.

   The procedure parameters are maintained on a stack - with
   boundaries between activation records.  Only current parameter
   values are visible.  Parameters can only be set at call time,
   the assignment statement operates on globals only.  Conformance
   between the formal and actual parameters is NOT required.

   Global variables are stored in the vars SymTab, params holds
   the procedure parameters.
*)

FROM UpdLex IMPORT StrPtr, StrType, ChkIdSyntax;
FROM Errors IMPORT Error, Assert;
IMPORT StrTable;
FROM Symbols IMPORT SymTab, InitSymTab, LookUp, Enter, 
                    Null, Mark, Release;
FROM FileTimes IMPORT CompareTimes, MarkCurrent;
FROM Exec IMPORT Execute;
FROM R2Storage IMPORT ALLOCATE, DEALLOCATE;
FROM R2Strings IMPORT Copy, Substring, Remove, Insert, Length,
                      Find, Concatenate;

VAR     params, vars : SymTab;          (* "memory"  *)
        TestFlag, QuietFlag : StrPtr;   (* pointers to magic names *)
        DidSomething : BOOLEAN;         (* for final message *)

PROCEDURE  PreAssign( VAR vl, vr : ARRAY OF CHAR);

(* Immediately executes vl := "vr" *)

VAR     l, r : StrPtr;

BEGIN   IF NOT ChkIdSyntax(vl) THEN
          Error("Invalid identifier syntax",vl);
          END;
        l := StrTable.Enter(vl);
        r := StrTable.Enter(vr);
        Enter(vars,l,r);
END PreAssign;

PROCEDURE  iREptr( r : REptr) : BOOLEAN;

VAR     s1, s2 : StrPtr;

BEGIN   WITH r^ DO
          s1 := iEXptr(e1);
          s2 := iEXptr(e2);
          CASE op OF
             eqOP   : RETURN s1 = s2;
           | neqOP  : RETURN s1 <> s2;
           | subsOP : RETURN Find(s1^,s2^) < Length(s2^);
             ELSE Assert(FALSE,"Unknown relational operator");
             END;
          END;
END iREptr;

PROCEDURE  iBFptr( b : BFptr) : BOOLEAN;

BEGIN   WITH b^ DO
          CASE kind OF
              parenBF : RETURN iBEptr(be);
           |  notBF   : RETURN NOT iBFptr(bf);
           |  relopBF : RETURN iREptr(re);
              ELSE Assert(FALSE,"Unknown boolean factor");
              END;
          END;
END iBFptr;

PROCEDURE  iBTptr( b : BTptr) : BOOLEAN;

VAR     t : BOOLEAN;

BEGIN   t := TRUE;
        WHILE (b <> NIL) AND t (* short circuit *) DO
          t := t AND iBFptr(b^.bf);
          b := b^.nx;
          END;
        RETURN t;
END iBTptr;

PROCEDURE  iBEptr( b : BEptr) : BOOLEAN;

VAR     t : BOOLEAN;

BEGIN   t := FALSE;
        WHILE (b <> NIL) AND (NOT t) (* short circuit *) DO
          t := t OR iBTptr(b^.bt);
          b := b^.nx;
          END;
        RETURN t;
END iBEptr;

PROCEDURE  iEXptr( e : EXptr) : StrPtr;

VAR     s : StrPtr;

BEGIN   WITH e^ DO
          IF rk = idenXP THEN
            s := LookUp(params,pt);
            IF s = Null THEN s := LookUp(vars,pt) END;
            RETURN s;
            END;
          IF rk = litrXP THEN RETURN pt END;
          Assert(FALSE,"Unknown expression type")
          END;
END iEXptr;

PROCEDURE  SetParams( pars : StrSeq; args : ExprSeq);

(* The arguments must be evaluated in the calling environment.
   So, we build a list of the args, then assign them. *)

VAR     s, t : StrSeq;

BEGIN   s := NIL;
        WHILE args <> NIL DO
          IF s = NIL 
            THEN NEW(s); t := s;
            ELSE NEW(t^.nx); t := t^.nx;
            END;
          t^.st := iEXptr(args^.ex);
          args := args^.nx;
          END;
        Mark(params);
        WHILE pars <> NIL DO
          IF s <> NIL THEN
            Enter(params,pars^.st,s^.st);
            t := s; s := s^.nx;
            DISPOSE(t);
            END;
          pars := pars^.nx;
          END;
END SetParams;

PROCEDURE  DoCall( pn : StrPtr; ps : ExprSeq);

VAR     p : ProcList;

BEGIN   p := Procs;
        WHILE (p <> NIL) AND (p^.nm <> pn) DO
          p := p^.nx END;
        Assert(p <> NIL,"Procedure missing");
        SetParams(p^.pl,ps);
        iStatSeq(p^.ss);
        Release(params);
END DoCall;

PROCEDURE  DoComm( cs : StrPtr);

(* We must locate and substitute any embedded variables before
   submitting the command to the system. 

   To fix:  the embedded idents should be checked in the
   compiler, not here.
*)

CONST   backtick = "`";

VAR     i, j, k : CARDINAL;
        cl, cv : StrType;
        v, s : StrPtr;

BEGIN   i := 0;
        cl[0] := 0C; Insert(cs^,cl,0);
        LOOP IF (i > HIGH(cl)) OR (cl[i] = 0C) THEN EXIT END;
             j := 1;
             IF cl[i] = backtick THEN
               j := i + 1;
               LOOP IF (j > HIGH(cl)) OR (cl[j] = 0C) THEN
                      Error("Substitution error in",cl);
                      END;
                    IF cl[j] = backtick THEN EXIT END;
                    INC(j);
                    END;
               (* now we have both ticks *)
               Substring(cl,cv,i+1,j-i-1);
               Remove(cl,i,j-i+1);
               v := StrTable.Enter(cv);
               s := LookUp(params,v);
               IF s = Null THEN s := LookUp(vars,v) END;
               j := Length(s^);
               Insert(s^,cl,i);
               END;
             INC(i,j);
             END;
        Execute(cl,(*echo*) LookUp(vars,QuietFlag)=Null,
                   (* do *) LookUp(vars,TestFlag)=Null);
END DoComm;

PROCEDURE  DoGuard( mi, mo : StrSeq; ss : StatSeq);

(* First, find the youngest input (yi) and the oldest output (oo).
   If oo is older than yi, then execute ss.  If the statements were
   executed, then set all the outputs to current time.
*)

(* NB:  This routine seems to consume a lot of CPU time - think
   about improvements here and in FileTimes.
*)

(* The success of each step is evaluated by checking for the existence
   of a new product.  This check is done by deleting the first product
   beforehand, and looking for it afterward.  Both operations are
   skipped if Test is not false (Null).  This should give reasonable
   behaviour on systems without the ability to check the status of
   spawned commands.
*)

CONST   crash = "Expected product not present";

VAR     i, o : StrSeq;
        yi, oo : StrPtr;

BEGIN   i := mi^.nx; (* there is at least one ! *)
        yi := mi^.st;
        WHILE i <> NIL DO
          IF CompareTimes(yi,i^.st) THEN yi := i^.st;
            END;
          i := i^.nx;
          END;
        o := mo^.nx;
        oo := mo^.st;
        WHILE o <> NIL DO
          IF CompareTimes(o^.st,oo) THEN oo := o^.st END;
          o := o^.nx;
          END;
        (* yi, oo now set *)
        IF CompareTimes(oo,yi) THEN (* needs rebuilding *)
          DidSomething := TRUE;
          iStatSeq(ss);
          o := mo;            (* now update file times *)
          WHILE o <> NIL DO
            MarkCurrent(o^.st);
            o := o^.nx;
            END;
          END;
END DoGuard;

PROCEDURE  DoAssign( vl : StrPtr; so : SopKind; vr, vx : EXptr);

(* This procedure implements a minor (and experimental) extension
   to the description language.  An assignment may optionally
   contain a string operation.  (These are not valid anywhere else.)
*)

VAR     s, t : StrPtr;
        r : StrType;
        i, j : CARDINAL;
   
BEGIN   s := iEXptr(vr);
        IF so <> noSO THEN
          t := iEXptr(vx);
          CASE so OF
             noSO  :  (* impossible *)
           | conSO : Concatenate(s^,t^,r);  (* concatenate strings *)
           | remSO : j := Length(t^);
                     i := Find(t^,s^);  (* look for t^ in s^ *)
                     r[0] := 0C; Insert(s^,r,0);
                     IF i <= HIGH(s^) THEN
                       Remove(r,i,j);  (* remove substring *)
                       END;
             ELSE Assert(FALSE,"Unknown string operation");
             END;
          s := StrTable.Enter(r);
          END;
        Enter(vars,vl,s);
END DoAssign;

PROCEDURE  iStatSeq( prog : StatSeq);

BEGIN   WHILE prog <> NIL DO
          WITH prog^ DO
            CASE sk OF
              nullST   : 
            | commST   : DoComm(cs);
            | callST   : DoCall(pn,ps);
            | abortST  : Error("ABORT statement executed","");
            | assignST : DoAssign(vl,so,vr,vx);
            | guardST  : DoGuard(mi,mo,ss);
            | condST   : IF iBEptr(bx)
                           THEN iStatSeq(tc)
                           ELSE iStatSeq(ec)
                           END;
              ELSE Assert(FALSE,"Unknown statement type");
              END;
            prog := nx;
            END;
          END;
END iStatSeq;

PROCEDURE  Interpret( prog : StatSeq);

(* Interpret the program pointed to by prog. *)

BEGIN   iStatSeq(prog);
        IF NOT DidSomething THEN
          Error("System is up to date","");
          END;
END Interpret;  

BEGIN   InitSymTab(params);
        InitSymTab(vars);
        TestFlag := StrTable.Enter("Test");
        QuietFlag := StrTable.Enter("Quiet");
        DidSomething := FALSE;
END Interpreter.
