IMPLEMENTATION MODULE TextIO;  (* Simple text file I/O *)

(* G.R. Guenther 1988 05 28  *)
(* Modified: GRG 1988 06 07  Added ReadLn *)
(*           GRG 1989 01 03  LUMOS-2 version *)

(* This module is similar to N. Wirth's InOut module.  It
   provides simple text-file I/O with support for file inclusion
   on input.  It hides the system interface and performs some
   simple file-name manipulation.  It also supports syntax error
   reporting.

   Reading or writing when no file is open will cause an assertion 
   to fail - "File not open when ...".
*)

FROM R2Streams IMPORT Stream, OpenStream, CloseStream, CreateStream,
                      ReadLine, WriteChar, EndStream;
IMPORT R2Streams;
FROM Errors IMPORT Error, Assert;
FROM R2Strings IMPORT Length, Copy, Remove, Concatenate, Insert;
FROM R2SysCalls IMPORT WriteString, Write, WriteLn, Translate, Status;
FROM R2NumbIO IMPORT WriteCard;
FROM R2Storage IMPORT ALLOCATE, DEALLOCATE;

CONST   fnl = 79;
        ll = 133;

TYPE    NameStr = ARRAY [0..fnl] OF CHAR;
        Line = ARRAY [0..ll] OF CHAR; 
        FileList = POINTER TO FCB;
        FCB = RECORD f : Stream;
                     name : NameStr;
                     text : Line;
                     curpos, lineno : CARDINAL;
                     eof, eoln, putbak : BOOLEAN;
                     next : FileList;
              END;

VAR     Files : FileList; (* Stack of open files *)

(* VAR     Eof, Done : BOOLEAN;  

   Eof reports whether the next GetChar operation would read
   beyond the end of the input.  Done gets the result of OpenInput.
*)

PROCEDURE  OpenInput( VAR filename, filetype, errortext : ARRAY OF CHAR;
                      forcetype, fatal : BOOLEAN );

(* If filename does not contain a file type, or forcetype is true,
   filename is edited to include the given type.  An attempt is made
   to open the file for input.  If the file cannot be opened, Done
   is set to FALSE.  If fatal is TRUE and Done is false an error message 
   containing errortext and the attempted filename is generated.
   If the open worked, subsequent calls to GetChar will read from
   this file.  If a file was already open, it is temporarily 
   hidden until this file is closed.  (Input files are stacked.)
*)
    
VAR     t : FileList;
        fn : ARRAY [0..133] OF CHAR;

BEGIN   NEW(t);  
        fn[0] := 0C; Insert(filename,fn,0);
        Done := FALSE;
        WITH t^ DO
          next := Files;
          IF Translate(fn,name) <> Ok THEN 
            IF fatal THEN Error(errortext,fn) END;
            DISPOSE(t);
            RETURN
            END;
          SetType(name,filetype,forcetype);
          OpenStream(f,name,FALSE);
          Done := R2Streams.Done;
          IF NOT Done THEN
            IF fatal THEN Error(errortext,name) END;
            DISPOSE(t);
            RETURN;
            END;
          eof := FALSE;
          Eof := FALSE;
          eoln := TRUE;
          putbak := FALSE;
          lineno := 0;
          curpos := 0;
          END;
        Files := t;
END OpenInput;

PROCEDURE  GetChar() : CHAR;

(* Providing that Eof is not TRUE, this function returns the
   next character in the input file.  Line boundaries are returned
   as blanks.  Calling GetChar when Eof is true also returns a blank.
*)

BEGIN   Assert(Files <> NIL,"No input file for GetChar");
        WITH Files^ DO
          IF NOT putbak THEN
            IF eoln
              THEN eof := EndStream(f);
                   IF NOT eof 
                     THEN ReadLine(f,text);
                          INC(lineno);
                     ELSE text[0] := 0C;
                     END;
                   curpos := 0;
                   Eof := eof;
                   eoln := FALSE;
              ELSE INC(curpos)
              END;
            END;
          putbak := eof;
          IF text[curpos] = 0C 
            THEN eoln := TRUE;
                 RETURN " ";
            ELSE RETURN text[curpos]
            END;
          END;
END GetChar;

PROCEDURE  PutBack;

(* Forces the single character just read back into the input stream. 
*)

BEGIN   Assert(Files <> NIL,"No input file for PutBack");
        Assert(Files^.lineno > 0,"Nothing to PutBack");
        Files^.putbak := TRUE;
END PutBack;

PROCEDURE  ErrorDisplay;

(* The contents of the current input line are written out.  The display
   is labelled with the name of the input file, the line number within
   that file and the current input position.  This is for identifying
   syntax errors.  Calling this procedure does not terminate a program.
*)

VAR     i, b : CARDINAL;

BEGIN   Assert(Files <> NIL,"Input file not open for error display");
        WITH Files^ DO
          Assert(lineno > 0,"No current line for error display");
          WriteString("File: ");  WriteString(name);
          IF eof THEN
            WriteString("  (at end of file.)"); WriteLn;
            RETURN;
            END;
          WriteString("  Line: "); WriteCard(lineno,1); WriteLn;
          WriteString(text); WriteLn;
          b := 1; 
          IF putbak THEN b := 2 END;  (* backup pointer *)
          FOR i := b TO curpos DO Write(" ") END;
          Write("^"); WriteLn;
          END;
END ErrorDisplay;

PROCEDURE  ReadLn( VAR s : ARRAY OF CHAR);

(* Reads an entire line into s.  Does not honour PutBack or maintain
   the error buffer.
*)

BEGIN   Assert(Files <> NIL,"No input file for ReadLn");
        WITH Files^ DO
          eof := EndStream(f);
          IF NOT eof
            THEN ReadLine(f,s);
            ELSE s[0] := 0C
            END;
          Eof := eof;
          END;
END ReadLn;

PROCEDURE  CloseInput;

(* If there is an input file open, this will close it and return
   to the previously open input file.  It is not an error to call
   CloseInput when no files are open.
*)
                     
VAR     t : FileList;

BEGIN   t := Files;
        IF t = NIL THEN RETURN END;
        WITH t^ DO
          Files := next;
          CloseStream(f);
          END;
        DISPOSE(t);
        IF Files <> NIL THEN Eof := Files^.eof END;
END CloseInput;

(* Output file section *)

VAR     Ofile : Stream;
        Oopen : BOOLEAN;
        ColPos : CARDINAL;  (* Where next char goes *)
        Oname : NameStr;

PROCEDURE  CreateOutput( VAR filename, filetype, errortext : ARRAY OF CHAR);

(* The filename parameter is edited to include the given filetype.
   An attempt is made to open the file; if it fails, the generated
   filename and the errortext are used to signal a fatal error.
   Otherwise, all subsequent output will go to this file.  (Output
   files are not stacked.)
*)     
            
VAR     fn : NameStr;

BEGIN   fn[0] := 0C; Insert(filename,fn,0);
        Assert(NOT Oopen,"Output file already open");
        IF Translate(fn,Oname) <> Ok THEN Error(errortext,fn) END;
        SetType(Oname,filetype,TRUE);
        CreateStream(Ofile,Oname);
        IF NOT R2Streams.Done THEN Error(errortext,Oname) END;
        Oopen := TRUE;
        ColPos := 1;
END CreateOutput;

PROCEDURE  PutChar( ch : CHAR);    

(* Writes a single character on the output file *)

BEGIN   Assert(Oopen,"Output file not open when writing");
        WriteChar(Ofile,ch);
        INC(ColPos);
END PutChar;

PROCEDURE  PutString( VAR s : ARRAY OF CHAR; margin, indent : CARDINAL);
 
(* PutString writes a string on the current output file. If it
   would push the output line beyond the margin a new line 
   is started first, and indented by the stated amount.  This
   feature is disabled if margin = 0.
*)
       
VAR     i, j : CARDINAL;

BEGIN   j := Length(s);
        IF (ColPos + j > margin) AND (margin > 0) THEN
          PutLn;
          FOR i := 1 TO indent DO PutChar(' ') END;
          END;
        FOR i := 1 TO j DO PutChar(s[i-1]) END;
END PutString;

PROCEDURE  PutLn;

(* Writes an end-of-line mark in the file.
*)

BEGIN   PutChar(15C); 
        ColPos := 1;
END PutLn;
 
PROCEDURE  CloseOutput;

(* The output file is closed.  It is not an error to call this
   procedure when no output file is open.
*)

BEGIN   IF Oopen THEN 
          IF ColPos <> 1 THEN PutLn END;
          CloseStream(Ofile);
          Oopen := FALSE;
          END;
END CloseOutput;

PROCEDURE  SetType( VAR nam, typ : ARRAY OF CHAR; force : BOOLEAN);

(* If force is true, any extension (.XXX) is stripped off nam.  If
nam has no extension, typ is added. *)

VAR     i, k : CARDINAL;
        dotseen : BOOLEAN;

BEGIN   k := Length(nam);
        IF k = 0 THEN RETURN END;
        i := k-1;
        LOOP IF nam[i] = '.' THEN
               dotseen := TRUE;
               EXIT;
               END;
             IF (nam[i] = '-') OR (i = 0) THEN EXIT END;
             DEC(i);
             END;
        IF force AND dotseen THEN 
          Remove(nam,i,k);
          dotseen := FALSE;
          END;
        IF NOT dotseen THEN Concatenate(nam,typ,nam) END;
END SetType;

BEGIN   Oopen := FALSE;
        Files := NIL;
END TextIO.
