IMPLEMENTATION MODULE TopoSort;  (* Topologically sorts relations *)

(* G.R. Guenther 1988 05 28  *)
(* Modified: GRG 1989 01 03  LUMOS-2 version *)

(* This module supports an abstract notion of a partially ordered
   set.  The elements of the underlying set are arbitrary addresses.
   (Elements are the same if the addresses are equal as numbers.)
   The relation is initially empty.  Ordered pairs may be Inserted
   into the relation at any time. Each time it is called, GetNode
   returns a minimal element (according to the partial order) and
   removes that element from the relation (along with all related
   pairs).

   As an efficiency hack, the Insert procedure also accepts an
   additional attribute to associate to the first member of the pair.
   The most recent value assigned is returned by GetNode.
         
   If TopoSort detects a cycle in the relation, the "Cycle in dependencies" 
   fatal error is issued. The "Out of memory" error can also occur if too 
   large a relation is being sorted.
*)        

FROM SYSTEM IMPORT ADDRESS;
FROM R2Storage IMPORT ALLOCATE, DEALLOCATE;
FROM Errors IMPORT Error;

TYPE    NPtr = POINTER TO Node;
        LPtr = POINTER TO Link;
        Link = RECORD e : NPtr;
                      n : LPtr;
               END;
        Node = RECORD elt, attr : ADDRESS;
                      ref : CARDINAL;
                      edges : LPtr;
                      next : NPtr;
               END;

VAR     Head : NPtr;
        Count : CARDINAL;
          
PROCEDURE  Newnode( VAR p : NPtr; el : ADDRESS);

(* helper for Insert *)

BEGIN   NEW(p);
        WITH p^ DO
          elt := el;
          attr := NIL;
          ref := 0;
          edges := NIL;
          next := Head;
          END;        
        Head := p;
        INC(Count);
END Newnode;

PROCEDURE  Insert( lesser, greater, lattr : ADDRESS);

(* Adds the ordered pair to the relation. *)

VAR     p, l, g : NPtr;
        e : LPtr;

BEGIN   l := NIL;  g := NIL;
        p := Head;
        WHILE p <> NIL DO
          IF p^.elt = lesser THEN l := p END;
          IF p^.elt = greater THEN g := p END;
          p := p^.next;
          END;
        IF l = NIL THEN Newnode(l,lesser) END;
        IF g = NIL THEN Newnode(g,greater) END;
        l^.attr := lattr;
        e := l^.edges;
        WHILE e <> NIL DO
          IF e^.e = g THEN RETURN (* edge is here already *) END;
          e := e^.n;
          END;
        NEW(e);
        e^.e := g;
        e^.n := l^.edges;
        INC(g^.ref);
        l^.edges := e;    (* new edge is in  *)
END Insert;
        
PROCEDURE  GetNode( VAR node, attr : ADDRESS);

(* node is NIL when no more nodes are available. *) 

VAR     q, r : NPtr;
        k, d : LPtr;

BEGIN   r := Head;  q := NIL;
        WHILE (r <> NIL) AND (r^.ref <> 0) DO 
          q := r;
          r := r^.next;
          END;
        IF r = NIL THEN
          IF Count > 0 THEN Error("Cycle in dependencies","") END;
          node := NIL;
          RETURN;
          END;
        k := r^.edges;
        WHILE k <> NIL DO
          DEC(k^.e^.ref);
          d := k;
          k := k^.n;
          DISPOSE(d);
          END;
        node := r^.elt;
        attr := r^.attr;
        DEC(Count);
        IF q = NIL 
          THEN Head := r^.next 
          ELSE q^.next := r^.next 
          END;
        DISPOSE(r);
END GetNode;

BEGIN   Head := NIL;
        Count := 0;
END TopoSort.
