IMPLEMENTATION MODULE UMode;  (* Update file generation mode of Uses *)

(* G.R. Guenther 1988 05 29  *)

(* This module contains the procedures to generate an Update description
   file for a suite of modules.  The required steps are outlined
   in section 4 of the Users' Manual for Uses:

#START
     4. Principles of Operation - U mode

     When Uses is called on to construct an Update description file,
     it performs the same basic analysis as described above.  In this
     mode, however, definition and implementation modules are
     considered as distinct modules.  The clauses written to the .UPD
     file take three different forms.

     A definition module generates a MODULE clause for its symbol file
     in which the USES part contains the .DEF file itself, and the
     .SYM files of any imports. For example, if Earl.DEF imports from
     SysCalls the following MODULE clause will be generated:

          MODULE "Earl.SYM" USES "Earl.DEF", "SysCalls.SYM"
            BY Build("Earl.DEF") END;

     An implementation module, similarly, generates a clause for its
     .OBJ file.  For Earl.MOD this would be:

          MODULE "Earl.OBJ" USES "Earl.MOD", "Earl.SYM",
            "SysCalls.SYM" BY Build("Earl.MOD") END;

     A program module has no corresponding definition module, so does
     not implicitly import its own .SYM file.  Each program module
     corresponds to an executable file, however.  King.MOD would
     produce the following two clauses:

          MODULE "King.OBJ" USES "King.MOD", "Queen.SYM",
            "Duke.SYM", "Earl.SYM" BY Build("King.MOD") END;
          MODULE "King.EXE" USES "King.OBJ", "Queen.OBJ",
            "Duke.OBJ", "Earl.OBJ", "SysCalls.OBJ"
            BY Build("King.OBJ") END;
#STOP
*)       

FROM UsesLex IMPORT IdType, IdPtr, LexText;
FROM StrTable IMPORT Enter;
FROM UsesParse IMPORT Parse;
FROM IdLists IMPORT IdList, IdPut, IdGet, IdInit;
FROM TextIO IMPORT PutString, PutChar, PutLn, CreateOutput, CloseOutput,
                   OpenInput, CloseInput, Done;
FROM Commands IMPORT GetArg;
FROM R2Strings IMPORT Concatenate;
FROM Errors IMPORT Error, Assert;
FROM Extensions IMPORT DEFext, MODext, SYMext, LNKext, LODext;

CONST   Margin = 72;       (* output format parameters *)
        Indent = 4;                                     

VAR     Include, Builder : IdType;  (* The command line arguments *)
        ModList, LnkList : IdList;  (* LnkList is used to build the
                                       MODULE clause for the .LOD file *)
        SYSTEMp : IdPtr;

PROCEDURE  Quote( VAR s1, s2 : ARRAY OF CHAR);   (* print helper *)

(* The two strings are concatenated, surrounded by quotes and passed
   to PutString with the standard margin and indent settings *)

VAR     s, sq : IdType;

BEGIN   sq[0] := '"'; sq[1] := 0C;
        Concatenate(sq,s1,s);
        Concatenate(s,s2,s);
        Concatenate(s,sq,s);
        PutString(s,Margin,Indent);
END Quote;
 
PROCEDURE  ProcessLod;

(* This procedure writes out the MODULE clause for root.LOD
*)     

VAR     z : BOOLEAN;
        mod, main : IdPtr;
        i : CARDINAL;

BEGIN   PutString("MODULE ",0,0);
        Assert(IdGet(LnkList,mod,z),"Empty Link file list");
        main := mod;
        Quote(mod^,LODext);
        PutString(" USES ",0,0);
        Quote(mod^,LNKext);
        WHILE IdGet(LnkList,mod,z) DO
          PutString(", ",0,0);
          Quote(mod^,LNKext);
          END;
        PutLn;
        FOR i := 1 TO Indent DO PutChar(' ') END;
        PutString("BY ",0,0);
        PutString(Builder,0,0);
        PutChar('('); Quote(main^,LNKext); PutChar(')');
        PutString(" END.",0,0); 
        PutLn;
END ProcessLod;

PROCEDURE  ProcessModule( VAR mod, sext, pext : ARRAY OF CHAR;
                          implem : BOOLEAN) : BOOLEAN;

(* This procedure generates the MODULE clause for mod.pext in
   terms of mod.sext and any imports determined by parsing
   the module. All imported modules are put onto the ModList.  

   When an implementation module is compiled to a link file,
   the link file depends on the symbol file of the corresponding
   definition module.  implem signals this extra dependence.
*)

VAR     z : BOOLEAN;
        mn : IdPtr;
        i : CARDINAL;

BEGIN   OpenInput(mod,sext,"",FALSE,FALSE);
        IF NOT Done THEN RETURN FALSE END;  (* No file *)
        z := Parse();
        mn := LexText;
        IdPut(LnkList,mn,z);
        PutString("MODULE ",0,0);
        Quote(mn^,pext);
        PutString(" USES ",0,0);
        Quote(mn^,sext);
        IF implem THEN
          PutString(", ",0,0);
          Quote(mn^,SYMext);
          END;
        WHILE Parse() DO
          IF LexText <> SYSTEMp THEN
            PutString(", ",0,0);
            Quote(LexText^,SYMext);
            IdPut(ModList,LexText,z);
            END;
          END;  
        PutLn;  
        FOR i := 1 TO Indent DO PutChar(' ') END;
        PutString("BY ",0,0);
        PutString(Builder,0,0);
        PutChar('('); Quote(mn^,sext); PutChar(')');
        PutString(" END;",0,0);
        PutLn;
        CloseInput;
        RETURN TRUE;
END ProcessModule;

PROCEDURE  ModeU( root : IdType);

(* ModeU gets the command line arguments, initialises the lists,
   creates the output file, writes the INCLUDE directive and 
   sequences the calls to ProcessModule and ProcessLod.
*)

VAR     junk : ARRAY [0..2] OF CHAR;
        mod : IdPtr;
        z : BOOLEAN;

BEGIN   Builder := "";
        GetArg("Build procedure",Builder,junk,(*reqd*)FALSE,(*val*)FALSE);
        IF Builder[0] = 0C
          THEN Builder := "Build";
               Include := "$Build";
          ELSE Concatenate(Builder,".UPD",Include);
          END;
        GetArg("Include file",Include,junk,FALSE,FALSE);
        IdInit(ModList);
        IdInit(LnkList);
        SYSTEMp := Enter("SYSTEM");
        CreateOutput(root,".UPD","Cannot create output file");
        PutString("INCLUDE ",0,0);
        Quote(Include,"");
        PutChar(';'); PutLn;
        IF NOT ProcessModule(root,MODext,LNKext,(*implem*)FALSE)
          THEN Error("File not found",root) END;
        WHILE IdGet(ModList,mod,z) DO
          z := ProcessModule(mod^,DEFext,SYMext,(*implem*)FALSE);
          z := ProcessModule(mod^,MODext,LNKext,(*implem*)TRUE);
          END;
        ProcessLod;
        CloseOutput;
END ModeU;

END UMode.
