MODULE Update;  (* Rebuilding tool for Modula-2 programs *)
                             
(* G.R. Guenther 1988 05 27 *)

(* This module obtains the command line, opens the update description
   file and invokes the compiler and interpreter.

   From the users' manual for Update:
#START
                       Users' Manual -- Update

     1. Introduction

     This manual describes the Update program.  Update regenerates
     programs when changes are made to their components.  Regeneration
     is directed by a description file.  Update does no more work than
     is required.  You should be familiar with Modula-2 programming
     and the programming conventions of LUMOS-2 before reading this
     manual.  The description files used by Update are often produced
     by the Uses tool.  You should read the manual for Uses along with
     this manual.

     2. Invoking Update

     Update may be started with a command of the form:

          Update <descriptor> [<ident>[="<string>"]] ...

     <descriptor>  is the name of an Update description file.  If you
                   do not specify a file type, .UPD will be assumed.
                   Update will use this file to direct the rebuilding
                   of the software system that it describes. If you do
                   not provide this argument, Update will prompt you
                   for your command. (The syntax of Update description
                   files is described below.)

     <ident>       is an identifier. (As in Modula-2, it must begin
                   with a letter and contain only letters or digits.)
                   Declaring an identifier on the command line gives
                   it a known value as a variable in the Update
                   description.  The value may be any character
                   string.  If a value is not given explicitly on the
                   command line, the value "T" is used.  These
                   variable assignments may be used to choose between
                   variations in the rebuilding method.

     <string>      is any string of characters that will be taken as
                   the value of the preceding identifier.

     Examples:

          Update King

               Will examine King.UPD and, if necessary, perform some
               or all of the rebuilding operations described there.

          Update MyOwnOS map install="C:"

               This command follows the rebuilding instructions in
               MyOwnOS.UPD with the (Update) variable 'map' set to "T"
               and the variable 'install' set to "C:".

#STOP
#INCLUDE Compiler.MOD
#START

#STOP
#INCLUDE Interpre.MOD
#START

     6. Error Messages

     Update detects many errors.  In each case, Update's operation is
     terminated.  If the rebuilding operation has been partially
     completed, no attempt is made to remove the intermediate
     products.  The following list contains the messages that are
     printed when an error occurs.

     ABORT statement executed

          The Update description file contains an ABORT statement.
          It has just been executed.  If you did not write the
          description file, look at the file for a possible
          explanation.

     Cannot open description file

          The description file named on the command line does not
          exist, or is protected.

     Cannot open INCLUDE file: ...

          A file named in an INCLUDE directive cannot be found or is
          protected.

     Command line syntax error

          If a "=" occurs on the command line, it must be followed by
          a quoted string.  If the string or a quote is missing you
          will receive this error.

     Cycle in dependencies

          The description file describes a system that cannot be
          constructed because of circular dependencies.  You must
          correct the file.

     Expected product not present: ...

          If the execution of a manufacturing step did not produce
          the expected output, this message will appear.  You should
          read the error messages produced by the rebuilding command.

     I/O error

          This message indicates that an unexpected error occurred
          while reading a file.  Check for hardware problems.

     Internal error: ...

          Update has detected an inconsistency in its own operation.
          Please report this to your system manager, giving as much
          detail as possible. 

     Invalid identifier syntax: ...

          You have attempted to set a variable on the command line.
          The variable name must begin with a letter and contain
          only letters and digits.

     Out of memory

          If a system is very large, Update may not be able to obtain
          enough memory to hold its tables.  There is no effective
          remedy for this problem.  This error can also occur if
          infinite recursion is attempted.  Check for recursive
          procedures in the description file.

     More than one way to rebuild: ...

          The named component appears as a product in more than one
          manufacturing step.

     Procedure multiply defined: ...

          The Update description file defines a procedure more
          than once.  Repair the file.

     Procedure not declared: ...
 
          The Update description file does not contain a declaration
          for the named procedure.  Repair the file.

     Stopped by user

          You pressed the system break key.

     Substitution error in: ...

          While performing variable substitution on a command,
          Update has encountered mismatched back ticks.  Repair
          the description file.
                                 
     System is up to date

          Update has determined that there is nothing for it to do.

     Syntax error: ...

          This message will contain a description of the error and the
          offending text.  Repair the description file and try again.
          (Syntax errors are detected before any statements are
          executed.)
#STOP
*)

FROM Commands IMPORT GetArg;
FROM Errors IMPORT SetName, Error;
FROM UpdLex IMPORT StrType, ChkIdSyntax;
FROM Interpreter IMPORT PreAssign, Interpret, StatSeq;
FROM Compiler IMPORT Parse;
FROM TextIO IMPORT OpenInput, CloseInput;

CONST   defext = ".UPD";

VAR     descrip, var, val : StrType;
        ss : StatSeq;

BEGIN   SetName("Update");
        GetArg("Description file",descrip,var,(*reqd*)TRUE,(*val*)FALSE);   
        LOOP  var := "";
              val[0] := "T"; val[1] := 0C;
              GetArg("",var,val,FALSE,TRUE);
              IF var[0] = 0C THEN EXIT END;   (* null argument *)
              IF NOT ChkIdSyntax(var) THEN 
                Error("Command line syntax error",var);
                END;
              PreAssign(var,val);
              END;                     
        OpenInput(descrip,defext,"Cannot open description file",
                  (*forcetype*)FALSE,(*fatal*)TRUE);
        ss := Parse();
        CloseInput;
        Interpret(ss);
END Update.
