IMPLEMENTATION MODULE UpdLex;  (* Lexical analyser for Update *)

(* G.R. Guenther 1988 05 30  *)
(* Modified: GRG 1988 06 01  Added plusSY and minusSY for string ops *)

(* This module is the lexical analyser for the Update program.  It
   recognises the lexical rules for the Update description language.

   Communication with this module is through the global variables
   LexSym and LexText.  They are set every time Lex is called.

   The lexer keeps only one copy of each identifier, so if
   it returns two equal pointers in different contexts, you
   may assume that the identifiers were equal without comparing
   the actual strings.

   Be cautious when using StrPtrs.  Do not assume that all the
   MaxIdLen characters have actually been allocated.  Never
   write on LexText^.

EXPORT QUALIFIED MaxStrLen, StrType, StrPtr, Symbol, LexText, 
                 ChkIdSyntax, LexSym, Lex;

CONST   MaxStrLen = 133;  (* Strings cannot cross line boundaries ... *)

TYPE    Symbol = ( semiSY,   (* ;              *)
                   dotSY,    (* .              *)
                   commaSY,  (* ,              *)
                   lparenSY, (* (              *)
                   rparenSY, (* )              *)
                   getsSY,   (* :=             *)
                   eqSY,     (* =              *)
                   neqSY,    (* # (also <>)    *)
                   subSY,    (* <=             *)
                   plusSY,   (* +              *)
                   minusSY,  (* -              *)
                   abortSY,  (* ABORT          *)
                   andSY,    (* AND            *)
                   beginSY,  (* BEGIN          *)
                   bySY,     (* BY             *)
                   elseSY,   (* ELSE           *)
                   endSY,    (* END            *)
                   ifSY,     (* IF             *)
                   includSY, (* INCLUDE        *)
                   moduleSY, (* MODULE         *)
                   notSY,    (* NOT            *)
                   orSY,     (* OR             *)
                   procSY,   (* PROCEDURE      *)
                   thenSY,   (* THEN           *)
                   usesSY,   (* USES           *)
                   identSY,  (* identifier     *) 
                   stringSY, (* literal string *)
                   comandSY, (* command string *) (* substs still in *)
                   otherSY   (* none of these  *) );

        StrType = ARRAY [0..MaxStrLen] OF CHAR;
        StrPtr  = POINTER TO StrType;

VAR     LexText : StrPtr;
        LexSym  : Symbol;    
*)

FROM TextIO IMPORT Eof, GetChar, PutBack;
FROM StrTable IMPORT Enter;

VAR     ifID, beginID, orID, thenID, elseID, abortID, notID,
        endID, andID, byID, includID, moduleID, procID, usesID : StrPtr;

PROCEDURE  IsLetter( c : CHAR) : BOOLEAN;

BEGIN   c := CAP(c);
        RETURN ('A' <= c) AND (c <= 'Z');
END IsLetter;

PROCEDURE  IsIdentChar( c : CHAR): BOOLEAN;

BEGIN   RETURN IsLetter(c) OR (('0' <= c) AND (c <= '9'));
END IsIdentChar;

PROCEDURE ChkIdSyntax( VAR str : ARRAY OF CHAR) : BOOLEAN;

(* Determines whether the string obeys the lexical rules for
   an identifier.
*)

VAR     i : CARDINAL;

BEGIN   IF NOT IsLetter(str[0]) THEN RETURN FALSE END;
        i := 1;
        WHILE str[i] <> 0C DO
          IF NOT IsIdentChar(str[i]) THEN RETURN FALSE END;
          INC(i);
          END;
        RETURN TRUE;
END ChkIdSyntax;

PROCEDURE  Lex;

(* Lex skips blanks, tabs and comments to find the next token in the
   current input file (in TextIO).  The type of the token is stored
   in LexSym. If LexSym is identSY, stringSY or comandSY then the
   corresponding text is stored in LexText^.
*)

VAR     text : StrType;
        ch, qc : CHAR;
        cd, i : CARDINAL;

BEGIN   LexSym := otherSY;
        cd := 0; (* comment nesting depth *)
        LOOP IF Eof THEN RETURN END;
             ch := GetChar();
             IF ch = '(' THEN
               ch := GetChar();
               IF ch = '*' 
                 THEN INC(cd)
                 ELSE IF cd = 0 THEN 
                        PutBack; 
                        LexSym := lparenSY;
                        RETURN;
                        END;
                 END;
             ELSIF ch = '*' THEN
               IF cd = 0 THEN  (* command line *)
                 i := 0;
                 LOOP text[i] := GetChar();
                      IF text[i] = ';' THEN EXIT END;
                      IF i < MaxStrLen THEN INC(i) END;
                      END;
                 text[i] := 0C;
                 PutBack;
                 LexText := Enter(text);
                 LexSym := comandSY;
                 RETURN;
                 END;
               ch := GetChar();
               IF ch = ')' THEN DEC(cd) ELSE PutBack END;
             ELSIF (cd = 0) AND ((ch = '"') OR (ch = "'")) THEN
               qc := ch;
               i := 0;
               LOOP text[i] := GetChar();
                    IF text[i] = qc THEN EXIT END;
                    IF i < MaxStrLen THEN INC(i) END;
                    END;
               text[i] := 0C;
               LexText := Enter(text);
               LexSym := stringSY;
               RETURN;
             ELSIF (ch > ' ') AND (cd = 0) THEN EXIT;
               END;
             END;
        IF NOT IsLetter(ch) THEN
          (* the following could be sped up marginally by using an
             array indexed by CHAR. *)
          IF ch = ';' THEN LexSym := semiSY; RETURN END;
          IF ch = '.' THEN LexSym := dotSY; RETURN END;
          IF ch = ',' THEN LexSym := commaSY; RETURN END;
          IF ch = ')' THEN LexSym := rparenSY; RETURN END;
          IF ch = '+' THEN LexSym := plusSY; RETURN END;
          IF ch = '-' THEN LexSym := minusSY; RETURN END;
          IF ch = ':' THEN 
            ch := GetChar();
            IF ch <> '=' THEN PutBack; RETURN END;
            LexSym := getsSY; 
            RETURN 
            END;
          IF ch = '=' THEN LexSym := eqSY; RETURN END;
          IF ch = '#' THEN LexSym := neqSY; RETURN END;
          IF ch = '<' THEN    
            ch := GetChar();
            IF ch = '=' THEN LexSym := subSY
              ELSIF ch = '>' THEN LexSym := neqSY
              ELSE PutBack
              END;
            RETURN 
            END;
        ELSE  (* IsLetter *)
          i := 0;
          REPEAT text[i] := ch;
                 IF i < MaxStrLen THEN INC(i) END;
                 ch := GetChar();
            UNTIL NOT IsIdentChar(ch);
          PutBack;
          text[i] := 0C;
          LexText := Enter(text);
          IF LexText = ifID THEN LexSym := ifSY; RETURN END;
          IF LexText = beginID THEN LexSym := beginSY; RETURN END;
          IF LexText = orID THEN LexSym := orSY; RETURN END;
          IF LexText = thenID THEN LexSym := thenSY; RETURN END;
          IF LexText = elseID THEN LexSym := elseSY; RETURN END;
          IF LexText = abortID THEN LexSym := abortSY; RETURN END;
          IF LexText = notID THEN LexSym := notSY; RETURN END;
          IF LexText = endID THEN LexSym := endSY; RETURN END;
          IF LexText = andID THEN LexSym := andSY; RETURN END;
          IF LexText = byID THEN LexSym := bySY; RETURN END;
          IF LexText = includID THEN LexSym := includSY; RETURN END;
          IF LexText = moduleID THEN LexSym := moduleSY; RETURN END;
          IF LexText = procID THEN LexSym := procSY; RETURN END;
          IF LexText = usesID THEN LexSym := usesSY; RETURN END;
          LexSym := identSY;
          RETURN;
          END;
END Lex;

BEGIN   ifID     := Enter("IF");
        beginID  := Enter("BEGIN");
        orID     := Enter("OR");
        thenID   := Enter("THEN");
        elseID   := Enter("ELSE");
        abortID  := Enter("ABORT");
        notID    := Enter("NOT");
        endID    := Enter("END");
        andID    := Enter("AND");
        byID     := Enter("BY");
        includID := Enter("INCLUDE");
        moduleID := Enter("MODULE");
        procID   := Enter("PROCEDURE");
        usesID   := Enter("USES");
END UpdLex.
