MODULE Uses;  (* Dependence analyser for Modula-2 programs *)
                             
(* G.R. Guenther 1988 05 26 *)    

(* This module performs the initial command line parsing.  It then
   invokes either ModeD or ModeU.
*)

(* From the users' manual for Uses:
#START
                       Users' manual -- Uses

     1. Introduction

     This manual describes the Uses program.  Uses analyses Modula-2
     programs and deduces the dependencies between modules.  You
     should be familiar with Modula-2 programming and the programming
     conventions of LUMOS-2 before reading this manual. The main
     application of Uses is to produce description files for the
     Update tool.  It is recommended that you read the documentation
     for Update along with this manual.

     2. Invoking Uses

     Uses may be started with a command of the form:

          Uses <root> [ <option> [ <builder> [ <include> ]]]

     <root>     is the name of a Modula-2 program module. If you do
                not specify a file type, .MOD will be assumed.  Uses
                will analyse this file and all modules that it
                explicitly or implicitly imports.  If you omit this
                filename altogether, Uses will prompt you for all the
                command arguments.

     <option>   is one of the two letters: D or U.  If you enter D,
                Uses will construct an ordered table of the module
                dependencies and write it on a file with a name
                derived from the <root> by replacing .MOD with .DEP.
                If you enter U, which is also the default value for
                this argument, Uses will produce a .UPD file in the
                format required by Update.

     <builder>  is the procedure name that will be written in the BY
                part of every MODULE clause generated by Uses.  This
                name is only used in the U mode.  The default value of
                this argument, which is used if nothing is entered
                here, is the name "Build".

     <include>  is a file name for which an INCLUDE directive will be
                generated in the .UPD file.  The default value is
                derived from the <builder> argument by suffixing .UPD.
                If the <builder> argument was not specified <include>
                is given the value "$Build".  (This is a logical name
                that usually refers to a file containing a general
                purpose Build procedure.) This argument, too, is used
                only in the U mode.

     Examples:

          Uses King

               Examines King.MOD and produces King.UPD for Update. The
               lines of King.UPD might look like:

               INCLUDE "$Build";
               MODULE "King.EXE" USES "King.OBJ", "SysCalls.OBJ"
                 BY Build("King.OBJ") END;

          Uses King D

               Examines King.MOD and writes out King.DEP containing
               the module dependencies of King.  The format of this
               file is explained below.

          Uses King.OLD U SysBuild MyBuild

               This command will try to find a program module in
               King.OLD.  It will generate an Update description file
               using "SysBuild" in the BY phrases and "MyBuild" in the
               INCLUDE directive.

#STOP
#INCLUDE dmode.mod
#START

#STOP
#INCLUDE umode.mod
#START

     5. Error messages

     Uses can stumble for several reasons.  In every case, the
     computation is abandoned and the output file is deleted.  The
     following list contains the messages that are printed when an
     error occurs, and the possible causes.


     Cannot create output file

          Uses is unable to create the output .DEP or .UPD file.
          Check that you have write permission in the current
          directory and that the volume has room for the file.

     Cycle in dependencies

          In the D mode, Uses may encounter circular dependencies in
          the modules.  This is bad programming practice, and causes
          Uses to fail.  (It can happen if an implementation module
          imports a module that itself imports the definition of the
          first module.) The program should be redesigned to avoid the
          circular dependencies.

     File not found: ...

          This message is printed if the <root> file cannot be found
          in the current directory.  It can also occur if an
          implementation module is missing.  Missing definition
          modules are assumed to be in a library and do not generate
          this error.

     I/O error

          Unexpected failures while reading or writing files can
          generate this error message.  Check for hardware problems.

     Internal error: ...

          Uses is confused. Report this error to your system manager,
          include as much information about the error as you can.

     Modula-2 syntax error in: ...

          Although Uses recognises only a small portion of the
          Modula-2 syntax, it can detect that some modules are
          syntactically invalid.  Use the Modula-2 compiler to
          identify the error and repair it.

     Out of memory

          Rarely, a system may be so large that Uses' internal tables
          exhaust available memory.  There is no convenient remedy for
          this problem.

     Stopped by user

          You pressed the system break key.

     Unrecognised option

          At present, Uses accepts only the D and U option letters.
#STOP
*)

FROM Commands IMPORT GetArg;
FROM UMode IMPORT ModeU;
FROM DMode IMPORT ModeD;
FROM UsesLex IMPORT IdType;
FROM Errors IMPORT SetName, Error;

VAR     root : IdType;
        opt, junk : ARRAY [0..1] OF CHAR;

BEGIN   SetName("Uses");
        GetArg("Root file",root,junk,(*reqd*)TRUE,(*val*)FALSE);   
        opt[0] := "U"; opt[1] := 0C;
        GetArg("Mode (U or D)",opt,junk,(*reqd*)FALSE,(*val*)FALSE);
        opt[0] := CAP(opt[0]);
        IF opt[0] = "U" THEN ModeU(root)
          ELSIF opt[0] = "D" THEN ModeD(root)
          ELSE Error("Unrecognised option",opt);
          END;
END Uses.
