MODULE Backup;  (* Backup utility for cartridge tape *)
                               
FROM SYSTEM IMPORT ADDRESS;
FROM R2SysCalls IMPORT GetCommand, Time, Handle, Status, ModeBits, FileMode,
                       Exit, Open, GetMode, ReadBlock, WriteLn, WriteString,
                       GetMessage, GetTime, FormatTrack, Create, Result,
                       ReadString, Translate, FastSize, Close, WriteBlock,
                       SetMode;
FROM R2NewCalls IMPORT SetFileDate;
FROM R2NumbIO IMPORT WriteCard;
FROM R2TimeSubs IMPORT CompareTime, FormatTime, ScanTime;
FROM R2Streams IMPORT Stream, CreateStream, WriteLine, CloseStream, Done;
FROM R2Format IMPORT FormatCard;
FROM R2Strings IMPORT Copy, Find, Length, Remove, Substring, Insert;

CONST   TapeBlockSize = 32256; (* Bytes, 63 Blocks *)

TYPE    Line = ARRAY [0..79] OF CHAR;
        DirRec = RECORD Name : ARRAY [0..19] OF CHAR;
                        First : CARDINAL;
                        Date : Time;
                        Owner : CARDINAL;
                        Mode : FileMode;
                        END;
        Block = RECORD 
                  CASE BOOLEAN OF
                    TRUE  : a : ARRAY [0..511] OF CHAR  |
                    FALSE : d : ARRAY [0..15] OF DirRec
                    END;
                  END;
        Header = RECORD Name : Line;
                        Size, Owner : CARDINAL;
                        Mode : FileMode;
                        CDate, BDate : Time;             (* 98 bytes *)
                 END;
        Hdr = RECORD 
                CASE BOOLEAN OF
                  TRUE  : a : ARRAY [0..97] OF CHAR |    
                  FALSE : h : Header
                  END;
                END;
        Trailer = RECORD
                    CASE BOOLEAN OF
                      TRUE  : a : ARRAY [0..3] OF CHAR |
                      FALSE : cs : ADDRESS
                      END;
                    END;

VAR     ComLine, Param, Root : Line;
        Hyphen : ARRAY [0..1] OF CHAR;
        Tape : Handle;
        TapeBlock : ARRAY [0..TapeBlockSize-1] OF CHAR;
        Tp : CARDINAL;
        CheckSum : ADDRESS;
        CutOff : Time;
        os : Stream;
        inter, log : BOOLEAN;

PROCEDURE  StartWriteTape;

BEGIN   Tp := 0;
        Check(Open("#C0",TRUE,Tape),'Opening tape','');
END StartWriteTape;

PROCEDURE WriteTape( VAR b : ARRAY OF CHAR);

VAR     i : CARDINAL;

BEGIN   FOR i := 0 TO HIGH(b) DO
          TapeBlock[Tp] := b[i];
          INC(Tp);
          (*$T-*) CheckSum := CheckSum + ADDRESS(ORD(b[i])) (*$T=*) ;
          IF Tp = TapeBlockSize THEN
            Check(WriteBlock(Tape,TapeBlock,FALSE),'Writing tape','');
            Tp := 0;
            END;
          END;
END WriteTape;

PROCEDURE  EndWriteTape;

BEGIN   IF Tp <> 0 THEN 
          Check(WriteBlock(Tape,TapeBlock,FALSE),'Writing tape','') END;
        Check(Close(Tape,FALSE),'Closing tape','');
END EndWriteTape;

PROCEDURE  StartReadTape;

BEGIN   Check(Open("#C0",TRUE,Tape),'Opening tape','');
        Tp := TapeBlockSize;
END StartReadTape;

PROCEDURE  ReadTape( VAR b : ARRAY OF CHAR);

VAR     i : CARDINAL;   (* EOF should not be encountered *)

BEGIN   FOR i := 0 TO HIGH(b) DO
          IF Tp = TapeBlockSize THEN
            Check(ReadBlock(Tape,TapeBlock),'Reading tape','');
            Tp := 0;
            END;
          b[i] := TapeBlock[Tp];
          INC(Tp);
          (*$T-*) CheckSum := CheckSum + ADDRESS(ORD(b[i])) (*$T=*) ;
          END;
END ReadTape;

PROCEDURE  EndReadTape;

BEGIN   Check(Close(Tape,FALSE),'Closing tape','');
END EndReadTape;

PROCEDURE  Scan( VAR fn : ARRAY OF CHAR);  (* Recursive directory scan *)

VAR     h, fh : Handle;
        rs : Status;
        db : Block;
        tm : Time;
        mo : FileMode;
        sp, rsp : Line;
        i, j, k : CARDINAL;
        head : Hdr;
        tail : Trailer;
        blk : Block;

BEGIN   Copy(fn,rsp);
        Insert(Root,rsp,0);
        i := Find(Hyphen,rsp);
        j := Length(rsp);
        IF i=j  THEN Insert("-DIRECTORY.SYS",rsp,j) END;
        Check(Open(rsp,FALSE,h),'Opening directory: ',rsp);
        Check(GetMode(h,tm,mo,i),'Checking mode of: ',rsp);
        IF NOT (DIR IN mo) THEN Exit(BadDirect) END;
        LOOP rs := ReadBlock(h,db.a);
             IF rs <> Ok THEN EXIT END;
             FOR i := 0 TO 15 DO
               WITH db.d[i] DO
                 j := Length(Name);
                 IF (j <> 0) AND (NOT (SYS IN Mode)) THEN
                   Copy(Name,sp);
                   Insert(Hyphen,sp,0);
                   Insert(fn,sp,0);         
                   IF CompareTime(Date,CutOff) >= 0 THEN
                     Copy(sp,rsp);
                     Insert(Root,rsp,0);
                     IF DIR IN Mode
                       THEN k := 0
                       ELSE Check(FastSize(Root,First,k),'Sizing file: ',rsp)
                       END;
                     WITH head DO
                       Copy(sp,h.Name);
                       h.Size := k;
                       h.Owner := Owner;
                       h.Mode := Mode;
                       h.CDate := Date;
                       rs := GetTime(h.BDate);  
                       END;
                     CheckSum := 0;
                     WriteTape(head.a);
                     Check(Open(rsp,FALSE,fh),'Opening file: ',rsp);
                     FOR j := 1 TO k DO
                       Check(ReadBlock(fh,blk.a),'Reading file: ',rsp);
                       WriteTape(blk.a);
                       END;
                     Check(Close(fh,FALSE),'Closing file: ',rsp);
                     tail.cs := CheckSum;
                     WriteTape(tail.a);
                     IF log THEN WriteString(rsp); WriteLn; END;
                     END;
                   IF DIR IN Mode THEN Scan(sp) END;
                   END;
                 END;
               END;
             END; 
        IF (rs <> Ok) AND (rs <> EndFile) THEN 
          Check(rs,'Reading directory: ',fn) END;
        rs := Close(h,FALSE);
END Scan;

PROCEDURE  BackupFiles;

VAR     eof : Trailer;
        d, t : Line;

BEGIN   Component(ComLine,Param);
        Component(ComLine,d);
        Component(ComLine,t);
        Insert("  ",t,0);
        Insert(d,t,0);
        Component(ComLine,d);
        log := CAP(d[0]) = 'L';
        IF Param[0] = 0C THEN
          WriteString('Root directory: ');
          ReadString(Param);
          WriteString('Since: ');
          ReadString(t);
          log := TRUE;
          END;
        IF Param[0] = '*' THEN Param[0] := 0C END;
        IF Param[0] = 0C THEN Copy("$CURRENT",Param) END;
        Check(Translate(Param,Root),'Translating root specification: ',Param);
        WITH CutOff DO 
          day := 0;
          min := 0;
          csec := 0;
          END;
        IF NOT ScanTime(t,CutOff) THEN Err("Invalid date syntax") END;
        StartWriteTape;
        Scan("");
        eof.cs := 0;
        WriteTape(eof.a);  (* Put a NUL in the next header *)
        EndWriteTape;
END BackupFiles;

PROCEDURE  FindScan( VAR fn : ARRAY OF CHAR);  (* Recursive directory scan *)

VAR     h, fh : Handle;
        rs : Status;
        db : Block;
        tm : Time;
        mo : FileMode;
        sp, rsp : Line;
        i, j, k : CARDINAL;

BEGIN   Copy(fn,rsp);
        Insert(Root,rsp,0);
        i := Find(Hyphen,rsp);
        j := Length(rsp);
        IF i=j  THEN Insert("-DIRECTORY.SYS",rsp,j) END;
        Check(Open(rsp,FALSE,h),'Opening directory: ',rsp);
        Check(GetMode(h,tm,mo,i),'Checking mode of: ',rsp);
        IF NOT (DIR IN mo) THEN Exit(BadDirect) END;
        LOOP rs := ReadBlock(h,db.a);
             IF rs <> Ok THEN EXIT END;
             FOR i := 0 TO 15 DO
               WITH db.d[i] DO
                 j := Length(Name);
                 IF (j <> 0) AND (NOT (SYS IN Mode)) THEN
                   Copy(Name,sp);
                   Insert(Hyphen,sp,0);
                   Insert(fn,sp,0);
                   IF CompareTime(Date,CutOff) >= 0 THEN
                     Copy(sp,rsp);
                     Insert(Root,rsp,0);
                     WriteLine(os,rsp);
                     END;
                   IF DIR IN Mode THEN FindScan(sp) END;
                   END;
                 END;
               END;
             END; 
        IF (rs <> Ok) AND (rs <> EndFile) THEN 
          Check(rs,'Reading directory: ',fn) END;
        rs := Close(h,FALSE);
END FindScan;

PROCEDURE  FindFiles;    (* Make a list of files that would be selected *)

VAR     eof : Trailer;
        d, t, f : Line;

BEGIN   Component(ComLine,Param);
        Component(ComLine,d);
        Component(ComLine,t);
        Component(ComLine,f);
        Insert("  ",t,0);
        Insert(d,t,0);
        IF Param[0] = 0C THEN
          WriteString('Root directory: ');
          ReadString(Param);
          WriteString('Since: ');
          ReadString(t);
          WriteString('Listing file: ');
          ReadString(f);
          END;                
        IF Param[0] = '*' THEN Param[0] := 0C END;
        IF Param[0] = 0C THEN Copy("$CURRENT",Param) END;
        Check(Translate(Param,Root),'Translating root specification: ',Param);
        WITH CutOff DO 
          day := 0;
          min := 0;
          csec := 0;
          END;
        IF NOT ScanTime(t,CutOff) THEN Err("Invalid date syntax") END;
        IF f[0] = 0C THEN Copy("$TI",f) END;
        CreateStream(os,f);
        IF NOT Done THEN Err("Unable to create listing file") END;
        FindScan("");
        CloseStream(os);
END FindFiles;

PROCEDURE  Restore;

VAR     head : Hdr;
        blk : ARRAY [0..511] OF CHAR;
        tail : Trailer;
        cs : ADDRESS;
        ds, p, i : CARDINAL;
        pattern, resnam : Line;
        h : Handle;
        ok, log : BOOLEAN;

BEGIN   Component(ComLine,pattern);
        Component(ComLine,Param);
        Component(ComLine,resnam);
        log := CAP(resnam[0]) = 'L';
        IF Param[0] = 0C THEN
          WriteString('Tape directory: ');
          ReadString(pattern);
          WriteString('Target directory: ');
          ReadString(Param);
          log := TRUE;
          END;
        IF pattern[0] = '*' THEN pattern[0] := 0C END;
        IF Param[0] = '*' THEN Param[0] := 0C END;
        Check(Translate(Param,Root),'Translating target specification','');
        ds := Length(pattern);
        IF (ds > 0) AND (pattern[ds-1] = '-') THEN
          pattern[ds-1] := 0C;
          DEC(ds);
          END;      
        StartReadTape;
        LOOP CheckSum := 0;
             ReadTape(head.a);
             IF head.h.Name[0] = 0C THEN EXIT END;
             Copy(head.h.Name,resnam);
             ok := (ds = 0);
             IF NOT ok THEN
               p := Find(pattern,resnam);
               IF p <= 1 THEN
                 Remove(resnam,0,ds+p);
                 ok := (resnam[0] <> 0C);
                 END;
               END;
             IF ok THEN 
               Insert(Root,resnam,0);           
               Check(Create(resnam,head.h.Mode,h),'Creating: ',resnam);
               Check(SetMode(h,head.h.Mode,head.h.Owner),'SetMode: ',resnam);
               Check(SetFileDate(h,head.h.CDate),'SetFileDate: ',resnam);
               END;
             FOR i := 1 TO head.h.Size DO 
               ReadTape(blk);
               IF ok THEN Check(WriteBlock(h,blk,TRUE),'Writing: ',resnam) END;
               END;
             cs := CheckSum;
             ReadTape(tail.a);
             IF tail.cs <> cs THEN Err("Checksum error") END;
             IF ok THEN Check(Close(h,TRUE),'Closing: ',resnam) END;
             IF ok AND log THEN WriteString(resnam); WriteLn; END;
             END;
        EndReadTape;
END Restore;  
  
PROCEDURE  Directory;

VAR     head : Hdr;
        blk : ARRAY [0..511] OF CHAR;
        tail : Trailer;
        cs : ADDRESS;
        ds, p, i : CARDINAL;
        pattern, resnam : Line;
        h : Handle;
        ok, first : BOOLEAN;

BEGIN   Component(ComLine,pattern);
        Component(ComLine,Param);
        IF pattern[0] = 0C THEN
          WriteString('Tape directory: ');
          ReadString(pattern);
          WriteString('Listing file: ');
          ReadString(Param);
          END;
        IF pattern[0] = '*' THEN pattern[0] := 0C END;
        IF Param[0] = 0C THEN Copy("$TI",Param) END;
        CreateStream(os,Param);
        IF NOT Done THEN Err("Creating listing file") END;
        ds := Length(pattern);
        IF (ds > 0) AND (pattern[ds-1] = '-') THEN
          pattern[ds-1] := 0C;
          DEC(ds);
          END;  
        first := TRUE;
        StartReadTape;
        LOOP CheckSum := 0;
             ReadTape(head.a);
             IF head.h.Name[0] = 0C THEN EXIT END;
             IF first THEN
               FormatTime(head.h.BDate,resnam);
               Insert('Backup made: ',resnam,0);
               WriteLine(os,resnam);
               WriteLine(os,"");
               first := FALSE;
               END;
             Copy(head.h.Name,resnam);
             ok := (ds = 0);
             IF NOT ok THEN
               p := Find(pattern,resnam);
               IF p <= 1 THEN
                 Remove(resnam,0,ds+p);
                 ok := (resnam[0] <> 0C);
                 END;
               END;
             IF ok THEN   
               WITH head.h DO
                 Copy(resnam,Param);
                 Param[0] := ' ';        (* remove leading hyphen *)
                 Insert("  ",Param,0);
                 FormatCard(Size,10,5,resnam);
                 Insert(resnam,Param,0);
                 FormatCard(Owner,10,5,resnam);
                 Insert(resnam,Param,5);
                 FormatCard(CARDINAL(Mode),16,5,resnam);
                 Insert(resnam,Param,10);
                 WriteLine(os,Param);
                 END;
               END;
             FOR i := 1 TO head.h.Size DO 
               ReadTape(blk);
               END;
             cs := CheckSum;
             ReadTape(tail.a);
             IF tail.cs <> cs THEN Err("Checksum error") END;
             END;
        EndReadTape;  
        WriteLine(os,"");
        CloseStream(os);
        IF first THEN WriteString('Backup -- Empty tape'); WriteLn; END;
END Directory;

PROCEDURE  Erase;

VAR     eof : Trailer;

BEGIN   Check(Open("#C0",TRUE,Tape),'Opening tape','');
        Check(FormatTrack(Tape,0),'Erasing tape','');
        Check(Close(Tape,FALSE),'Closing tape','');
        StartWriteTape;
        eof.cs := 0;
        WriteTape(eof.a);
        EndWriteTape;
END Erase;

PROCEDURE  Verify;  (* Read the tape, verifying checksum *)

VAR     head : Hdr;
        blk : ARRAY [0..511] OF CHAR;
        tail : Trailer;
        cs : ADDRESS;
        i : CARDINAL;

BEGIN   StartReadTape;
        LOOP CheckSum := 0;
             ReadTape(head.a);
             IF head.h.Name[0] = 0C THEN EXIT END;
             FOR i := 1 TO head.h.Size DO ReadTape(blk) END;
             cs := CheckSum;
             ReadTape(tail.a);
             IF tail.cs <> cs THEN Err("Checksum error") END;
             END;
        EndReadTape;
END Verify;

PROCEDURE  Check( rs : Status; VAR m, n : ARRAY OF CHAR);

BEGIN   IF rs <> Ok THEN
          WriteString('Backup -- ');
          GetMessage(rs,Param);
          WriteString(Param);
          WriteString(", ");
          WriteString(m);
          IF HIGH(n) <> 0 THEN WriteString(n) END;
          WriteLn;
          Exit(Status(-1));
          END;
END Check;

PROCEDURE  Err( VAR s : ARRAY OF CHAR);

BEGIN   WriteString("Backup -- ");
        WriteString(s);
        WriteLn;
        Exit(Status(-1));
END Err;

PROCEDURE  Component( VAR s, t : ARRAY OF CHAR);

VAR     a : CARDINAL;

BEGIN   a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] = ' ') DO INC(a) END;
        Remove(s,0,a);
        a := 0;
        WHILE (a <= HIGH(s)) AND (s[a] > ' ') DO INC(a) END;
        Substring(s,t,0,a);
        IF t[0] = ';' THEN t[0] := 0C; RETURN END;
        Remove(s,0,a);
END Component;

BEGIN   Hyphen[0] := '-'; Hyphen[1] := 0C;
        GetCommand(ComLine);
        inter := FALSE;
        Component(ComLine,Param);
        LOOP IF Param[0] = 0C THEN 
               WriteString('Backup> ');
               inter := TRUE;
               ReadString(ComLine);
               IF Result() <> Ok THEN EXIT END;
               Component(ComLine,Param);
               END;
             CASE CAP(Param[0]) OF
               0C  :               |
               ' ' :               |
               'B' : BackupFiles   |
               'R' : Restore       |
               'E' : Erase         |
               'F' : FindFiles     |
               'D' : Directory     |
               'V' : Verify        |
               'X' : EXIT
               ELSE WriteString('Backup -- Unknown function');
                    WriteLn;
               END;
             IF NOT inter THEN EXIT END;
             Param[0] := 0C;
             END;
END Backup.
