MODULE Tree;  (* Display a directory tree subject to various selections *)

(* Syntax:   Tree root [opt[="val"]]*

   where:    root   a (possibly) wildcard specification for the root of
                    a tree.

             opt    single letter selection (or formatting) options
                    possibly with an associated value

   options:  a="date"   all files after date
             b="date"   all files before date
             d          show directories only
             f          show regular files only
             m="pat"    files must match a wildcard pattern
             o="file"   write the output on file
             p="text"   prefix each line with text
             s          print statistics at end of listing   
             u="numb"   show files with this owner number only
             v          visible files only
             z          append output to file (if o in effect)
*)

FROM R2SysCalls IMPORT WriteString, WriteLn, FileMode, ModeBits, Time, 
                       GetTime, GetMessage, Status, GetUser, GetPID;
FROM R2Directories IMPORT NMode, FullName, SetWild, NextWild, WildMatch,
                          NName, NDate, NOwner;
FROM R2Streams IMPORT Stream, CreateStream, CloseStream, SetPos, 
                      OpenStream, Done, WriteLine;
FROM R2NumbIO IMPORT WriteCard;
FROM R2Format IMPORT ScanCard;
FROM R2Strings IMPORT Concatenate, Copy;
FROM R2TimeSubs IMPORT CompareTime, ScanTime;
FROM R2Storage IMPORT ALLOCATE, DEALLOCATE;
FROM Commands IMPORT GetArg;
FROM Errors IMPORT SetName, Error;

PROCEDURE  Display( VAR tree : ARRAY OF CHAR);

TYPE    ptr = POINTER TO drec;
        drec = RECORD d : ARRAY [0..79] OF CHAR;
                      n : ptr;
               END;

VAR     tmp : ARRAY [0..79] OF CHAR;
        hd, tl, z : ptr;
        go, isd : BOOLEAN;

BEGIN   hd := NIL; tl := NIL;
        IF SetWild(tree) <> Ok THEN RETURN END;
        INC(dirs);
        WHILE NextWild() = Ok DO
          isd := (DIR IN FileMode(NMode)) AND NOT (SYS IN FileMode(NMode));
          IF isd THEN
            NEW(z);
            WITH z^ DO
              n := NIL;
              IF tl = NIL
                THEN hd := z;
                ELSE tl^.n := z;
                END;
              tl := z;
              Copy(FullName,tmp);
              Concatenate(tmp,"-*",d);
              END;
            END;    
          go := TRUE;
          go := go AND (CompareTime(NDate,b) <= 0);
          go := go AND (CompareTime(a,NDate) <= 0);
          go := go AND ((isd AND NOT fonly) OR (NOT isd AND NOT donly));
          go := go AND WildMatch(match,NName);
          go := go AND ( NOT ucheck OR (NOwner = owner));
          go := go AND ( NOT vcheck OR (NOwner = user) OR (user = 0) OR
                         ODD(NMode));
          IF go THEN
            INC(fils);
            Concatenate(prefix,FullName,tmp);
            IF tt
              THEN WriteString(tmp); WriteLn
              ELSE WriteLine(os,tmp)
              END;
            END;
          END;
        WHILE hd <> NIL DO
          z := hd;
          WITH z^ DO
            hd := n;
            Display(d);
            DISPOSE(z);
            END;
          END;
END Display;

VAR     root, match, prefix, outfile, tmp, val : ARRAY [0..79] OF CHAR;
        fonly, donly, appfil, stats, ucheck, vcheck, tt, dum : BOOLEAN;
        a, b : Time;
        fils, dirs, owner, user, pid : CARDINAL;
        rs : Status;  
        os : Stream;

BEGIN   fils := 0;
        dirs := 0;
        a.day := 0; a.min := 0; a.csec := 0;
        root[0] := '*'; root[1] := 0C; match[0] := '*'; match[1] := 0C;
        fonly := FALSE; donly := FALSE; tt := TRUE; vcheck := FALSE;
        prefix[0] := 0C;
        appfil := FALSE;
        stats := FALSE;
        ucheck := FALSE;
        rs := GetTime(b);
        rs := GetPID(pid);
        rs := GetUser(pid,user);
        SetName("Tree");
        GetArg("",root,val,FALSE,FALSE);
        LOOP  val[0] := 0C;  tmp[0] := 0C;
              GetArg("",tmp,val,FALSE,TRUE);
              IF tmp[0] = 0C THEN EXIT END;
              CASE CAP(tmp[0]) OF
                 'A' : dum := ScanTime(val,a);
               | 'B' : dum := ScanTime(val,b);
               | 'D' : donly := TRUE;
               | 'F' : fonly := TRUE;
               | 'M' : IF val[0] <> 0C THEN Copy(val,match) END;
               | 'O' : tt := FALSE;  Copy(val,outfile);
               | 'P' : IF val[0] <> 0C THEN Concatenate(val,"  ",prefix) END;
               | 'S' : stats := TRUE;
               | 'U' : ucheck := TRUE;
                       IF NOT ScanCard(val,10,owner) THEN owner := user END;
               | 'V' : vcheck := TRUE;
               | 'Z' : appfil := TRUE;
                ELSE Error("Unknown option",tmp);
                END;
              END;
        IF fonly AND donly THEN Error("Inconsistent options","") END;
        IF NOT tt THEN
          IF appfil
            THEN OpenStream(os,outfile,TRUE);
                 IF NOT Done THEN Error("Cannot open",outfile) END;
                 SetPos(os,0FFFFFFFFH);
            ELSE CreateStream(os,outfile);
                 IF NOT Done THEN Error("Cannot create",outfile) END;
            END;
          END;
        Display(root);
        IF NOT tt THEN CloseStream(os) END;
        IF tt AND stats THEN
          WriteLn;
          WriteString("Total of ");
          WriteCard(fils,1);
          WriteString(" files in ");
          WriteCard(dirs,1);
          IF dirs = 1 
            THEN WriteString(" directory.");
            ELSE WriteString(" directories.");
            END;
          WriteLn;
          END;
END Tree.
